/*
 *	EXAM8.C
 *
 *	Four Transputer "pipeline" prototype application WITH I/O buffer
 *	processes as described in the "EXAMPLES" section of the Logical Systems
 *	"TRANSPUTER 'C' LIBRARY DESCRIPTION" documentation.  The "pipe4.nif"
 *	file contains a suitable topology file for use with LD-NET when running
 *	this program.
 */

#include	<stdio.h>
#include	<conc.h>

#define		DATA_SETS	250	/* # of data sets to do */
#define		NUM_POINTS	2048	/* # of data points */
#define		WS_SIZE		8192	/* Workspace size/process */

int		comp_data[NUM_POINTS];	/* Buffer for compute */
int		in_data[NUM_POINTS];	/* Buffer for input */
int		out_data[NUM_POINTS];	/* Buffer for output */

inbuf(Process *p)			/* Root node: Write data to pipeline */
	{
	int			i;

	for(i = 0; i < DATA_SETS; i++)	/* Random data */
		{
		ChanOut(LINK1OUT,
			(char *) out_data,sizeof(out_data));
		}
	}

outbuf(Process *p)			/* Root node: Read data from pipe */
	{
	int			i;

	for(i = 0; i < DATA_SETS; i++)	/* Read and toss */
		ChanIn(LINK2IN,(char *) in_data,sizeof(in_data));
	}

pinbuf(Process *p,Channel *cin)
	{		/* Pipeline node: Read from buffer process */
	while(1)
		{
		ChanIn(LINK2IN,(char *) in_data,sizeof(in_data));
		ChanOut(cin,(char *) in_data,sizeof(in_data));
		}
	}

poutbuf(Process *p,Channel *cout)
	{		/* Pipeline node: Write to buffer process */
	while(1)
		{
		ChanIn(cout,(char *) out_data,sizeof(out_data));
		ChanOut(LINK1OUT,(char *) out_data,
			sizeof(out_data));
		}
	}

main()
	{
	if(_node_number == 1)		/* Root node */
		{
		Process		*in,*out;

		if((in = ProcAlloc(inbuf,WS_SIZE,0)) == NULL)
			printf("No memory for process 'inbuf'\n");
		if((out = ProcAlloc(outbuf,WS_SIZE,0)) == NULL)
			printf("No memory for process 'outbuf'\n");
		ProcPar(in,out,NULL);	/* Start I/O processes */
		ProcFree(in);		/* Return heap storage */
		ProcFree(out);
		}
	else
		{
		Channel		*cin,*cout;
		Process		*in,*out;

		if((cin = ChanAlloc()) == NULL)
			exit(1);	/* No memory, quit */
		if((in = ProcAlloc(pinbuf,WS_SIZE,1,cin)) == NULL)
			exit(1);	/* No memory, quit */
		ProcRun(in);		/* Run asynch. input process */
		if((cout = ChanAlloc()) == NULL)
			exit(1);	/* No memory, quit */
		if((out = ProcAlloc(poutbuf,WS_SIZE,1,cout)) == NULL)
			exit(1);	/* No memory, quit */
		ProcRun(out);		/* Run asynch. output process */

		if(_node_number == 2)	/* 1st stage */
			{
			while(1)
				{
				ChanIn(cin,(char *) comp_data,
					sizeof(comp_data));
/*
 *	Add the computation for the 1st stage of the pipeline.
 */
				ChanOut(cout,(char *) comp_data,
					sizeof(comp_data));
				}
			}
		else if(_node_number == 3)	/* 2nd stage */
			{
			while(1)
				{
				ChanIn(cin,(char *) comp_data,
					sizeof(comp_data));
/*
 *	Add the computation for the 2nd stage of the pipeline.
 */
				ChanOut(cout,(char *) comp_data,
					sizeof(comp_data));
				}
			}
		else			/* 3rd stage */
			{
			while(1)
				{
				ChanIn(cin,(char *) comp_data,
					sizeof(comp_data));
/*
 *	Add the computation for the 3rd stage of the pipeline.
 */
				ChanOut(cout,(char *) comp_data,
					sizeof(comp_data));
				}
			}
		}
	printf("All done\n");
	}
