/*
 *	EXAM9.C
 *
 *	Four Transputer "pipeline" prototype application using the FORK/JOIN
 *	concurrency primitives as described in the "EXAMPLES" section of the
 *	Logical Systems	"TRANSPUTER 'C' LIBRARY DESCRIPTION" documentation.
 *	The "pipe4.nif" file contains a suitable topology file for use with
 *	LD-NET when running this program.
 */

#include	<stdio.h>
#include	<conc.h>

#define		DATA_SETS	250	/* # of data sets to do */
#define		NUM_POINTS	2048	/* # of data points */
#define		WS_SIZE		256	/* Workspace size/process */

int		comp_data[NUM_POINTS];	/* Buffer for compute */
int		in_data[NUM_POINTS];	/* Buffer for input */
int		out_data[NUM_POINTS];	/* Buffer for output */

char		inbufws[WS_SIZE];	/* Workspace for "inbuf" */

inbuf()				/* Root node: Write data to pipeline */
	{
	int			i;

	for(i = 0; i < DATA_SETS; i++)	/* Random data */
		{
		ChanOut(LINK1OUT,
			(char *) out_data,sizeof(out_data));
		}
	}

pinbuf(Channel *cin)
	{		/* Pipeline node: Read from buffer process */
	while(1)
		{
		ChanIn(LINK2IN,(char *) in_data,sizeof(in_data));
		ChanOut(cin,(char *) in_data,sizeof(in_data));
		}
	}

poutbuf(Channel *cout)
	{		/* Pipeline node: Write to buffer process */
	while(1)
		{
		ChanIn(cout,(char *) out_data,sizeof(out_data));
		ChanOut(LINK1OUT,(char *) out_data,
			sizeof(out_data));
		}
	}

main()
	{
	if(_node_number == 1)		/* Root node */
		{
		int		i;
		Forkblk		f;
		PDes		p;

		PForkInit(f,2);		/* 2 processes on root node */
		p = PSetup(inbufws,inbuf,WS_SIZE,0);
		PFork(f,p);		/* Start "inbuf" process */
		for(i = 0; i < DATA_SETS; i++)	/* Read/toss */
			ChanIn(LINK2IN,(char *) in_data,sizeof(in_data));
		PJoin(&f);		/* Synchronize processes */
		}
	else
		{
		Channel		cin = NOPROCESS;
		Channel		cout = NOPROCESS;
		char		inws[WS_SIZE],outws[WS_SIZE];

		PRun(PSetup(inws,pinbuf,WS_SIZE,1,&cin) | 1);
		PRun(PSetup(outws,poutbuf,WS_SIZE,1,&cout) | 1);

		if(_node_number == 2)	/* 1st stage */
			{
			while(1)
				{
				ChanIn(&cin,(char *) comp_data,
					sizeof(comp_data));
/*
 *	Add the computation for the 1st stage of the pipeline.
 */
				ChanOut(&cout,(char *) comp_data,
					sizeof(comp_data));
				}
			}
		else if(_node_number == 3)	/* 2nd stage */
			{
			while(1)
				{
				ChanIn(&cin,(char *) comp_data,
					sizeof(comp_data));
/*
 *	Add the computation for the 2nd stage of the pipeline.
 */
				ChanOut(&cout,(char *) comp_data,
					sizeof(comp_data));
				}
			}
		else			/* 3rd stage */
			{
			while(1)
				{
				ChanIn(&cin,(char *) comp_data,
					sizeof(comp_data));
/*
 *	Add the computation for the 3rd stage of the pipeline.
 */
				ChanOut(&cout,(char *) comp_data,
					sizeof(comp_data));
				}
			}
		}
	printf("All done\n");
	}
