/*
 *	File:	acos32.c
 *
 *	Implements T800/T414 32 bit floating point "acosf(x)/acos(x)".
 *
 *	Based on INMOS supplied code.  Modified by Logical Systems for correct
 *	ANSI "C" setting of "errno" in ERANGE/EDOM situations.
 */

#include	<inline.h>
#include	<math.h>
#include	<stddef.h>

#pragma	asm_expand			/* Recognize EDOM/ERANGE in "asm" */

float
acosf(float x)
	{
	#pragma	asm
#ifdef	DOUBLE32
	.pub	acos
acos:
#endif	DOUBLE32

#ifdef	T800
;
;	T800 32 bit version
;
	.ext	errno
	_asm_ajw(-8);
	.ldc	data
	stl	7
	.ldc	0
	stl	6
	fpldzerosn
	ldlp	[x]
	fpldnlsn
	fpgt
	stl	6
	.ldc	0
	stl	3
	ldlp	[x]
	fpldnlsn
	fpnan
	stl	3
	ldlp	[x]
	fpldnlsn
	fpuabs
	ldlp	4
	fpstnlsn
	ldl	3
	cj	@L1
	ldl	[x]
	stl	5
L0:	.ldc	EDOM			;Domain error (input is NaN)
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L14
	.align
data:	.dw	0x3f800000
	.dw	0x7f800010
	.dw	0x40490fdb
	.dw	0x3f000000
	.dw	0x39800000
	.dw	0xbf012066
	.dw	0x3f6f166c
	.dw	0xc0b18d0c
	.dw	0x40b350f0
	.dw	0x3fc90fdb
L1:	ldlp	4
	fpldnlsn
	ldl	7
	fpldnlsn
	fpgt
	cj	@L2
	ldl	7
	ldnl	1
	stl	5
	j	@L0			;Domain error (input > |1.0|)
L2:	ldlp	4
	fpldnlsn
	ldl	7
	fpldnlsn
	fpeq
	cj	@L5
	ldl	6
	cj	@L3
	ldl	7
	ldnl	2
	stl	5
	j	@L4
L3:	fpldzerosn
	ldlp	5
	fpstnlsn
L4:	j	@L14
L5:	ldlp	4
	fpldnlsn
	ldl	7
	ldnlp	3
	fpldnlsn
	fpgt
	cj	@L10
	ldlp	4
	ldl	7
	fpldnlsn
	fpldnlsn
	fpsub
	fpudivby2
	ldlp	1
	fpstnlsn
	ldlp	1
	fpldnlsn
	fpusqrtfirst
	fpusqrtstep
	fpusqrtstep
	fpusqrtlast
	ldlp	2
	fpstnlsn
	ldlp	2
	fpldnlsn
	fpumulby2
	ldlp	2
	fpstnlsn
	ldl	7
	ldnlp	4
	fpldnlsn
	ldlp	2
	fpldnlsn
	fpgt
	cj	@L6
	ldl	2
	stl	0
	j	@L7
L6:	ldl	7
	ldnlp	5
	ldlp	1
	fpldnlsn
	fpldnlmulsn
	ldl	7
	ldnlp	6
	fpldnladdsn
	ldlp	1
	fpldnlmulsn
	ldlp	1
	ldl	7
	ldnlp	7
	fpldnlsn
	fpldnladdsn
	ldlp	1
	fpldnlmulsn
	ldl	7
	ldnlp	8
	fpldnladdsn
	fpdiv
	ldlp	2
	fpldnlmulsn
	ldlp	2
	fpldnladdsn
	ldlp	0
	fpstnlsn
L7:	ldl	6
	cj	@L8
	ldlp	0
	ldl	7
	ldnlp	2
	fpldnlsn
	fpldnlsn
	fpsub
	ldlp	5
	fpstnlsn
	j	@L9
L8:	ldl	0
	stl	5
L9:	j	@L14
L10:	ldl	7
	ldnlp	4
	fpldnlsn
	ldlp	4
	fpldnlsn
	fpgt
	cj	@L11
	ldl	4
	stl	2
	j	@L12
L11:	ldl	4
	stl	1
	ldlp	1
	ldlp	1
	fpldnlsn
	fpldnlmulsn
	ldlp	0
	fpstnlsn
	ldl	7
	ldnlp	5
	ldlp	0
	fpldnlsn
	fpldnlmulsn
	ldl	7
	ldnlp	6
	fpldnladdsn
	ldlp	0
	fpldnlmulsn
	ldlp	0
	ldl	7
	ldnlp	7
	fpldnlsn
	fpldnladdsn
	ldlp	0
	fpldnlmulsn
	ldl	7
	ldnlp	8
	fpldnladdsn
	fpdiv
	ldlp	1
	fpldnlmulsn
	ldlp	1
	fpldnladdsn
	ldlp	2
	fpstnlsn
L12:	ldl	6
	cj	@L13
	ldl	7
	ldnlp	9
	ldlp	2
	fpldnlsn
	fpldnladdsn
	ldlp	5
	fpstnlsn
	j	@L14
L13:	ldlp	2
	ldl	7
	ldnlp	9
	fpldnlsn
	fpldnlsn
	fpsub
	ldlp	5
	fpstnlsn
L14:	ldlp	5
	fpldnlsn
	_asm_ajw(8);
#else	/* T800 */
;
;	T414 32 bit version
;
#endif	/* T800 */
	#pragma	endasm
	}
