/************************************************************************/
/*									*/
/*	File:	addfree.c						*/
/*									*/
/*	Add a selected region of memory to the free storage pool.	*/
/*									*/
/*	void	addfree(void *ptr, size_t size)				*/
/*									*/
/*	Where "ptr" points to the beginning of the region to add and	*/
/*	"size" is the number of bytes to include.  Note that both	*/
/*	values are rounded to the nearest 4 byte alignment break.	*/
/*									*/
/*	Uses an optimized boundary tag algorithm with an in-use		*/
/*	overhead of one word per block.					*/
/*									*/
/*	Copyright (C) 1988 by Logical Systems.				*/
/*									*/
/*	Last update:	1/26/88						*/
/*									*/
/************************************************************************/

#define		NOTMALLOC
#include	"../malloc.c"		/* Get dynamic memory types etc. */

void
addfree(ptr,size)
	void	*ptr;
	size_t	size;
	{
	register struct	header	*hp;

	if((ptr == NULL) || (size < (size_t) (sizeof(struct header) +
		(2 * sizeof(struct header *)))))
		{
		return;
		}
	hp = (struct header *) ptr;
/*
 *	Ensure block starting address is word aligned.  If not increase it up
 *	to the next correct boundary.
 */
	if((size_t) hp & (sizeof(size_t) - 1))
		{
		hp = (struct header *)
			((char *) ((size_t) hp & (~(sizeof(size_t) - 1))) +
				sizeof(size_t));
		size -= ((char *) hp - (char *) ptr);
		}
/*
 *	Modify the block size to take into account the space taken by the
 *	upper terminating tag word.
 */
	size -= sizeof(size_t);
/*
 *	Verify that the block size also satisfies the ALIGNMENT constraint, if
 *	not reduce the block size to the next lower boundary.
 */
	if(size & (ALIGNMENT - 1))
		size = (size & (~(ALIGNMENT - 1)));
/*
 *	After all this make sure enough storage is available to make this block
 *	satisfy the minimum size constraint on an allocated block.
 */
	if(size < (size_t) (sizeof(struct header) + sizeof(struct header *)))
		return;
/*
 *	Block looks OK, set the length and upper tag in so that the "free" call
 *	can add it to the free storage list.
 */
	hp->size = size | (THISINUSE | PREVINUSE);
	*((size_t *) ((char *) hp + size)) = (THISINUSE | PREVINUSE);
	free((void *) ((char *) hp + sizeof(struct header *)));
	}
