/*
 *	File:	atan232.c
 *
 *	Implements T800/T414 32 bit floating point "atan2f(x,y)/atan2(x,y)".
 *
 *	Based on INMOS supplied code.  Modified by Logical Systems for correct
 *	ANSI "C" setting of "errno" in ERANGE/EDOM situations.
 */

#include	<inline.h>
#include	<math.h>
#include	<stddef.h>

#pragma	asm_expand			/* Recognize EDOM/ERANGE in "asm" */

float
atan2f(float x,float y)
	{
	#pragma	asm
#ifdef	DOUBLE32
	.pub	atan2
atan2:
#endif	DOUBLE32

#ifdef	T800
;
;	T800 32 bit version
;
	.ext	atanf
	.ext	errno
	_asm_ajw(-8);
	.ldc	data
	stl	7
	.ldc	0
	stl	5
	.ldc	0
	stl	4
	.ldc	0
	stl	3
	.ldc	0
	stl	2
	ldlp	[y]
	fpldnlsn
	fpnan
	stl	5
	ldlp	[y]
	fpldnlsn
	fpnotfinite
	stl	3
	ldlp	[x]
	fpldnlsn
	fpnan
	stl	4
	ldlp	[x]
	fpldnlsn
	fpnotfinite
	stl	2
	ldl	5
	eqc	0
	cj	@L0
	ldl	4
	cj	@L2
L0:	ldlp	[y]
	ldlp	[x]
	fpldnlsn
	fpldnladdsn
	ldlp	6
	fpstnlsn
L1:	.ldc	EDOM			;Domain error (input is NaN)
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L12
	.align
data:	.dw	0x7f800010
	.dw	0x3fc90fdb
	.dw	0xbfc90fdb
	.dw	0x40490fdb
L2:	ldl	3
	cj	@L3
	ldl	2
	cj	@L3
	ldl	7
	ldnl	0
	stl	6
	j	@L1			;Domain error (both inputs infinite)
L3:	ldlp	[y]
	fpldnlsn
	fpldzerosn
	fpeq
	cj	@L9
	ldlp	[x]
	fpldnlsn
	fpldzerosn
	fpeq
	cj	@L5
	ldl	7
	ldnl	0
	stl	6
	j	@L1			;Domain error (both inputs zero)
L5:	ldlp	[x]
	fpldnlsn
	fpldzerosn
	fpgt
	cj	@L6
	ldl	7
	ldnl	1
	stl	6
	j	@L12
L6:	ldl	7
	ldnl	2
	stl	6
	j	@L12
L9:	ldlp	[y]
	fpldnlsn
	fpldzerosn
	fpgt
	cj	@L10
	ldlp	[y]
	ldlp	[x]
	fpldnlsn
	fpldnlsn
	fpdiv
	ldlp	1
	fpstnlsn
	ldl	1
#ifdef	STATIC_LINK
	ldl	[?slb]
#endif	STATIC_LINK
	call	@atanf
	ldlp	6
	fpstnlsn
	j	@L12
L10:	ldlp	[y]
	ldlp	[x]
	fpldnlsn
	fpldnlsn
	fpdiv
	fpuabs
	ldlp	0
	fpstnlsn
	ldl	0
#ifdef	STATIC_LINK
	ldl	[?slb]
#endif	STATIC_LINK
	call	@atanf
	ldlp	1
	fpstnlsn
	ldlp	[x]
	fpldnlsn
	fpldzerosn
	fpgt
	cj	@L11
	ldlp	1
	ldl	7
	ldnlp	3
	fpldnlsn
	fpldnlsn
	fpsub
	ldlp	6
	fpstnlsn
	j	@L12
L11:	ldl	7
	ldnlp	3
	ldlp	1
	fpldnlsn
	fpldnlsn
	fpsub
	ldlp	6
	fpstnlsn
L12:	ldlp	6
	fpldnlsn
	_asm_ajw(8);
#else	/* T800 */
;
;	T414 32 bit version
;
#endif	/* T800 */
	#pragma	endasm
	}
