/************************************************************************/
/*									*/
/*	File:	atol.c							*/
/*									*/
/*	Contains "atoi","atol","strtol" and "strtoul" functions.	*/
/*									*/
/*	Copyright 1987 by Logical Systems.				*/
/*									*/
/*	Written by Kirk Bailey, last update 12/15/87.			*/
/*									*/
/************************************************************************/

#include	<ctype.h>
#include	<stdlib.h>
#include	<limits.h>

typedef	int		BOOL;
#define	FALSE		0
#define	FOREVER		for(;;)
typedef	long		SLONG;
#define	TRUE		1
typedef	unsigned long	ULONG;

#ifdef	DEBUG
/* Local to MSC */
#define	ERANGE		1
#ifdef	LATER
#define	LONG_MAX	((SLONG) 0x7FFFFFFF)	/* (2**31)-1, max SLONG */
#define	LONG_MIN	((SLONG) 0x80000000)	/* (2**31), min SLONG */
#define	ULONG_MAX	((ULONG) 0xFFFFFFFF)	/* (2**32)-1, max ULONG */
#endif	/* LATER */
#define	const	
#define	NULL	0

long	strtol();
char	*xatol();

main()
	{
	char	buf[81],*cptr;
	int	base;
	long	result;

	for(;;)
		{
		printf("Base:");
		base = atoi(gets(buf));
		printf("Number:");
		gets(buf);
		printf("strtol = %ld,\n",strtol(buf,&cptr,base));
		printf("Remaining string[%s],errno=%d\n",cptr,errno);
		errno = 0;
		cptr = xatol(buf,&result,base,FALSE);
		printf("xatol = %ld, hex = %.8lX\n",result,result);
		printf("Remaining string[%s],errno=%d\n",cptr,errno);
		errno = 0;
		cptr = xatol(buf,&result,base,TRUE);
		printf("Unsigned xatol = %lu, hex = %.8lX\n",result,result);
		printf("Remaining string[%s],errno=%d\n",cptr,errno);
		}
	}
#endif	/* DEBUG */
/*
 *	Determine if a character can be represented in the given base.
 */
static	BOOL
isbdigit(digit,base)
	int	digit;
	int	base;
	{
	if(base <= 10)
		{
		if((digit >= '0') && (digit < ('0' + base)))
			return (TRUE);
		else
			return (FALSE);
		}
	else
		{
		digit = (islower(digit) ? toupper(digit) : digit);
		if(isdigit(digit) ||
			((digit >= 'A') && (digit < ('A' + base - 10))))
			{
			return (TRUE);
			}
		else
			return (FALSE);
		}
	}

/*
 *	Convert an ASCII character which is assumed to lie within the base
 *	range of 2-36 into the corresponding binary equivalent.
 */
static	int
tobdigit(digit)
	int	digit;
	{
	if(isdigit(digit))
		return (digit - '0');
	else
		return ((islower(digit) ? toupper(digit) : digit) - 'A' + 10);
	}

/*
 *	Convert a numeric string in the specified base to a signed or unsigned
 *	long as directed.
 *
 *	Parameters:
 *			"parse_ptr"	A pointer to the input string.
 *			"result"	A pointer to where to store the result.
 *			"base"		Base of input number (ala "strtol").
 *			"to_unsigned"	TRUE if result is to be unsigned.
 */
static	char	*
xatol(parse_ptr,result,base,to_unsigned)
	char	*parse_ptr;
	long	*result;
	int	base;
	BOOL	to_unsigned;
	{
	extern	int	errno;

	register char	*cptr;
	register int	c;
	register ULONG	ul;
	BOOL	lead_zero,done;
	char	*old_cptr;
	int	sign;
	ULONG	over_div,over_rem;
/*
 *	Ensure legal base.
 */
	if((base != 0) && ((base < 2) || (base > 36)))
		{
		*result = 0;
		errno = ERANGE;
		return (parse_ptr);
		}
/*
 *	Initialize stuff and skip leading whitespace.
 */
	ul = 0;				/* Result accumulator */

	cptr = parse_ptr;
	while(isspace(*cptr)) cptr++;
	old_cptr = cptr;
/*
 *	Process sign indication for number (if positive or signed).
 */
	if(((c = *cptr++) == '-') && (! to_unsigned))
		{
		sign = -1;
		c = *cptr++;
		}
	else if(c == '+')
		{
		sign = 1;
		c = *cptr++;
		}
	else	sign = 1;
/*
 *	See if we have a leading zero.
 */
	if(c == '0')
		{
		lead_zero = TRUE;
		c = *cptr++;
		}
	else
		lead_zero = FALSE;
/*
 *	Check for hexadecimal prefix ("0x" or "0X").
 */
	if(lead_zero && ((base == 0) || (base == 16)) &&
		((c == 'x') || (c == 'X')))
		{
		base = 16;		/* We know its hexadecimal now */
		c = *cptr++;
		}
/*
 *	Check for octal prefix (leading "0").
 */
	else if(lead_zero && (base == 0))
		base = 8;		/* We know its octal */
/*
 *	If the base is still unknown it must be decimal.
 */
	else if(base == 0)
		base = 10;		/* Must be decimal */
/*
 *	We've determined the base, now convert the number.  First compute the
 *	overflow limit.
 */
	if(to_unsigned)
		{
		over_div = ULONG_MAX / base;
		over_rem = ULONG_MAX % base;
		}
	else if(sign > 0)
		{
		over_div = LONG_MAX / base;
		over_rem = LONG_MAX % base;
		}
	else
		{
		over_div = (((ULONG) LONG_MAX) + 1) / base;
		over_rem = (((ULONG) LONG_MAX) + 1) % base;
		}
	done = FALSE;
	while(isbdigit(c,base))
		{
		if(! done)
			{
			c = tobdigit(c);
			if((ul > over_div) ||
				((ul == over_div) && (c > over_rem)))
				{
				if(to_unsigned)
					*result = ULONG_MAX;
				else if(sign > 0)
					*result = LONG_MAX;
				else
					*result = LONG_MIN;
				errno = ERANGE;
				done = TRUE;
				}
			else
				ul = (ul * base) + c;
			}
		c = *cptr++;
		}
	if(! done)
		{
		if(sign < 0)
			ul = -ul;
		*result = ul;
		}
	if(--cptr == old_cptr)		/* Any conversion performed? */
		cptr = parse_ptr;;	/* No, return original pointer */
	return (cptr);
	}
#ifndef	DEBUG
/*
 *	atoi
 */
int
atoi(cptr)
	const	char	*cptr;
	{
	long	result;

	(void) xatol(cptr,&result,10,FALSE);
	return ((int) result);
	}

/*
 *	atol
 */
long
atol(cptr)
	const	char	*cptr;
	{
	long	result;

	(void) xatol(cptr,&result,10,FALSE);
	return (result);
	}

/*
 *	strtol
 */
long
strtol(nptr,endptr,base)
	const	char	*nptr;
	char	**endptr;
	int	base;
	{
	long	result;

	if(endptr != NULL)
		*endptr = xatol(nptr,&result,base,FALSE);
	else
		(void) xatol(nptr,&result,base,FALSE);
	return (result);
	}

/*
 *	strtoul
 */
ULONG
strtoul(nptr,endptr,base)
	const	char	*nptr;
	char	**endptr;
	int	base;
	{
	long	result;

	if(endptr != NULL)
		*endptr = xatol(nptr,&result,base,TRUE);
	else
		(void) xatol(nptr,&result,base,TRUE);
	return ((ULONG) result);
	}
#endif	/* !DEBUG */
