/*
 *	fprintf,printf,vfprintf,vprintf
 *
 *	Copyright (c) 1987 by Logical Systems
 *
 *	Written by Kirk Bailey
 *
 *	Last update:	9/15/87
 */

#include	<stdio.h>
#include	<stddef.h>
#include	<stdarg.h>

/*
 *	Buffering information for "fprintf"/"printf".
 */

#define	PFBUFSIZE	90		/* "printf"/"fprintf" buffer size */

struct	pf_buf
	{
	char	buf[PFBUFSIZE];		/* Actual buffer */
	char	*ptr;			/* Current buffer pointer */
	FILE	*fp;			/* File pointer to use */
	};

/*
 *	pf_help
 *
 *	Buffering helper function for "fprintf"/"printf".
 */
static	int
pf_help(c,s)
	int	c;
	FILE	*s;
	{
	register struct	pf_buf	*sptr;

	sptr = (struct pf_buf *) s;

	if(sptr->ptr >= &(sptr->buf[PFBUFSIZE]))	/* Full buffer? */
		{
		sptr->ptr = sptr->buf;
		if(fwrite(sptr->buf,1,PFBUFSIZE,sptr->fp) != PFBUFSIZE)
			return (EOF);
		}
	*(sptr->ptr++) = c;
	return (! EOF);
	}

/*
 *	fprintf
 */
int
fprintf(FILE *fp,const char *format,...)
	{
	extern	int	_fmt();

	struct	pf_buf	info;
	int	length;
	int	rval;
	va_list	vargs;

	info.ptr = info.buf;		/* Set buffer pointer */
	info.fp = fp;			/* File pointer to use */
	va_start(vargs,format);		/* Start of var args stuff */
	rval = _fmt(pf_help,&info,format,vargs);
	va_end(vargs);			/* Cleanup for var args stuff */
/*
 *	Flush out buffer.
 */
	if((rval != EOF) && (info.ptr != info.buf))	/* Anything in buf? */
		{
		length = info.ptr - info.buf;
		if(fwrite(info.buf,1,length,fp) != length)
			return (EOF);
		}
	return (rval);
	}

/*
 *	printf
 */
int
printf(const char *format,...)
	{
	extern	int	_fmt();

	struct	pf_buf	info;
	int	length;
	int	rval;
	va_list	vargs;

	info.ptr = info.buf;		/* Set buffer pointer */
	info.fp = stdout;		/* File pointer to use */
	va_start(vargs,format);		/* Start of var args stuff */
	rval = _fmt(pf_help,&info,format,vargs);
	va_end(vargs);			/* Cleanup for var args stuff */
/*
 *	Flush out buffer.
 */
	if((rval != EOF) && (info.ptr != info.buf))	/* Anything in buf? */
		{
		length = info.ptr - info.buf;
		if(fwrite(info.buf,1,length,stdout) != length)
			return (EOF);
		}
	return (rval);
	}

/*
 *	vfprintf
 */
int
vfprintf(fp,format,vargs)
	FILE	*fp;
	const	char	*format;
	va_list	vargs;
	{
	extern	int	_fmt();

	struct	pf_buf	info;
	int	length;
	int	rval;

	info.ptr = info.buf;		/* Set buffer pointer */
	info.fp = fp;			/* File pointer to use */
	rval = _fmt(pf_help,&info,format,vargs);
/*
 *	Flush out buffer.
 */
	if((rval != EOF) && (info.ptr != info.buf))	/* Anything in buf? */
		{
		length = info.ptr - info.buf;
		if(fwrite(info.buf,1,length,fp) != length)
			return (EOF);
		}
	return (rval);
	}

/*
 *	vprintf
 */
int
vprintf(format,vargs)
	const	char	*format;
	va_list	vargs;
	{
	extern	int	_fmt();

	struct	pf_buf	info;
	int	length;
	int	rval;

	info.ptr = info.buf;		/* Set buffer pointer */
	info.fp = stdout;		/* File pointer to use */
	rval = _fmt(pf_help,&info,format,vargs);
/*
 *	Flush out buffer.
 */
	if((rval != EOF) && (info.ptr != info.buf))	/* Anything in buf? */
		{
		length = info.ptr - info.buf;
		if(fwrite(info.buf,1,length,stdout) != length)
			return (EOF);
		}
	return (rval);
	}
