/************************************************************************/
/*									*/
/*	File:	free.c (free,cfree)					*/
/*									*/
/*	Return to free storage part of the "C" heap previously		*/
/*	allocated by "calloc", "malloc" or "realloc".			*/
/*									*/
/*	Uses an optimized boundary tag algorithm with an in-use		*/
/*	overhead of one word per block.					*/
/*									*/
/*	Copyright (C) 1988 by Logical Systems.				*/
/*									*/
/*	Last update:	1/26/88						*/
/*									*/
/************************************************************************/

#define		NOTMALLOC
#include	"../malloc.c"		/* Get dynamic memory types etc. */

void
cfree(ptr)
	void	*ptr;
	{
	free(ptr);
	}

void
free(ptr)
	void	*ptr;
	{
	register size_t	bsize;
	register struct	header	*hp,*righthp,*lefthp;
	int	old_priority;

	if(ptr == NULL)
		return;

	hp = (struct header *) ((char *) ptr - sizeof(struct header *));
	bsize = hp->size & (~(THISINUSE | PREVINUSE));
	old_priority = ProcToHigh();	/* Protect critical section */
	if(hp->size & PREVINUSE)
		{
		if(*((size_t *) (righthp = (struct header *)
			((char *) hp + bsize))) & THISINUSE)
/*
 *	Both adjacent blocks busy...
 */
			{
			*((size_t *) righthp) &= (~PREVINUSE);
			hp->size &= (~THISINUSE);
			*((struct header **) ((char *) righthp -
				sizeof(struct header *))) = hp;
			hp->left = _freelist;
			hp->right = _freelist->right;
			(hp->right)->left = hp;
			(hp->left)->right = hp;
			}
		else
/*
 *	Upper (right) adjacent block is free, merge them.
 */
			{
			(righthp->left)->right = hp;
			(righthp->right)->left = hp;
			hp->left = righthp->left;
			hp->right = righthp->right;
			if(_freelist == righthp)
				_freelist = hp;
			bsize += (righthp->size & (~(THISINUSE | PREVINUSE)));
			hp->size = bsize | PREVINUSE;
			*((struct header **) ((char *) hp + bsize -
				sizeof(struct header *))) = hp;
			}
		}
	else
		{
		lefthp = *((struct header **) ((char *) hp -
			sizeof(struct header *)));
		if(*((size_t *) (righthp = (struct header *)
			((char *) hp + bsize))) & THISINUSE)
/*
 *	Lower (left) adjacent block is free, merge them.
 */
			{
			lefthp->size = ((lefthp->size &
				(~(THISINUSE | PREVINUSE))) + bsize) |
					PREVINUSE;
			*((struct header **) ((char *) righthp -
				sizeof(struct header *))) = lefthp;
			*((size_t *) righthp) &= (~PREVINUSE);
			}
		else
/*
 *	Both adjacent blocks are free, merge all three.
 */
			{
			(righthp->left)->right = righthp->right;
			(righthp->right)->left = righthp->left;
			if(_freelist == righthp)
				_freelist = righthp->left;
			bsize += ((lefthp->size & (~(THISINUSE | PREVINUSE))) +
				(righthp->size & (~(THISINUSE | PREVINUSE))));
			lefthp->size = bsize | PREVINUSE;
			*((struct header **) ((char *) lefthp + bsize -
				sizeof(struct header *))) = lefthp;
			}
		}
	if(old_priority)
		ProcToLow();		/* End of critical section */
	}
