/*
 *	nsstart.c
 *
 *	Handle initialization prior to calling "main".  This file is similar in
 *	function to the "start.c" file except this file assumes the CIO server
 *	will not be used with this program and avoids dragging the server
 *	interface library in.  A simple "printf"-like output formatting
 *	routine is provided (called "_ns_printf"), which is functionally
 *	equivalent to the "_printf" routine but only writes messages to the
 *	channel defined by "_boot_chan_out".  A entrypoint is provided to the
 *	equivalent of "exit" under these conditions (ie. "_ns_exit" which is a
 *	jump to point).
 *
 *	Copyright 1988, 1989 by Logical Systems.
 *
 *	WARNING:  THE "_ns_main" ROUTINE NEEDS MODIFICATION FOR USE WITH THE
 *	"static link" GLOBAL DATA MODEL (IT MUST SETUP THE INITIAL VALUE)!
 */

#include	<inline.h>
#include	<stdio.h>
#include	<conc.h>

#define	TRUE	1
#define	FALSE	0

#pragma	asm
	.ext	main			;Main program entry point
	.ext	_boot_chan_in		;Input boot channel address
	.ext	_boot_chan_out		;Output boot channel address
	.ext	_node_number		;User supplied network node #
	.pub	_ns_exit		;Replacement for "exit"
#pragma	endasm

/*
 *	_ns_main
 *
 *	Non-Server version of "_main".  MUST BE MODIFIED FOR USE WITH THE
 *	STATIC LINK DATA MODEL (THE INITIAL VALUE OF THE STATIC LINK MUST BE
 *	SET AND THE DATA ACCESSES MUST BE CHANGED FROM ".ldc" TO THE
 *	APPROPRIATE TECHNIQUE).
 */
void
_ns_main(void)
	{
	#pragma	asm
	ajw	-2			;A bit of temp space
;
;	At this point WS(1) is the user supplied node #.
;
;	Input channel address is initially in "C"
;
	or
	stl	0			;Input channel address in "A"
	stl	0			;Save for a bit
	ldl	0
	.ldc	_boot_chan_in
	stnl	0			;Save input address
	.ldc	4
	bcnt				;Areg will be 8 for T2, 16 for T4/T8
	ldl	0			;Get input channel address
	xor				;Compute output channel address
	.ldc	_boot_chan_out
	stnl	0			;Save output address
	ldl	1			;Processor node #
	.ldc	_node_number
	stnl	0			;Save processor node #
;
;	Dummy up the argv/argc stuff.
;
	ldlp	1
	stl	0			;argv[0] value
	.ldc	0
	stl	1			;argv[0] points to this
	.ldc	0			;Environment pointer
	ldlp	0			;ARGV
	.ldc	1			;ARGC
	call	@main
_ns_exit:
	stopp				;Kill process
	#pragma	endasm
	}

/*
 *	Write a character to the boot link.
 */
static	void
_putchar(int c)
	{
	#pragma	asm
	ldlp	[c]
	_asm_eval1(&_boot_chan_out);		;Output channel pointer address
	ldnl	0
	.ldc	1
	out
	#pragma	endasm
	}

/*
 *	Write a string to the boot link.
 */
static	void
_puts(char *s)
	{
	#pragma	asm
p_loop
	ldl	[s]
	lb				;Load byte
#ifdef	T800
	dup				;Save a copy
	stl	3
#else
	stl	3			;Save a copy
	ldl	3
#endif	/* T800 */
	cj	@done			;If '\0' terminator
	ldl	3			;Fresh copy of byte
	call	@_putchar
	ldl	[s]			;Increment string pointer
	adc	1
	stl	[s]
	j	@p_loop
done
	#pragma	endasm
	}

/*
 *	Convert an integer to a string in any base (2-36)
 */
static	char	*
itob(int n, char *s, int base)
	{
	int		c;
	int		negative;
	char		*p;
	char		*q;
	unsigned int	u;

	if((n < 0) && (base == -10))
		{
		negative = TRUE;
		u = -n;
		}
	else
		{
		negative = FALSE;
		u = n;
		}
	if(base == -10)			/* Signals signed conversion */
		base = 10;
	p = q = s;
	do
		{			/* Generate digits in reverse order */
		if((*p = u % base + '0') > '9')
			*p += ('A' - ('9' + 1));
		++p;
		} while((u /= base) > 0);
	if(negative)
		*p++ = '-';
	*p = '\0';			/* Terminate the string */
	while(q < --p)
		{			/* Reverse the digits */
		c = *q;
		*q++ = *p;
		*p = c;
		}
	return (s);
	}

/*
 *	_ns_printf
 *
 *	Equivalent to "_printf" except all output goes to the "_boot_chan_out"
 *	link unconditionally.
 */

void
_ns_printf(char *arg, ...)
	{
	char		**argp;
	int		base;
	int		c;
	char		*fmt;
	char		s[80];

	argp = &arg;

	fmt = *argp;
	while(c = *fmt++)
		{
		if(c != '%')
			{
			_putchar(c);
			continue;
			}
		while((*fmt >= '0') && (*fmt <= '9'))
			fmt++;
		switch(c = *fmt++)
			{
			case 'l':
				switch(c = *fmt++)
					{
					case 'd':
						base = -10;
						goto prt;
					case 'o':
						base = 8;
						goto prt;
					case 'u':
						base = 10;
						goto prt;
					case 'x':
						base = 16;
						goto prt;
					default:
						_putchar(c);
						continue;
					}
			case 'c':
				_putchar((int) (*++argp));
				continue;
			case 'd':
				base = -10;
				goto prt;
			case 'o':
				base = 8;
				goto prt;
			case 'u':
				base = 10;
				goto prt;
			case 'x':
				base = 16;
		prt:
				_puts(itob((int) (*++argp), s, base));
				continue;
			case 's':
				_puts(*++argp);
				continue;
			default:
				_putchar(c);
				continue;
			}
		}
	}
