/*
 *	File:	pow32.c
 *
 *	Implements T800/T414 32 bit floating point "powf(x,y)/pow(x,y)".
 *
 *	Based on INMOS supplied code.  Modified by Logical Systems for correct
 *	operation with negative values of "x" and integral values of "y" per
 *	the ANSI "C" specification.  Also modified to correctly set "errno"
 *	for the various ERANGE/EDOM conditions.  This code is more than a
 *	little obscure, good luck!  (After all, "rewriting" the output of
 *	compiler generated machine language is what "real" programmers LOVE
 *	doing when they run out of hard jobs...).
 */

#include	<inline.h>
#include	<math.h>
#include	<stddef.h>

#pragma	asm_expand			/* Recognize EDOM/ERANGE in "asm" */

float
powf(float x,float y)
	{
	#pragma	asm
#ifdef	DOUBLE32
	.pub	pow
pow:
#endif	DOUBLE32

#ifdef	T800
;
;	T800 32 bit version
;
	.ext	errno
	.ext	_scalef
	.ext	_unpkf
	_asm_ajw(-17);
	.ldc	A3
	stl	16
	.ldc	0
	stl	14
	.ldc	0
	stl	13
	.ldc	0
	stl	12
	.ldc	0
	stl	11
	ldlp	[x]
	fpldnlsn
	fpnan
	stl	14
	ldlp	[x]
	fpldnlsn
	fpnotfinite
	stl	13
	ldlp	[y]
	fpldnlsn
	fpnan
	stl	12
	ldlp	[y]
	fpldnlsn
	fpnotfinite
	stl	11
	ldl	14
	eqc	0
	cj	@L1
	ldl	12
	cj	@L2
L1:	ldlp	[x]
	ldlp	[y]
	fpldnlsn
	fpldnladdsn
	ldlp	15
	fpstnlsn
	j	@L2B			;Input(s) are NaN's set "errno" to EDOM
	.align
A1:	.dw	0x3f800000
	.dw	0x3f75257e
	.dw	0x3f6ac0c8
	.dw	0x3f60cce0
	.dw	0x3f5744fe
	.dw	0x3f4e248c
	.dw	0x3f45672a
	.dw	0x3f3d08a4
	.dw	0x3f3504f3
	.dw	0x3f2d583f
	.dw	0x3f25fed7
	.dw	0x3f1ef533
	.dw	0x3f1837f0
	.dw	0x3f11c3d3
	.dw	0x3f0b95c3
	.dw	0x3f05aac3
	.dw	0x3f000000

A2:	.dw	0x31a92436
	.dw	0xb19eab59
	.dw	0x31a8fc24
	.dw	0xb2c14fe9
	.dw	0xb1adeaf6
	.dw	0x32c12342
	.dw	0x32e75625
	.dw	0x32cf9892

A3:	.dw	0x7f800010
	.dw	0x7f800008
	.dw	0x3f800000
	.dw	0x7f800000
	.dw	0xbf800000
	.dw	0x3f3504f3
	.dw	0x3daab75c
	.dw	0x3ee2a8ed
	.dw	0x41800000
	.dw	0x00000000
	.dw	0x40dfffc0
	.dw	0x00000000
	.dw	0xc0dfffc0
	.dw	0x3d800000
	.dw	0x3d5ea8ff
	.dw	0x3e75f316
	.dw	0x3f317211
L2:	.ldc	0
	stl	14			;Once held "BaseIsNan", now "sign" flag
	fpldzerosn
	ldlp	[x]
	fpldnlsn
	fpgt
	cj	@L3
;
;	INMOS just called it a domain error if the base < 0.  We need to
;	see if the exponent value was integral before deciding whether to
;	call it a domain error or not.  If the value was integral we compute
;	what the sign of the resulting expression will be and store it in
;	local variable offset 14 (which has been initialized to "positive").
;
	ldlp	[y]			;Point to exponent
	fpldnlsn
	fpdup				;Need three copies of exponent
	fpdup
	fpint				;See if integral
	fpeq
	eqc	0
	cj	@L2C			;Jump if so (not domain error time)
L2A:	ldl	16			;Indicate domain error
	ldnl	0
	stl	15
L2B:	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L39
L2C:	ldlp	14			;Compute sign of result
	fpstnli32			;Convert to integer
	ldl	14			;Get integer value
	.ldc	1
	and
	stl	14			;Save positive/negative status in flag
;
;	We have computed the sign of the result, now negate the "base".
;
	ldlp	[x]			;Point to base
	dup				;A copy of the address
	fpldnlsn			;Get base value
	fpuabs				;Make base positive
	fpstnlsn			;Update base value
L3:	ldlp	[x]
	fpldnlsn
	fpldzerosn
	fpeq
	cj	@L8
	ldlp	[y]
	fpldnlsn
	fpldzerosn
	fpgt
	eqc	0
	cj	@L4
	j	@L2A			;Return NaN and set "errno" to EDOM
L4:	ldl	11
	cj	@L5
	ldl	16
	ldnl	1
	stl	15
	j	@L2B			;Return NaN and set "errno" to EDOM
L5:	ldl	[x]
	stl	15
	j	@L39
L8:	ldl	16
	ldnlp	2
	fpldnlsn
	ldlp	[x]
	fpldnlsn
	fpgt
	cj	@L11
	ldl	11
	cj	@L11
	ldlp	[y]
	fpldnlsn
	fpldzerosn
	fpgt
	cj	@L21A			;Return NaN and set "errno" to ERANGE
	fpldzerosn
	ldlp	15
	fpstnlsn
	j	@L39
L11:	ldlp	[x]
	fpldnlsn
	ldl	16
	ldnlp	2
	fpldnlsn
	fpeq
	cj	@L15
	ldl	11
	cj	@L13
	ldl	16
	ldnl	1
	stl	15
	j	@L21B			;Return NaN and set "errno" to ERANGE
L13:	ldl	16
	ldnl	2
	stl	15
	j	@L39
L15:	ldl	13
	cj	@L20
	ldl	16
	ldnlp	2
	fpldnlsn
	ldlp	[y]
	fpldnlsn
	fpgt
	cj	@L21A			;Return NaN and set "errno" to ERANGE
	ldlp	[y]
	fpldnlsn
	ldl	16
	ldnlp	4
	fpldnlsn
	fpgt
	eqc	0
	cj	@L2A
	fpldzerosn
	ldlp	15
	fpstnlsn
	j	@L39
L20:	ldlp	[y]
	fpldnlsn
	ldl	16
	ldnlp	2
	fpldnlsn
	fpeq
	cj	@L21
	ldl	[x]
	stl	15
	j	@L38A
L21:	ldl	11
	cj	@L24
	ldlp	[y]
	fpldnlsn
	fpldzerosn
	fpgt
	cj	@L22
L21A:	ldl	16			;Indicate range error
	ldnl	3
	stl	15
L21B:	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
	j	@L39
L22:	fpldzerosn
	ldlp	15
	fpstnlsn
	j	@L39
L24:	ldl	[x]
	ldlp	7
	ldlp	9
	call	@_unpkf
	ldlp	7
	fpldnlsn
	fpudivby2
	ldlp	7
	fpstnlsn
	ldl	9
	adc	1
	stl	9
	.ldc	0
	stl	10
	ldlp	7
	fpldnlsn
	ldl	16
	ldnlp	5
	fpldnlsn
	fpgt
	eqc	0
	cj	@L25
	.ldc	8
	stl	10
L25:	ldlp	7
	fpldnlsn
	ldl	10
	.ldc	A1
	ldnlp	4
	fpldnlsni
	fpgt
	eqc	0
	cj	@L26
	ldl	10
	adc	4
	stl	10
L26:	ldlp	7
	fpldnlsn
	ldl	10
	.ldc	A1
	ldnlp	2
	fpldnlsni
	fpgt
	eqc	0
	cj	@L27
	ldl	10
	adc	2
	stl	10
L27:	ldl	10
	.ldc	A1
	wsub
	ldnl	1
	stl	3
	ldl	10
	.ldc	2
	div
	.ldc	A2
	wsub
	ldnl	0
	stl	2
	ldlp	3
	ldlp	7
	fpldnlsn
	fpldnlsn
	fpsub
	ldlp	2
	fpldnlsn
	fpsub
	ldlp	7
	ldlp	3
	fpldnlsn
	fpldnladdsn
	fpdiv
	fpumulby2
	ldlp	1
	fpstnlsn
	ldlp	1
	ldlp	1
	fpldnlsn
	fpldnlmulsn
	ldl	16
	ldnlp	6
	fpldnlmulsn
	ldlp	1
	fpldnlmulsn
	ldlp	0
	fpstnlsn
	ldl	16
	ldnlp	7
	ldlp	0
	fpldnlsn
	fpldnlmulsn
	ldlp	0
	fpldnladdsn
	ldlp	1
	ldl	16
	ldnlp	7
	fpldnlsn
	fpldnlmulsn
	fpadd
	ldlp	1
	fpldnladdsn
	ldlp	5
	fpstnlsn
	ldl	9
	.ldc	16
	mul
	ldl	10
	adc	1
	sub
	stl	9
	ldlp	9
	fpi32tor32
	ldl	16
	ldnlp	8
	fpldnlsn
	fpdiv
	ldlp	6
	fpstnlsn
	ldlp	6
	fpldnlsn
	fpur32tor64
	ldlp	5
	fpldnlsn
	fpur32tor64
	fpadd
	ldlp	[y]
	fpldnlsn
	fpur32tor64
	fpmul
	ldlp	2
	fpstnldb
	ldlp	2
	fpldnldb
	ldl	16
	ldnlp	9
	fpldnldb
	fpgt
	cj	@L28
	.ldc	2031
	stl	8
	j	@L31
L28:	ldl	16
	ldnlp	11
	fpldnldb
	ldlp	2
	fpldnldb
	fpgt
	cj	@L29
	.ldc	-2385
	stl	8
	j	@L31
L29:	ldlp	2
	fpldnldb
	fpur64tor32
	ldlp	1
	fpstnlsn
	ldlp	1
	ldl	16
	ldnlp	8
	fpldnlsn
	fpldnlmulsn
	fprtoi32
	ldlp	8
	fpstnli32
	ldl	8
	eqc	0
	cj	@L30
	ldl	1
	stl	4
	j	@L31
L30:	ldlp	8
	fpi32tor32
	ldlp	0
	fpstnlsn
	ldlp	2
	fpldnldb
	ldl	16
	ldnlp	8
	ldlp	0
	fpldnlsn
	fpldnlsn
	fpdiv
	fpur32tor64
	fpsub
	fpur64tor32
	ldlp	4
	fpstnlsn
L31:	ldl	8
	.ldc	2030
	gt
	cj	@L34
	j	@L21A			;Return Nan & set "errno" to ERANGE
L34:	.ldc	-2384
	ldl	8
	gt
	cj	@L35
	fpldzerosn
	ldlp	15
	fpstnlsn
	j	@L39
L35:	ldlp	4
	fpldnlsn
	fpldzerosn
	fpgt
	cj	@L36
	ldl	8
	adc	1
	stl	8
	ldl	16
	ldnlp	13
	ldlp	4
	fpldnlsn
	fpldnlsn
	fpsub
	ldlp	4
	fpstnlsn
L36:	.ldc	0
	ldl	8
	gt
	cj	@L37
	ldl	8
	.ldc	16
	div
	stl	9
	j	@L38
L37:	ldl	8
	.ldc	16
	div
	adc	1
	stl	9
L38:	ldl	9
	.ldc	16
	mul
	ldl	8
	sub
	stl	10
	ldl	10
	.ldc	A1
	wsub
	ldnl	0
	stl	3
	ldl	16
	ldnlp	14
	ldlp	4
	fpldnlsn
	fpldnlmulsn
	ldl	16
	ldnlp	15
	fpldnladdsn
	ldlp	4
	fpldnlmulsn
	ldl	16
	ldnlp	16
	fpldnladdsn
	ldlp	4
	fpldnlmulsn
	ldlp	3
	fpldnlmulsn
	ldlp	3
	fpldnladdsn
	ldlp	2
	fpstnlsn
	ldl	9
	ldl	2
	ldlp	15
	call	@_scalef
L38A:	ldl	14			;See if we have to negate the result
	cj	@L39			;Jump if not (no negative base stuff)
	fpldzerosn			;Compute -result
	ldlp	15
	fpldnlsn
	fpsub
	ajw	17			;Function cleanup code
	ret
L39:	ldlp	15
	fpldnlsn
	_asm_ajw(17);
#else	/* T800 */
;
;	T414 32 bit version
;
#endif	/* T800 */
	#pragma	endasm
	}
