/*
 *	sprintf,vsprintf and _fmt
 *
 *	Copyright (c) 1987, 1988, 1989 by Logical Systems
 *
 *	Written by Kirk Bailey
 *
 *	Last update:	12/15/89
 */

#ifdef	ISTC
#include	<inline.h>
#endif	/* ISTC */

#include	<stdio.h>
#include	<stddef.h>
#include	<stdarg.h>
#include	<ctype.h>
#include	<string.h>

#define	FALSE	0
#define	TRUE	1
#define	FOREVER	for(;;)			/* Or at least awhile */
#define	MAXFLD	509			/* Max. conversion field size */

#ifdef	MSDOS

#define	DOUBLE	1			/* Use 64 bit IEEE floating point */
#define	MEMMOVE	memcpy			/* Copy which works for overlaps */

#else

#ifdef	DOUBLE32			/* Library control variable */
#define	DOUBLE	0			/* Use 32 bit IEEE floating point */
#else
#define	DOUBLE	1			/* Use 64 bit IEEE floating point */
#endif	/* DOUBLE32 */

#define	MEMMOVE	memmove			/* ANSI copy for overlaps */

#endif	/* MSDOS */

#define	INT32	long			/* 32 bit signed "integer" */
#define	UINT32	unsigned long		/* 32 bit unsigned "integer" */

/*
 *	_to_hex
 *
 *	Convert a unsigned 32 bit binary # into ASCII hexadecimal
 *	representation.
 */
static	void
_to_hex(buf,value,alt_form,precision,prefix)
	char	*buf;
	UINT32	value;
	int	alt_form,precision,prefix;
	{
	register int	i,temp;

	if(precision < 0) precision = 1;
	*buf = '\0';
	if(alt_form)
		{
		if(prefix == 'X') strcpy(buf,"0X");
		else strcpy(buf,"0x");
		buf += 2;
		}
#ifdef	T212
	for(i = 3; i >= 0; i--)
#else	/* !T212 */
	for(i = 7; i >= 0; i--)
#endif	/* T212 */
		{
		temp = (value >> (i * 4)) & 0xF;
		if((precision > i) || (temp != 0))
			{
			precision = i;
			if(temp < 10) *buf++ = temp + '0';
			else
				{
				if(prefix == 'X')
					*buf++ = temp - 10 + 'A';
				else
					*buf++ = temp - 10 + 'a';
				}
			}
		}
	*buf = '\0';
	}
/*
 *	_to_octal
 *
 *	Convert a unsigned 32 bit binary # into ASCII octal representation.
 */
static	void
_to_octal(buf,value,alt_form,precision)
	char	*buf;
	UINT32	value;
	int	alt_form,precision;
	{
	register int	i,temp;

	if(precision < 0) precision = 1;
	*buf = '\0';
#ifdef	T212
	for(i = 5; i >= 0; i--)
#else	/* !T212 */
	for(i = 10; i >= 0; i--)
#endif	/* T212 */
		{
		temp = (value >> (i * 3));
#ifdef	T212
		if(i == 5) temp &= 0x1;
#else	/* !T212 */
		if(i == 10) temp &= 0x3;
#endif	/* T212 */
		else temp &= 0x7;
		if((precision > i) || (temp != 0))
			{
			precision = i;
			if((temp != 0) && alt_form)
				*buf++ = '0';
			alt_form = FALSE;
			*buf++ = temp + '0';
			}
		}
	*buf = '\0';
	}
/*
 *	_to_udec
 *
 *	Convert a unsigned 32 bit binary # into ASCII decimal representation.
 */
static	void
_to_udec(buf,value,precision)
	char	*buf;
	UINT32	value;
	int	precision;
	{
	register int	i,temp;
	register UINT32	divisor;

#ifdef	T212
	divisor = 10000;		/* Largest possible power of 10 */

	if(precision < 0) precision = 1;
	for(i = 4; i >= 0; i--, divisor /= 10)
#else	/* !T212 */
	divisor = 1000000000;		/* Largest possible power of 10 */

	if(precision < 0) precision = 1;
	for(i = 9; i >= 0; i--, divisor /= 10)
#endif	/* T212 */
		{
		temp = value / divisor;
		value -= (temp * divisor);
		if((precision > i) || (temp != 0))
			{
			precision = i;
			*buf++ = temp + '0';
			}
		}
	*buf = '\0';
	}
/*
 *	_to_dec
 *
 *	Convert a signed 32 bit binary # into ASCII decimal representation.
 */
static	void
_to_dec(buf,value,fplus,fspace,precision)
	char	*buf;
	INT32	value;
	int	fplus,fspace,precision;
	{
	if(value < 0)
		{
		*buf++ = '-';
		value = -value;
		}
	else if(fplus)
		*buf++ = '+';
	else if(fspace)
		*buf++ = ' ';
	_to_udec(buf,(UINT32) value,precision);
	}
#ifndef	T212
/*
 *	_llshift
 *
 *	64 bit left shift helping function for _to_float.
 */
static	void
_llshift(value)
	UINT32	value[];
	{
#ifndef	ISTC
	if(value[0] & 0x80000000)
		{
		value[1] = (value[1] << 1) | 1;
		}
	else	value[1] <<= 1;
	value[0] <<= 1;
#else	/* ISTC */
	#pragma	asm
	ldl	[value]
	ldnl	1			;High order 32 bits of operand
	ldl	[value]
	ldnl	0			;Low order 32 bits of result
	.ldc	1			;Shifting by 1
	lshl
	ldl	[value]
	stnl	0			;Low 32 bits of result
	ldl	[value]
	stnl	1			;High 32 bits of result
	#pragma	endasm
#endif	/* !ISTC */
	}
/*
 *	_lrshift
 *
 *	64 bit right shift helping function for _to_float.
 */
static	void
_lrshift(value)
	UINT32	value[];
	{
#ifndef	ISTC
	if(value[1] & 1)
		{
		value[0] = (value[0] >> 1) | 0x80000000;
		}
	else	value[0] = (value[0] >> 1) & 0x7FFFFFFF;
	value[1] = (value[1] >> 1) & 0x7FFFFFFF;
#else	/* ISTC */
	#pragma	asm
	ldl	[value]
	ldnl	1			;High order 32 bits of operand
	ldl	[value]
	ldnl	0			;Low order 32 bits of result
	.ldc	1			;Shifting by 1
	lshr
	ldl	[value]
	stnl	0			;Low 32 bits of result
	ldl	[value]
	stnl	1			;High 32 bits of result
	#pragma	endasm
#endif	/* !ISTC */
	}
/*
 *	_ladd
 *
 *	64 bit addition helping function for _to_float.
 */
static	void
_ladd(result,value)
	UINT32	result[],value[];
	{
#ifndef	ISTC
	UINT32	carry,temp;

	carry = 0;
	temp = result[0] + value[0];

	if(result[0] & 0x80000000)
		{
		if((value[0] & 0x80000000) || ((temp & 0x80000000) == 0))
			carry = 1;
		}
	else
		{
		if((value[0] & 0x80000000) && ((temp & 0x80000000) == 0))
			 carry = 1;
		}
	result[0] = temp;
	result[1] = result[1] + value[1] + carry;
#else	/* ISTC */
	#pragma	asm
	.ldc	0			;Initial "carry"
	ldl	[result]
	ldnl	0			;Low order 32 bits of "result"
	ldl	[value]
	ldnl	0			;Low order 32 bits of "value"
	lsum
	ldl	[result]
	stnl	0			;Low order 32 bits of "result"
	ldl	[result]
	ldnl	1			;High order 32 bits of "result"
	ldl	[value]
	ldnl	1			;High order 32 bits of "value"
	lsum
	ldl	[result]
	stnl	1			;High 32 bits of "result"
	#pragma	endasm
#endif	/* !ISTC */
	}
/*
 *	_rlrshift
 *
 *	64 bit rounded right shift helping function for _to_float.
 */
static	void
_rlrshift(value)
	UINT32	value[];
	{
	UINT32	temp[2];

	temp[0] = value[0] & 1;
	temp[1] = 0;
	_lrshift(value);
	_ladd(value,temp);		/* Round */
	}
/*
 *	_ldiv5
 *
 *	64 bit divide by 5 function for _to_float.  The result is ROUNDED,
 *	not TRUNCATED.
 */
static	void
_ldiv5(value)
	UINT32	value[];
	{
#ifndef	ISTC
	register int	shift;
	UINT32	result[2],temp1[2],temp2[2];

	result[0] = 0;			/* Result accumulator */
	result[1] = value[1] / 5;

	temp1[0] = value[0];		/* Dividend for this pass */
	temp1[1] = value[1] % 5;

	temp2[1] = 0;

	FOREVER
		{
/*
 *	Now do the general Newton-Raphson iteration case.
 */
		for(shift = 0; temp1[1] != 0; shift++)
			_lrshift(temp1);
		temp2[0] = temp1[0] / 5;
		if(temp2[0] == 0)
			{
			if(temp1[0] % 5 > (5 / 2))
				{
				temp1[0] = 1;
				_ladd(result,temp1);
				}
			break;
			}
		temp1[0] = temp2[0];
		while(shift-- != 0)
			_llshift(temp1);
		_ladd(result,temp1);	/* Update result accumulator */
		temp1[0] = result[0];
		temp1[1] = result[1];
		_llshift(temp1);		/* Compute (current_result*5) */
		_llshift(temp1);
		_ladd(temp1,result);
		temp1[0] = ~temp1[0];	/* Compute -(current_result*5) */
		temp1[1] = ~temp1[1];
		temp2[0] = 1;
		_ladd(temp1,temp2);
		_ladd(temp1,value);	/* Compute #-(current_result*5) */
		}
/*
 *	All done, return result.
 */
	value[0] = result[0];
	value[1] = result[1];
#else	/* ISTC */
	#pragma	asm
	.ldc	0			;High order operand of 64/32 division
	ldl	[value]
	ldnl	1			;High order 32 bits of operand
	.ldc	5			;Divisor
	ldiv				;Top high order 32 bits by 5
	ldl	[value]
	stnl	1			;High order 32 bits of result
	ldl	[value]
	ldnl	0			;Low order 32 bits of result
	.ldc	5			;Divisor
	ldiv				;Old remainder in C, low 32 in B
	ldl	[value]
	stnl	0			;Low 32 bits of results
	#pragma	endasm
#endif	/* !ISTC */
	}
/*
 *	_get_digit
 *
 *	Digit processing helper function for "_to_float".
 */
static	int
_get_digit(fract,digit_count)
	UINT32	fract[];
	int	*digit_count;
	{
	int	rval;
	UINT32	temp[2];

	if(*digit_count > 0)
		{
		*digit_count -= 1;
		temp[0] = fract[0];
		temp[1] = fract[1];
		_llshift(fract);		/* Multiply by 10 */
		_llshift(fract);
		_ladd(fract,temp);
		_llshift(fract);
		rval = ((fract[1] >> 28) & 0xF) + '0';
		fract[1] &= 0x0FFFFFFF;
		}
	else
		rval = '0';
	return (rval);
	}
/*
 *	_to_float
 *
 *	Convert a floating point # to ASCII.
 *
 *	Parameters:
 *		"buf"		Buffer to write result into.
 *		"double_temp"	# to convert (either IEEE single or double).
 *		"full"		TRUE if IEEE double, else IEEE single.
 *		"c"		The conversion type (one of e,E,f,g,G).
 *		"falt"		TRUE if "#" conversion flag in effect.
 *		"fplus"		TRUE if "+" conversion flag in effect.
 *		"fspace"	TRUE if " " conversion flag in effect.
 *		"precision"	Desired precision (negative if undefined).
 */

/*
 *	The following two constants define the simulated binary floating
 *	point limit for the first stage of the conversion (fraction times
 *	power of two becomes fraction times power of 10), and the second
 *	stage (pulling the resulting decimal digits outs).
 */

#define	MAXFP1	0xFFFFFFFF		/* Largest # if first fp format */
#define	MAXFP2	0x0FFFFFFF		/* Largest # in second fp format */

static	void
_to_float(buf,double_temp,full,c,falt,fplus,fspace,precision)
	char	*buf;
	UINT32	double_temp[];
	int	full,c,falt,fplus,fspace,precision;
	{
	register int	exp,decexp;
	int	prune_zero,digit_count;
	UINT32	fract[2],ltemp[2];
/*
 *	Pull fraction and exponent out into individual fields.
 */
	if(full)			/* If IEEE double */
		{
		exp = (double_temp[1] >> 20) & 0x7FF;
		fract[1] = (double_temp[1] << 11) & 0x7FFFF800;
		fract[1] |= ((double_temp[0] >> 21) & 0x000007FF);
		fract[0] = double_temp[0] << 11;
		}
	else
		{
		exp = (double_temp[0] >> 23) & 0xFF;
		fract[1] = (double_temp[0] << 8) & 0x7FFFFF00;
		fract[0] = 0;
		}
/*
 *	Check for infinities and NaN's.
 */
	if((full && (exp == 0x7FF)) || ((! full) && (exp == 0xFF)))
		{
		if((fract[1] | fract[0]) == 0)
			{
			if((full && (double_temp[1] & 0x80000000)) ||
				((! full) && (double_temp[0] & 0x80000000)))
				strcpy(buf,"-INF");
			else
				strcpy(buf,"+INF");
			}
		else
			strcpy(buf,"NaN");
		return;
		}
/*
 *	Check for explicit zero and don't do the exponent unbiasing or
 *	addition of implicit leading 1 bit.
 */
	if((exp | fract[1] | fract[0]) != 0)
		{
		if(full)
			exp -= (1023 - 1);	/* +1 since .1 vs 1. */
		else
			exp -= (127 - 1);	/* +1 since .1 vs 1. */
		fract[1] |= 0x80000000;
		decexp = TRUE;		/* Wasn't zero */
		}
	else
		decexp = FALSE;		/* It was zero */
/*
 *	Put out a sign if needed.
 */
	if(decexp && ((full && (double_temp[1] & 0x80000000)) ||
		((! full) && (double_temp[0] & 0x80000000))))
		*buf++ = '-';
	else if(fplus)
		*buf++ = '+';
	else if(fspace)
		*buf++ = ' ';
/*
 *	Convert the fraction from a fraction times a base 2 exponent to a
 *	fraction times a base 10 exponent.
 */
	decexp = 0;
	while(exp <= -3)
		{
/*
 *	Make room in the fraction for the multiplication by 10 by
 *	shifting it right, rounding the result, and incrementing the base 2
 *	exponent.
 */
		while(fract[1] >= (MAXFP1 / 5))
			{
			_rlrshift(fract);
			exp++;		/* Increment binary exponent */
			}
		ltemp[0] = fract[0];
		ltemp[1] = fract[1];
		_llshift(fract);
		_llshift(fract);
		_ladd(fract,ltemp);
		exp++;			/* *10 = (*5) * (*2) */
		decexp--;
/*
 *	Normalize fraction.
 */
		while(fract[1] <= (MAXFP1 / 2))
			{
			_llshift(fract);
			exp--;
			}
		}
	while(exp > 0)
		{
		_ldiv5(fract);
		exp--;			/* Last /2 factor to get /10 */
		decexp++;
/*
 *	Normalize fraction.
 */
		while(fract[1] <= (MAXFP1 / 2))
			{
			_llshift(fract);
			exp--;
			}
		}
/*
 *	Number is >= .1 and < 1.  Denormalize the fraction enough to make the
 *	binary exponent be 0 (actually 4 since we are also converting to the
 *	second floating point format which is shifted four bits to the right).
 */
	while(exp < (0 + 4))
		{
		_rlrshift(fract);
		exp++;
		}
/*
 *	We've converted the #, now figure out how we should display it.
 */
	if(precision < 0)
		precision = 6;		/* Default precision if none given */
	prune_zero = FALSE;		/* Assume trailing 0's allowed */
	if((c == 'g') || (c == 'G'))
		{
		if(! falt && (precision > 0))
			prune_zero = TRUE;	/* Remove trailing 0's */
		if((decexp < (-4 + 1)) || (decexp > (precision + 1)))
			{
			if(c == 'g')	/* +1 factor above since .XX vs X.XX */
				c = 'e';
			else
				c = 'E';
			}
		else
			c = 'f';
		}
/*
 *	Figure out how many significant digits to round result to.
 */
	if(c == 'f')
		{
		exp = precision + decexp;
		if(exp < 0) exp = 0;
		}
	else
		exp = precision + 1;
	if(full)
		{
		digit_count = 16;
		if(exp > 16)
			exp = 16;
		}
	else
		{
		digit_count = 8;
		if(exp > 8)
			exp = 8;
		}
/*
 *	Compute the .5*10^X factor needed to round the result for display.
 */
	ltemp[0] = 0;
	ltemp[1] = 0x08000000;		/* .5 in this format */
	while(exp--)
		{
		_ldiv5(ltemp);		/* /5 */
		_rlrshift(ltemp);	/* /5 * /2 = /10 */
		}
/*
 *	Round the fractional part.  Note that this operation may have produced
 *	an overflow in which case we need to divide the fraction by 10 and
 *	increment the decimal exponent.
 */
	_ladd(fract,ltemp);		/* Round to display precision */
	if(fract[1] & 0xF0000000)
		{
		_ldiv5(fract);		/* /5 */
		_rlrshift(fract);	/* /5 * /2 = /10 */
		decexp++;
		}
/*
 *	Display the fractional part in 'f' or 'e'/'E' format as desired.
 */
	if(c == 'f')
		{
		if(decexp > 0)
			{
			while(decexp > 0)
				{
				*buf++ = _get_digit(fract,&digit_count);
				decexp--;
				}
			}
		else
			*buf++ = '0';
		if(falt || (precision > 0))
			{
			*buf++ = '.';
			}
		while(precision-- > 0)
			{
			if(decexp < 0)
				{
				*buf++ = '0';
				decexp++;
				}
			else
				*buf++ = _get_digit(fract,&digit_count);
			}
		}
	else				/* Must be 'e' or 'E' */
		{
		*buf = _get_digit(fract,&digit_count);
		if(*buf++ != '0')
			decexp--;
		if(falt || (precision > 0))
			*buf++ = '.';
		while(precision-- > 0)
			*buf++ = _get_digit(fract,&digit_count);
		}
/*
 *	Prune trailing zeros (and possibly ".") from fractional part if needed.
 */
	if(prune_zero)
		{
		while(*--buf == '0')	/* Remove zeroes */
			;
		if(*buf != '.')
			buf++;		/* Remove decimal point if adjacent */
		}
/*
 *	Display the exponent if the format was 'e'/'E'.
 */
	if((c == 'e') || (c == 'E'))
		{
		*buf++ = c;
		if(decexp < 0)
			{
			decexp = -decexp;
			*buf++ = '-';
			}
		else
			*buf++ = '+';
		*buf++ = (decexp / 100) + '0';
		decexp %= 100;
		*buf++ = (decexp / 10) + '0';
		decexp %= 10;
		*buf++ = decexp + '0';
		}
	*buf = '\0';
	}
#endif	/* !T212 */
/*
 *	_out_byte
 *
 *	Write a character to the "output" device.
 */
static	int
_out_byte(func,dest,c)
	int	(*func)();
	FILE	*dest;
	int	c;
	{
	return ((*func)(c,dest));
	}
/*
 *	_atoi
 *
 *	Helper routine for "_fmt" to convert a decimal ASCII value to 
 *	binary and update an associated pointer.
 */
static	int
_atoi(sptr)
	char	**sptr;
	{
	register char	*p;
	register int	i;

	i = 0;
	p = *sptr;
	p--;				/* Ptr was advanced past to begin */
	while(isdigit(*p))
		i = 10 * i + *p++ - '0';
	*sptr = p;
	return (i);
	}
/*
 *	_fmt
 *
 *	Formatting routine for fprintf,printf,sprintf.  Conforms to draft
 *	ANSI standard specification.
 */
int
_fmt(func,dest,format,vargs)
	int	(*func)();
	FILE	*dest;
	char	*format;
	va_list	vargs;
	{
	register char	*cptr;
	register int	c,i;
	char	buf[MAXFLD + 1];
	int	count,width,precision,prefix;
	int	fminus,fplus,fspace,falt,fconv,need_justifying,pad;

	char	*cptr_temp;
	INT32	*int32ptr_temp;
	INT32	int32_temp;
	UINT32	uint32_temp;
	UINT32	double_temp[2];
	
	count = 0;

	while((c = *format++) != '\0')
		{
		if(c != '%')		/* Conversion specification? */
			if(_out_byte(func,dest,c) == EOF)
				return (EOF);
			else
				count++;
		else
			{
			fminus = fplus = fspace = falt = FALSE;
			fconv = 0;	/* Conversion size modifier */
			pad = ' ';	/* Default pad character */
			precision = -1;	/* No precision specified */
/*
 *	First handle zero or more "flags".
 */
			while(strchr("-+ #",(c = *format++)) != NULL)
				{
				switch(c)
					{
					case '-':
						fminus = TRUE;
						break;
					case '+':
						fplus = TRUE;
						break;
					case ' ':
						fspace = TRUE;
						break;
					case '#':
						falt = TRUE;
						break;
					case '\0':
						return (count);
					}
				}
/*
 *	Handle an optional "width".
 */
			if(c == '*')	/* Is the width a parameter? */
				{
				width = (INT32) va_arg(vargs,INT32);
				if(width < 0)
					{
					fminus = TRUE;
					width = - width;
					}
				c = *format++;
				}
			else if(! isdigit(c))	/* No width? */
				width = 0;
			else
				{
				if(c == '0')
					pad = '0';	/* New pad character */
				width = _atoi(&format);	/* Find width */
				c = *format++;
				}
/*
 *	Handle an optional "precision".
 */
			if(c == '.')	/* Precision specified? */
				{
				if((c = *format++) == '*')
					precision = (INT32)
						va_arg(vargs,INT32);
				else
					precision = _atoi(&format);
				c = *format++;
				}
/*
 *	Handle optional "conversion size" modifier.  These are currently
 *	ignored since TCX doesn't have all the different sizes of variables
 *	that some compilers implement (no 16 bit or 64 bit stuff).
 */
			if(strchr("hlL",c) != NULL)
				{
				i = c;
				c = *format++;
				switch(i)
					{
					case 'h':
						if(strchr("diouxX",c) == NULL)
							break;
						fconv = 'h';
						break;
					case 'l':
						if(strchr("diouxX",c) == NULL)
							break;
						fconv = 'l';
						break;
					case 'L':
						if(strchr("eEfgG",c) == NULL)
							break;
						fconv = 'L';
						break;
					case '\0':
						return (count);
					}
				}
/*
 *	Finally, we recognize the conversion character.
 */
			need_justifying = FALSE;
			prefix = 0;
			switch(c)
				{
				case 'c':
					buf[0] = (INT32)
						va_arg(vargs,INT32);
					buf[1] = '\0';
					need_justifying = TRUE;
					break;
				case 'd':
				case 'i':
					int32_temp = (INT32)
						va_arg(vargs,INT32);
					_to_dec(buf,int32_temp,fplus,fspace,
						precision);
					if(fplus || fspace || (int32_temp < 0))
						prefix = 1;
					need_justifying = TRUE;
					break;
#ifndef	T212
				case 'e':
				case 'E':
				case 'f':
				case 'g':
				case 'G':
					double_temp[0] = (UINT32)
						va_arg(vargs,UINT32);
#if	DOUBLE
					double_temp[1] = (UINT32)
						va_arg(vargs,UINT32);
#endif	/* DOUBLE */
					_to_float(buf,double_temp,DOUBLE,c,
						falt,fplus,fspace,precision);
					if(fplus || fspace || (buf[0] == '-'))
						prefix = 1;
					need_justifying = TRUE;
					break;
#endif	/* !T212 */
				case 'n':
					int32ptr_temp = (INT32 *)
						va_arg(vargs,INT32 *);
					*int32ptr_temp = count;
					break;
				case 'o':
					uint32_temp = (UINT32)
						va_arg(vargs,UINT32);
					_to_octal(buf,uint32_temp,falt,
						precision);
					need_justifying = TRUE;
					break;
				case 'p':
					uint32_temp = (UINT32)
						va_arg(vargs,UINT32);
					_to_hex(buf,uint32_temp,FALSE,
#ifdef	T212
						4,(int) 'X');
#else	/* !T212 */
						8,(int) 'X');
#endif	/* T212 */
					need_justifying = TRUE;
					break;
				case 's':
					cptr_temp = (char *)
						va_arg(vargs,char *);
					if(precision < 0)
						strcpy(buf,cptr_temp);
					else
						{
						strncpy(buf,cptr_temp,
							(size_t) precision);
						buf[precision] = '\0';
						}
					need_justifying = TRUE;
					break;
				case 'u':
					uint32_temp = (UINT32)
						va_arg(vargs,UINT32);
					_to_udec(buf,uint32_temp,precision);
					need_justifying = TRUE;
					break;
				case 'x':
					uint32_temp = (UINT32)
						va_arg(vargs,UINT32);
					_to_hex(buf,uint32_temp,falt,precision,
						(int) 'x');
					if(falt) prefix = 2;
					need_justifying = TRUE;
					break;
				case 'X':
					uint32_temp = (UINT32)
						va_arg(vargs,UINT32);
					_to_hex(buf,uint32_temp,falt,precision,
						(int) 'X');
					if(falt) prefix = 2;
					need_justifying = TRUE;
					break;
				case '%':
					if(_out_byte(func,dest,'%') == EOF)
						return (EOF);
					else
						count++;
					break;
				case '\0':
					return (count);
				default:
					break;
				}
			if(need_justifying)
				{
/*
 *	Justify converted field.
 */
				c = strlen(buf);	/* Field size */
				if(fminus)	/* Left justify? */
					{
					for(i = c; i < width; i++)
						buf[i] = ' ';
					buf[i] = '\0';
					}
				else		/* Right justify */
					{
					if(c < width)
						{
						(void) MEMMOVE((buf+(width-c)),
							buf,(size_t) (c + 1));
						if(pad == ' ')
							prefix = 0;
						c = width - c + prefix;
						for(i = prefix; i < c; i++)
							buf[i] = pad;
						}
					}
/*
 *	Write converted and justified field to the "output" device.
 */
				for(cptr = buf; *cptr != '\0'; cptr++, count++)
					if(_out_byte(func,dest,*cptr) == EOF)
						return (EOF);
				}
			}
		}
	return (count);
	}

/*
 *
 *	sprintf_help
 *
 *	Helper routine for "sprintf" which copies characters into the
 *	destination buffer.
 */
static	int
sprintf_help(c,s)
	int	c;
	FILE	*s;
	{
	**((char **) s) = c;
	(*((char **) s))++;
	return (! EOF);
	}

/*
 *	sprintf
 */
int
sprintf(char *s,const char *format, ...)
	{
	int	rval;
	char	*sptr;
	va_list	vargs;

	va_start(vargs,format);		/* Start of var args stuff */
	sptr = s;
	rval = _fmt(sprintf_help,(FILE *) &sptr,format,vargs);
	*sptr = '\0';
	va_end(vargs);			/* Cleanup for var args stuff */
	return (rval);
	}

/*
 *	vsprintf
 */
int
vsprintf(s,format,vargs)
	char	*s;
	const	char	*format;
	va_list	vargs;
	{
	int	rval;
	char	*sptr;

	sptr = s;
	rval = _fmt(sprintf_help,(FILE *) &sptr,format,vargs);
	*sptr = '\0';
	return (rval);
	}
