/*
 *	File:	rand.c
 *
 *	Implements "rand()" and "srand()".
 *
 *	The seed is also used in the "frand" function to generate random
 *	floating point values.  Uses linear congruential generator specified
 *	by Draft ANSI "C" Standard" for the 32 bit version (except for removal
 *	of lowest common denominator scaling):
 *
 *	32 bit seed:
 *
 *	Seed = (Seed * 1103515245) + 12345
 *	Result = Seed & 0x7FFFFFFF
 *
 *	16 bit seed:
 *
 *	Seed = (Seed * 31413) + 13849
 *	Result = Seed & 0x7FFF
 *
 *	Copyright 1987-1989 by Logical Systems.
 */

#include	<stddef.h>

static	unsigned int	_seed = 1;	/* Seed for "frand","rand","srand" */

int
rand(/* void */)
	{
#ifdef	T212
	return ((_seed = (_seed * 31413) + 13849) & 0x7FFF);
#else	/* !T212 */
	return ((_seed = (_seed * 1103515245) + 12345) & 0x7FFFFFFF);
#endif	/* T212 */
	}

void
srand(seed)
	unsigned int	seed;
	{
	_seed = seed;
	}
