/*
 *	tstream.c
 *
 *	"stream" I/O routines for client (Transputer) side of "C" remote file
 *	I/O driver.  Uses the "tcio.c" I/O primitives.
 *
 *	Copyright 1988-1990 by Logical Systems.
 */

#include	<stdio.h>
#include	<string.h>
#include	<cio.h>

#define	UCHAR	unsigned char

extern	void	_enc_strm(UCHAR *,FILE *); /* Convert strm to 2 bytes	*/
extern	void	_enc_2b(UCHAR *,int);	/* Convert int to 2 bytes	*/
extern	void	_get_str(UCHAR *,int);	/* Get # of bytes from host	*/
extern	int	_get_1b(void);		/* Get one byte from host	*/
extern	int	_get_2b(void);		/* Get 2 bytes from host	*/
extern	int	_get_2beof(void);	/* Get 2 bytes/EOF from host	*/
extern	void	_put_str(UCHAR *,int);	/* Send "raw" string to host	*/
extern	void	_put_1b(int);		/* Send "raw" byte to host	*/
extern	void	_request_errno(void);	/* Get host "errno" value	*/
extern	void	_send_cmd(UCHAR *,int,int); /* Send command to host	*/
extern	void	_send_cmdstr(UCHAR *,int,int,int); /* Send cmd/str	*/

int
fgetc(FILE *stream)
	{
	UCHAR			buf[5];

	_enc_strm(&buf[3],stream);
	_send_cmd(buf,FN_FGETC,5);
	return (_get_2beof());
	}

char	*
fgets(char *s,int n,FILE *stream)
	{
	UCHAR			buf[7];
	int			length;

	_enc_strm(&buf[3],stream);
	if(n > MAX_MSG_DATA)
		return (NULL);
	_enc_2b(&buf[5],n);
	_send_cmd(buf,FN_FGETS,7);
	length = _get_2b() - 1;		/* Get return string length */
	if(_get_1b() == RPC_NULL)	/* Figure out if string/NULL */
		{
		(void) _get_1b();	/* Toss dummy '\0' byte */
		_request_errno();
		return (NULL);
		}
	else				/* Real string, read */
		{
		if(length > 0)
			_get_str((UCHAR *) s,length);
		}
	return (s);
	}

int
fputc(int c,FILE *stream)
	{
	UCHAR			buf[6];

	_enc_strm(&buf[3],stream);
	buf[5] = c;
	_send_cmd(buf,FN_FPUTC,6);
	return (_get_2beof());
	}

int
fputs(const char *s,FILE *stream)
	{
	UCHAR			buf[9];
	int			length;

	if(s == NULL)
		return (EOF);
	if((length = strlen(s)) > MAX_MSG_DATA)
		return (EOF);
	_enc_strm(&buf[3],stream);
	_enc_2b(&buf[5],1);
	_enc_2b(&buf[7],length);
	_send_cmdstr(buf,FN_FWRITE,9,length);
	if(length > 0)
		_put_str((UCHAR *) s,length);
	if(_get_2b() != length)
		{
		_request_errno();
		return (EOF);
		}
	return (0);
	}

int
getch(void)
	{
	UCHAR			buf[3];

	_send_cmd(buf,FN_GETCH,3);
	return (_get_2b());
	}

int
getche(void)
	{
	UCHAR			buf[3];

	_send_cmd(buf,FN_GETCHE,3);
	return (_get_2b());
	}

char	*
gets(char *s)
	{
	UCHAR			buf[3];
	int			length;

	_send_cmd(buf,FN_GETS,3);
	length = _get_2b() - 1;		/* Get return string length */
	if(_get_1b() == RPC_NULL)	/* Figure out if string/NULL */
		{
		(void) _get_1b();	/* Toss dummy '\0' byte */
		_request_errno();
		return (NULL);
		}
	else				/* Real string, read */
		{
		if(length > 0)
			_get_str((UCHAR *) s,length);
		}
	return (s);
	}

int
kbhit(void)
	{
	UCHAR			buf[3];

	_send_cmd(buf,FN_KBHIT,3);
	return (_get_2b());
	}

int
puts(const char *s)
	{
	UCHAR			buf[9];
	int			length;

	if(s == NULL)
		return (EOF);
	if((length = strlen(s)) >= MAX_MSG_DATA)
		return (EOF);
	_enc_strm(&buf[3],stdout);
	_enc_2b(&buf[5],1);
	_enc_2b(&buf[7],length + 1);
	_send_cmdstr(buf,FN_FWRITE,9,length + 1);
	if(length > 0)
		_put_str((UCHAR *) s,length);
	_put_1b('\n');
	if(_get_2b() != (length + 1))
		{
		_request_errno();
		return (EOF);
		}
	return (0);
	}
