/************************************************************************/
/*									*/
/*	File:	make.h							*/
/*									*/
/*	Include file for the MAKE program.				*/
/*									*/
/*	Written by Greg Yachuk.  Placed in public domain 10/1/88.	*/
/*	Modeled after a MAKE written by Neil Russell which in turn was	*/
/*	modeled after UN*X MAKE...					*/
/*									*/
/*	Modified/maintained by Kirk Bailey of Logical Systems.		*/
/*									*/
/*	NOTE: This program is EXPLICITLY placed in the public domain.	*/
/*		***Do what you wish with it, and have fun!***		*/
/*									*/
/************************************************************************/

/*
 *	Macro definitions.
 */

#define	VERSION		"Logical Systems Version 89.1"
#define MAKEINI 	"default.mk"


#define	READBUFSIZE	2048		/* Makefile input read buffer size */
#define	MAXFNSIZE	80		/* Maximum filename size */
#define	MAXLNSIZE	512		/* Maximum input line size */
#define	MAXNEGTIME	0x80000000
#define	MAXSYMSIZE	512		/* Maximum macro symbol size */
#define	REGISTER	register

#ifndef	TRUE
#define	TRUE		1
#define	FALSE		0
#endif

#ifdef	MSDOS
#define	FILE_SEPARATOR	"/\\"
#define	PATH_SEPARATOR	";"
#define	STDERR		stdout
#else
#define	FILE_SEPARATOR	"/"
#define	PATH_SEPARATOR	":"
#define	STDERR		stderr
#endif

#define	append_preq(t,p) (fileptr*)append_node((char**)(t),(char**)(p),sizeof(fileptr*))
#define	append_shell(t,s) (shellptr*)append_node((char**)(t),(char**)(s),sizeof(shellptr*))
#define	equal(s,t)	(!strcmp((s),(t)))
#define	get_target(t)	hash_target((t), NULL)
#define	get_file(f)	hash_file((f), NULL)
#define	tfree(t)	free((char *) t)
#define	tnew(t)		((t *) talloc((int) sizeof(t)))
#define	tstrcat(s,p)	(strcat(strcpy(talloc((int) (strlen(s)+strlen(p)+1)),(s)),(p)))
#define	tstrcpy(s)	(strcpy(talloc((int) (strlen(s)+1)), (s)))

#ifndef	MSDOS
#define	P_WAIT		1
#define	P_NOWAIT	2
#define	P_OVERLAY	3
#endif

/*
 *	Macro priority selection.
 */

#define	COMMAND_MACRO	4		/* Command line (highest priority) */
#define	HIGH_MACRO	3		/* Either user file or environment */
#define	LOW_MACRO	2		/* Either user file or environment */
#define	DEFAULT_MACRO	1		/* Default makefile (low priority) */

/*
 *	Type definitions.
 */

typedef	unsigned short t_mask;

typedef	struct	targnode
	{
	t_mask	tmask;			/* Mask to avoid string compares */
	struct	targnode *tnext;	/* Next target in global target list */
	struct	filenode *tfile;	/* File node for this target */
	struct	filenode **tpreq;	/* Pre-req list for this target */
	struct	shellnode **tshell;	/* Command list for this target */
	} targnode,*targptr;

typedef	struct	filenode
	{
	t_mask	fmask;			/* Mask to avoid string compares */
	char	*fname;			/* Name of this file (targ, preq...) */
	long	ftime;			/* Last MODIFY time for this file */
	struct	filenode *fnext;	/* Next file node in global file list */
	} filenode,*fileptr;

typedef	struct	shellnode
	{
	char	*scmd;			/* Text of command */
	unsigned s_silent:1;		/* Don't echo before executing */
	unsigned s_ignore:1;		/* Ignore exit status */
	unsigned s_shell:1;		/* Force spawning of command.com */
	unsigned s_force:1;		/* Force execution */
	struct	shellnode *snext;	/* Next shell node for a target */
	struct	shellnode *slink;	/* Next shell node in global list */
	} shellnode,*shellptr;

typedef	struct	symnode
	{
	t_mask	smask;			/* Mask to avoid string compares */
	char	*sname;			/* Name of a symbol */
	char	*svalue;		/* Value of a symbol */
	int	sprior;			/* Macro value priority */
	struct	symnode *snext;		/* Next symbol node in global list */
	} symnode,*symptr;

struct	msi				/* Make program state information */
	{
	int	depend;			/* -d option */
	int	dispcount;		/* Used for -D option */
	int	display;		/* -D option */
	fileptr	file_list;		/* List of file nodes */
	targptr	first_targ;		/* 1st target, if nothing explicit */
	char	**flist;		/* Command line make files */
	int	fmax;			/* Max size of flist */
	int	ignore;			/* -i option */
	int	inlen;			/* Input char counter in "input" */
	char	*input;			/* Current input line pointer */
	int	mac_env;		/* Environment macro priority */
	int	mac_usr;		/* User makefile macro priority */
	char	**mlist;		/* Command line macros */
	int	mmax;			/* Max size of mlist */
	int	noexec;			/* -n option */
	long	now;			/* Time at startup */
	char	*olds;			/* Token internal pointer */
	int	print;			/* -p option */
	int	readdef;		/* -r option */
	shellptr shell_list;		/* List of shell nodes */
	int	silent;			/* -s option */
	targptr	suffix_targ;		/* .SUFFIXES target pointer */
	symptr	symbol_list;		/* List of symbol nodes */
	targptr	target_list;		/* List of target nodes */
	char	**tlist;		/* Command line targets */
	int	tmax;			/* Max size of tlist */
	int	touch;			/* -t option */
	jmp_buf	up_one;			/* 'longjmp' target */
	};

/*
 *	External data definitions.
 */

extern	struct	msi	ms;		/* Make program internal state */
extern	int		nesting;	/* Nesting level for recursive MAKEs */
extern	char		*shell_cmds[];	/* List of commands requiring shell */

/*
 *	External function definitions.
 */

extern	fileptr		add_file();
extern	void		add_metas();
extern	shellptr	add_shell();
extern	void		add_symbol();
extern	targptr		add_target();
extern	char		**append_node();
extern	char		*breakout();
extern	void		build();
extern	long		curr_time();
extern	int		default_rule();
extern	void		display_prereq();
extern	void		dump_data();
extern	long		file_time();
extern	FILE		*fopenp();
extern	char		*getenv();
extern	symptr		get_symbol();
extern	char		**grow_list();
extern	fileptr		hash_file();
extern	t_mask		hash_name();
extern	targptr		hash_target();
extern	void		link_targs();
extern	void		main();
extern	void		make_args();
extern	int		new_make();
extern	void		nexit();
extern	int		spawnvp();
extern	void		terror();
extern	char		*tgets();
extern	char		*token();
extern	char		**tokenize();
extern	void		touch_file();
extern	char		*trealloc();
extern	char		*tstrncpy();
extern	void		usage();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

extern	int		add_macro();
extern	int		make();
extern	int		parse();
extern	char		*talloc();

/*lint -fvr */
