/************************************************************************/
/*									*/
/*	File:	make2.c							*/
/*									*/
/*	Makefile parsing code for MAKE.					*/
/*									*/
/*	Written by Greg Yachuk.  Placed in public domain 10/1/88.	*/
/*	Modeled after a MAKE written by Neil Russell which in turn was	*/
/*	modeled after UN*X MAKE...					*/
/*									*/
/*	Modified/maintained by Kirk Bailey of Logical Systems.		*/
/*									*/
/*	NOTE: This program is EXPLICITLY placed in the public domain.	*/
/*		***Do what you wish with it, and have fun!***		*/
/*									*/
/************************************************************************/

#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>
#include	<setjmp.h>

#include	"make.h"

/*
 *	add_file -	Return a found or created file node.
 */
fileptr
add_file(name)
	char			*name;
	{
	REGISTER fileptr	filep;
	t_mask			mask;

	/* See if file node already exists */
	filep = hash_file(name,&mask);
	if(filep != NULL)
		return (filep);

	filep = tnew(filenode);		/* Allocate new file node */
	filep->fmask = mask;		/* Save mask for later */
	filep->fname = tstrcpy(name);	/* Allocate string and copy name */
	filep->ftime = MAXNEGTIME;	/* Init MODIFY time to long time ago */

	filep->fnext = ms.file_list;	/* Link to head of file list */
	ms.file_list = filep;

	return (filep);
	}

/*
 *	add_macro -
 *
 *	Macros must have the format:
 *
 *		WORD = more stuff
 *	     	WORD= more stuff
 *		WORD =more stuff
 *		WORD=more stuff
 *	or:	WORD += more stuff
 *		WORD +=more stuff
 *
 *	It is assumed that there is no leading whitespace in `input'.
 */
int
add_macro(priority,input)
	int			priority;
	char			*input;
	{
	REGISTER char		*eqsign;
	symptr			symp;
	REGISTER char		*value;

	/* Gotta have an '=' to be a macro */
	eqsign = strchr(input, '=');
	if(eqsign == NULL)
		return (FALSE);

	/* Make sure we catch imbedded '='s (e.g. MACRO=STUFF) */
	for(value = input; ! isspace(*value); ++value)
		;
	if(value > eqsign)
		value = eqsign;

	/* Terminate the macro name */
	*value++ = '\0';

	/* Find start of value */
	for(value = eqsign + 1; isspace(*value); ++value)
		;

	/* Look for concat character */
	--eqsign;

	if((eqsign < input) || ((eqsign == input) && (*eqsign == '+')))
		terror("Badly formed macro","");

	if(*eqsign == '+')
		{
		/* Append to the current macro definition */
		symp = get_symbol(input);
		if((symp->svalue != NULL) && (priority >= symp->sprior))
			{
			eqsign = tstrcat(symp->svalue," ");
			value = tstrcat(eqsign,value);
			tfree(eqsign);
			tfree(symp->svalue);
			symp->svalue = value;
			symp->sprior = priority;
			return (TRUE);
			}
		}

	add_symbol(priority,input,value);
	return (TRUE);
	}

/*
 *	add_shell -	Create a new shell node.
 */
shellptr
add_shell(input)
	char			*input;
	{
	REGISTER shellptr	snode;

	snode = tnew(shellnode);	/* Allocate a new shell node */
	snode->s_shell = snode->s_ignore = snode->s_silent = snode->s_force =0;

	for(; isspace(*input); ++input)
		;			/* Skip over leading spaces */
	for(; ; ++input)
		{
		if(*input == '+')
			snode->s_shell = 1;	/* Must use command.com */
		else if(*input == '-')
			snode->s_ignore = 1;	/* Ignore return value */
		else if(*input == '@')
			snode->s_silent = 1;	/* Don't echo command */
		else if(*input == '^')
			snode->s_force = 1;	/* Force execution */
		else
			break;
		}

	snode->scmd = tstrcpy(input);	/* Allocate string and copy command */
	snode->snext = NULL;		/* No list of shells for a targ, yet */

	snode->slink = ms.shell_list;
	ms.shell_list = snode;

	return (snode);
	}

/*
 *	add_symbol -	Add a <name,value> pair to the symbol table.
 *			Override existing symbol value if priority is
 *			greater than or equal to existing priority.
 */
void
add_symbol(priority,name,value)
	int			priority;
	char			*name;
	char			*value;
	{
	REGISTER symptr		symp;

	symp = get_symbol(name);
	if(symp->svalue != NULL)
		{
		if(priority < symp->sprior)
			return;
		else
			tfree(symp->svalue);
		}
	symp->svalue = tstrcpy(value);
	symp->sprior = priority;
	}

/*
 *	add_target -	Return extant target node, or create new one.
 */
targptr
add_target(name)
	char			*name;
	{
	fileptr			filep;
	t_mask			mask;
	REGISTER targptr	targp;

	/* Each target must have a file node */
	filep = add_file(name);

	/* See if target already exists */
	targp = hash_target(name, &mask);
	if(targp != NULL)
		return (targp);

	/* Oh well, gotta create one */
	targp = tnew(targnode);		/* Allocate a target node */
	targp->tmask = mask;		/* Save mask for later */
	targp->tfile = filep;		/* Save pointer to file node */
	targp->tpreq = NULL;		/* No pre-req's yet */
	targp->tshell = NULL;		/* No shell lines yet */

	targp->tnext = ms.target_list;	/* Link to front of target list */
	ms.target_list = targp;

	return (targp);
	}

/*
 *	append_node -	Add a node to the end of an array of nodes
 */
char	**
append_node(node,adds,size)
	char			**node;
	char			**adds;
	int			size;
	{
	REGISTER int		addlen;
	REGISTER int		len;

	for(addlen = 0; adds[addlen] != NULL; ++addlen)
		;
	if(addlen++ == 0)
		return (node);

	len = 0;

	if(node != NULL)
		{
		for(; node[len] != NULL; ++len)
			;
		node = (char **)
			trealloc((char *) node, (len + addlen) * size);
		}
	else
		node = (char **) talloc(addlen * size);

	memcpy((char *) (node + len), (char *) adds,
		(unsigned) (addlen * size));
	return (node);
	}

/*
 *	breakout -	Replace macro names with values (apply recursively).
 *
 *	Note:		allocates (and returns) a string which must be freed.
 */
char	*
breakout(input)
	REGISTER char		*input;
	{
	char			*dend;
	char			*dest;
	int			dlen;
	REGISTER char		*dp;
	char			endch;
	int			slen;
	char			*sp;
	int			state;
	char			symname[MAXSYMSIZE + 1];
	symptr			symp;
	int			tlen;

	/* Allocate a string twice as long as input string */
	dlen = strlen(input) * 2;
	dest = dp = talloc(dlen);
	dend = dest + dlen;

/*
 *	State machine with 4 states:
 *
 *		0)	normal text	-- just copy
 *		1)	starting macro	-- define end char (e.g. ')', '}')
 *		2)	macro name	-- copy to a buffer
 *		3)	end of macro	-- look up value, and copy
 */
	state = 0;

	while((*input != '\0') || (state == 3))
		{
		/* If we don't have enough room, double size of string */
		if(dp == dend)
			{
			dlen *= 2;
			tlen = dp - dest;
			dest = trealloc(dest,dlen);
			dp = dest + tlen;
			dend = dest + dlen;
			}

		switch((int) state)
			{
			case 0:
				if(*input == '$')
					state = 1;	/* Found a macro */
				else
					*dp++ = *input++;
				break;

			case 1:
				state = 2;	/* Only in state for 1 char */
				sp = symname;
				switch((int) *++input)
					{
					case '(':
						endch = ')';
						break;
					case '{':
						endch = '}';
						break;
					default:
				/* Single char; go to state 3 immediately */
						*sp++ = *input;
						state = 3;
						break;
					}
				++input;	/* Skip '[' (or character) */
				break;

			case 2:
				if(*input == endch)
					state = 3;
				else
					*sp++ = *input;
				++input;	/* Skip end char */
				break;

			case 3:
				*sp = '\0';
				symp = get_symbol(symname);
				sp = symp->svalue;
				slen = -1;
				while((sp != NULL) && *sp)
					{
/*
 *	If value has a macro in it, we must process recursively.
 */
					if(*sp == '$')
						{
						sp = breakout(symp->svalue);
					/* Now guaranteed not to have a '$' */
						slen = strlen(sp);
						break;
						}
					++sp;
					}

				if(slen == -1)
					{
					/* Value did NOT have a macro */
					slen = (sp - symp->svalue);
					sp = symp->svalue;
					}

				/* If we have not enough room, expand */
				if(slen >= (dend - dp))
					{
				/* Use slen to make sure that we can fit */
					dlen = dlen * 2 + slen;
					tlen = dp - dest;
					dest = trealloc(dest,dlen);
					dp = dest + tlen;
					dend = dest + dlen;
					}

				/* If length is zero, don't bother to copy */
				if(slen)
					{
					strcpy(dp,sp);
					dp += slen;
					}

				if(sp != symp->svalue)
				tfree(sp);	/* Must've called `breakout' */

				state = 0;	/* And we are back to text */
				break;
			}
		}

	if(state != 0)
		terror("Improper macro: ",dest);

	*dp = '\0';			/* Terminate the string */
	return (dest);			/* And return it */
	}

/*
 *	dump_data -	Print macro definitions and target descriptions.
 */
void
dump_data()
	{
	REGISTER fileptr	*preqp;
	REGISTER shellptr	*shellp;
	REGISTER symptr		symp;
	REGISTER targptr	targp;

	printf("\nMacro definitions:\n\n");
	for(symp = ms.symbol_list; symp != NULL; symp = symp->snext)
		{
		printf("%-20s= %s\n", symp->sname,
			(symp->svalue != NULL) ? symp->svalue : "");
		}

	printf("\nEnd of macro definitions.  Target rules:\n\n");
	for(targp = ms.target_list; targp != NULL; targp = targp->tnext)
		{
		printf("%-20s:",targp->tfile->fname);
		for(preqp = targp->tpreq;
			((preqp != NULL) && (*preqp != NULL)); preqp++)
			{
			printf(" %s",(*preqp)->fname);
			}
		printf("\n");
		for(shellp = targp->tshell;
			((shellp != NULL) && (*shellp != NULL)); shellp++)
			{
			printf("                    %s\n",(*shellp)->scmd);
			}
		}
	printf("\nEnd of target rules.\n\n");
	}

/*
 *	get_symbol -	Find a symbol in the symbol table.  If non-extant,
 *			create <name,NULL>.
 *
 *	Return created or found symbol node.
 */
symptr
get_symbol(name)
	char			*name;
	{
	REGISTER t_mask		mask;
	REGISTER symptr		symp;

	/* Use `mask' to screen out most string comparisons */
	mask = hash_name(name);

	/* Linear search through symbol list */
	for(symp = ms.symbol_list; symp != NULL; symp = symp->snext)
		{
		if(mask != symp->smask)
			continue;

		if(equal(name,symp->sname))
			return (symp);
		}

	symp = tnew(symnode);		/* Allocate symbol node */
	symp->smask = mask;		/* Record mask for later */
	symp->sname = tstrcpy(name);	/* Allocate string and copy name */
	symp->svalue = NULL;		/* NULL value */
	symp->sprior = 0;		/* NULL value has LOWEST priority */

	symp->snext = ms.symbol_list;	/* Link to head of symbol list */
	ms.symbol_list = symp;

	return (symp);
	}

/*
 *	hash_file -	Look up file (by name) in file list.
 *
 *	Return file node or NULL, if requested, also return the mask.
 */
fileptr
hash_file(name,maskp)
	char			*name;
	t_mask			*maskp;
	{
	REGISTER fileptr	filep;
	REGISTER t_mask		mask;

	/* Use `mask' to screen out most string comparisons */
	mask = hash_name(name);

	/* See if we gotta return it */
	if(maskp != NULL)
		*maskp = mask;

	/* Linear search through file list */
	for(filep = ms.file_list; filep != NULL; filep = filep->fnext)
		{
		if(filep->fmask != mask)
			continue;

		if(equal(filep->fname,name))
			return (filep);
		}

	/* Nope, no file here */
	return (NULL);
	}

/*
 *	hash_name -	Transform a character string into an integer.
 */
t_mask
hash_name(name)
	char			*name;
	{
	REGISTER char		c;
	REGISTER t_mask		mask;
/*
 *	Take the sum of each pair of characters, and set a bit
 *	for this sum modulo the mask size (word size) in bits.
 *
 *	e.g.	string:	c0, c1, c2, c3
 *	set bit for (c0+c1) % mask_size
 *	set bit for (c1+c2) % mask_size
 *	set bit for (c2+c3) % mask_size
 *	set bit for (c3)    % mask_size	(end condition)
 *
 *	This gives a fair distribution, with not TOO much work...
 */
	mask = 0;
	while(*name)
		{
		c = *name++;
		mask |= (1L << ((c + *name) % (sizeof(mask) * 8)));
		}

	return (mask);
	}

/*
 *	hash_target -	Look up target (by name) in target list.
 *
 *	Return target node or NULL.  If requested, also return the mask.
 */
targptr
hash_target(name,maskp)
	char			*name;
	t_mask			*maskp;
	{
	REGISTER t_mask		mask;
	REGISTER targptr	targp;

	/* Use `mask' to screen out most string comparisons */
	mask = hash_name(name);

	/* See if we gotta return it */
	if(maskp != NULL)
		*maskp = mask;

	/* Linear search through target list */
	for(targp = ms.target_list; targp != NULL; targp = targp->tnext)
		{
		if(mask != targp->tmask)
			continue;

		/* Target name is ONLY stored in the file node */
		if(equal(name,targp->tfile->fname))
			return(targp);
		}

	/* Nope, no target here */
	return (NULL);
	}

/*
 *	link_targs -	Force a list of targs to point to same preq's and
 *			shell's.
 */
void
link_targs(targs,preqs,shells)
	REGISTER targptr	*targs;
	fileptr			*preqs;
	shellptr		*shells;
	{
	REGISTER int		plen;

	for(plen = 0; preqs[plen] != NULL; ++plen)
		;

	while(*targs != NULL)
		{
		/* Process some special targets */
		if((*targs)->tfile->fname[0] == '.')
			{
			if(equal((*targs)->tfile->fname,".SILENT"))
				ms.silent = 1;
			else if(equal((*targs)->tfile->fname,".IGNORE"))
				ms.ignore = 1;
			else if(equal((*targs)->tfile->fname,".SUFFIXES"))
			     /* Set `suffix_targ' to speed up `default_rule' */
				ms.suffix_targ = *targs;

			/* Special rule has preq's reset */
			/* Normally, preq's are merely appended */
			if(plen == 0)
				{
				tfree((*targs)->tpreq);
				(*targs)->tpreq = NULL;
				}
			}

/*
 *	Each target in the list points to the preq's and shell's for this
 *	rule PLUS the preq's from previous rules (NOT THE OLD SHELL'S HOWEVER).
 */
		(*targs)->tpreq = append_preq((*targs)->tpreq,preqs);
		tfree((*targs)->tshell);
		(*targs)->tshell = NULL;
		(*targs)->tshell = append_shell((*targs)->tshell,shells);
		++targs;
		}
	}

/*
 *	parse -		Read (text) makefile, and parse (close file before
 *			returning).
 *
 *	Lines have the following format:
 *
 *		# with or without preceeding spaces/tabs	(comment line)
 *		<TAB> commands					(shell line)
 *		name = stuff					(macro)
 *		name += stuff					(macro)
 *		targ [targ...] : [pre-req...]			(target line)
 */
int
parse(priority,fd)
	int			priority;
	FILE			*fd;
	{
	char			*colonp;
	char			*input;
	char			*ip;
	int			ltmax;
	int			npreqs;
	int			nshell;
	int			ntargs;
	int			pmax;
	fileptr			*preqs;
	shellptr		*shells;
	int			smax;
	targptr			*targs;

	if(fd == NULL)
		return (FALSE);

	/* Start off with a short list of targets */
	targs = (targptr *) grow_list(NULL,&ltmax);
	preqs = (fileptr *) grow_list(NULL,&pmax);
	shells = (shellptr *) grow_list(NULL,&smax);

	ntargs = npreqs = nshell = 0;

	/* Maximize buffering */
	if(setvbuf(fd,NULL,_IOFBF,READBUFSIZE))
		terror("Unable to create input buffer","");

	while((input = tgets(fd)) != NULL)
		{
		/* Punt on comments and blank lines */
		for(ip = input; isspace(*ip); ++ip)
			;
		if((*ip == '#') || (*ip == '\0'))
			continue;
		/* Process include files */
		if(! strncmp(ip,"include ",8))
			{
			/* Skip spaces AFTER "include " */
			for(ip += 8; isspace(*ip); ++ip)
				;
			if(! parse(priority,fopen(ip,"r")))
				terror("Unable to open file: ",ip);
			continue;		/* Get next input line */
			}

		/* Display the makefile line ? */
		if(ms.display)
			puts(input);

		/* See if we have a shell command */
		if(isspace(*input))
			{
			if(ntargs == 0)
				terror("Rules must be after target","");

			if(nshell == smax)
				shells = (shellptr *)
						grow_list((char **) shells,
							&smax);
			shells[nshell++] = add_shell(ip);
			continue;
			}

		/* Not a shell line, so must be a target or a macro */
		if(ntargs != 0)
			{
			/* Link previous preq's and shell's */
			targs[ntargs] = NULL;
			preqs[npreqs] = NULL;
			shells[nshell] = NULL;
			link_targs(targs,preqs,shells);
			ntargs = npreqs = nshell = 0;
			}

		if((colonp = strchr(ip,'#')) != NULL)
			*colonp = '\0';		/* Get rid of comment */

		/* Don't break out symbols until macro is invoked */
		if(add_macro(priority,ip))
			continue;

		/* Okay, we have a target line; break out macro symbols */
		input = breakout(ip);

		/* Just look for tokens with standard isspace() separators */
		ip = token(input,NULL);
		while(ip != NULL)
			{
			colonp = strchr(ip,':');
#ifdef	MSDOS
			/* Need to allow c:/bin/make.exe as a target */
			if((colonp != NULL) && ((colonp - ip) == 1))
				colonp = strchr(colonp + 1,':');
#endif	/* MSDOS */
			if(colonp != NULL)
				{
				/* Got a separator */
				*colonp = '\0';

				/* If at front of token, target is done */
				if(colonp == ip)
					break;
				}

			if(ntargs == ltmax)
				targs = (targptr *) grow_list((char **) targs,
						&ltmax);
			targs[ntargs] = add_target(ip);

			/* Make sure we don't save .INIT as our 1st target */
			if((ms.first_targ == NULL) && (*ip != '.'))
				ms.first_targ = targs[ntargs];
			++ntargs;

			if(colonp != NULL)
				break;
			ip = token(NULL,NULL);
			}
/*
 *	Taking care of four possible cases:
 *
 *	 	1)	object : source
 *		2)	object: source
 *		3)	object :source
 *		4)	object:source
 */
		if((colonp != NULL) && *++colonp)
			ip = colonp;
		else
			ip = token(NULL,NULL);

		/* Link the pre-req's */
		while(ip != NULL)
			{
			if(npreqs == pmax)
				preqs = (fileptr *) grow_list((char **) preqs,
						&pmax);
			preqs[npreqs++] = add_file(ip);
			ip = token(NULL,NULL);
			}

		/* Gotta free the line allocated by breakout() */
		tfree(input);
		}

	/* Link up any dangling dependants */
	if(ntargs != 0)
		{
		targs[ntargs] = NULL;
		preqs[npreqs] = NULL;
		shells[nshell] = NULL;
		link_targs(targs,preqs,shells);
		}

	/* Clean up our mallocs */
	tfree(targs);
	tfree(preqs);
	tfree(shells);

	if(fclose(fd))
		terror("Unable to close makefile","");

	return (TRUE);
	}
