/************************************************************************/
/*									*/
/*	File:	make3.c							*/
/*									*/
/*	Token processing routines for MAKE.				*/
/*									*/
/*	Written by Greg Yachuk.  Placed in public domain 10/1/88.	*/
/*	Modeled after a MAKE written by Neil Russell which in turn was	*/
/*	modeled after UN*X MAKE...					*/
/*									*/
/*	Modified/maintained by Kirk Bailey of Logical Systems.		*/
/*									*/
/*	NOTE: This program is EXPLICITLY placed in the public domain.	*/
/*		***Do what you wish with it, and have fun!***		*/
/*									*/
/************************************************************************/

#include	<stdio.h>
#include	<ctype.h>
#include	<malloc.h>
#include	<string.h>
#include	<setjmp.h>

#include	"make.h"

/*
 *	nexit -		Exit with supplied return code if at top level.  If
 *			in a nested MAKE use 'setjmp' to simulate a program
 *			exit to the next higher level of MAKE.
 */
void
nexit(n)
	int			n;
	{
	if(nesting == 0)
		exit(n);
	else
		{
		if(n == 0)
			n = -1;		/* OK flag to "setjmp" */
		longjmp(ms.up_one,n);	/* Simulate MAKE "exit" */
		}
	}

/*
 *	talloc -	Allocate space from heap and bitch if none available.
 */
char	*
talloc(n)
	int	n;
	{
	REGISTER char		*s;

	s = malloc((unsigned) n);
	if(s == NULL)
		terror("Insufficient free memory","");
	return (s);
	}

/*
 *	terror -	Print error message to user.
 */
void
terror(s1,s2)
	char			*s1;
	char			*s2;
	{
	fprintf(STDERR,"FATAL: %s%s\n",s1,s2);
	nexit(1);
	}

/*
 *	tgets -		Read input, swallowing escaped newlines as necessary.
 */
char	*
tgets(fd)
	FILE			*fd;
	{
	REGISTER char		*ep;
	int			len;

	if(ms.inlen == 0)
		ms.input = talloc(ms.inlen = MAXLNSIZE);

	ms.input[ms.inlen-2] = '\n';
	ep = ms.input - 1;
	while((fgets(ms.input,ms.inlen,fd)) != NULL)
		{
		for(;;)
			{
			while((ms.input[ms.inlen-2] != '\n') &&
				(ms.input[ms.inlen-2] != '\0'))
				{
				len = ms.inlen;
				ms.input = trealloc(ms.input,(ms.inlen <<= 1));
				ep = &ms.input[len - 2];
				ms.input[ms.inlen - 2] = '\n';
				fgets(ep + 1,len + 1,fd);
				}

			while(*++ep);
				*--ep = '\0';
			do
				{
				--ep;
				} while (isspace(*ep) && (ep >= ms.input));

			if((*ep == '\\') && (*--ep != '\\'))
				fgets(ep + 1,ms.inlen - (ep - ms.input) - 1,fd);
			else
				break;
			}

		return (ms.input);
		}

	ms.inlen = 0;
	tfree(ms.input);
	ms.input = NULL;

	return (NULL);
	}

/*
 *	token -		Take an input string and return a token each call.  The
 *			default token delimiter characters are `isspace()'.
 *			Text between quotes (" and ') is a single token.
 *
 *	Called as	s = token(string,seps);
 *	or		s = token(string,NULL);
 *
 *	Followed by	s = token(NULL,seps);
 *	or		s = token(NULL,NULL);
 *
 *	Returns NULL when no more tokens are available.
 */
char	*
token(s,sep)
	char			*s;
	char			*sep;
	{
	char			quote = 0;	/* Processing a quoted token */

	if(s != NULL)
		ms.olds = s;		/* We are starting all over again */

	if((ms.olds == NULL) || (! *ms.olds))
		return (NULL);		/* No tokens left */

	while(isspace(*ms.olds) || ((sep != NULL) && (strchr(sep,*ms.olds) != NULL)))
		++ms.olds;			/* Skip leading spaces and sep's */
	s = ms.olds;

	while(*ms.olds)
		{
		if(isspace(*ms.olds) || ((sep != NULL) &&
					(strchr(sep,*ms.olds) != NULL)))
			{
			*ms.olds++ = '\0';	/* Delimit the token */
			return (s);
			}
		else if((*ms.olds == '"') || (*ms.olds == '\''))
			{
			quote = *ms.olds++;
			while((*ms.olds != quote) && (*ms.olds != '\0'))
				++ms.olds;	/* Search for the proper quote char */
			if(*ms.olds != '\0')
				++ms.olds;	/* Didn't hit EOS, so skip quote */
			}
		else
			++ms.olds;		/* Otherwise, pass over char */
		}

	ms.olds = NULL;
	return (s);			/* Return last token */
	}

/*
 *	tokenize -	chop a string up into an array of (char *)'s.
 */
char	**
tokenize(input)
	char			*input;
	{
	REGISTER int		alen;
	REGISTER int		argc = 0;
	char			**argv;

	alen = 20;			/* Good initial guess */
	argv = (char **) talloc((alen + 1) * sizeof (char *));

	input = token(input,NULL);	/* Use default separators */
	while(input != NULL)
		{
		if(alen == argc)
			argv = (char **) trealloc((char *) argv,
					(alen <<= 1) * sizeof (char *));
/*
 *	Remove quotes from around tokens.  Note that quotes must be same sex.
 *	The trailing quote isn't required.
 */
		if((*input == '\'') || (*input == '"'))
			{
			*strrchr(input, *input) = '\0';
			input++;
			}
		argv[argc++] = input;
		input = token(NULL,NULL);
		}

	argv[argc] = NULL;		/* Mark end of array */

	return (argv);
	}

/*
 *	trealloc -	(Re)allocate space from heap and bitch if unavailable.
 */
char   *
trealloc(s,n)
	REGISTER char		*s;
	int			n;
	{
	s = realloc(s,(unsigned) n);
	if(s == NULL)
		talloc(n);		/* Force an error */
	return (s);
	}

/*
 *	tstrncpy -	Get the storage for a string from the heap and copy
 *			it into it.
 */
char   *
tstrncpy(s,n)
	REGISTER char		*s;
	int			n;
	{
	s = strncpy(talloc(n + 1),s,(unsigned) n);
	s[n] = '\0';
	return (s);
	}

