/************************************************************************/
/*									*/
/*	File:								*/
/*		PP3.C	General I/O for PP				*/
/*									*/
/*	Name:								*/
/*		PP - A preprocessor, principally for the C language.	*/
/*									*/
/*	Written by:							*/
/*			Gary Oliver					*/
/*			3420 NW Elmwood Dr.				*/
/*			PO Box 826					*/
/*			Corvallis, Oregon 97339				*/
/*			(503)758-5549					*/
/*	Maintained by:							*/
/*			Kirk Bailey					*/
/*			Logical Systems					*/
/*			P.O. Box 1702					*/
/*			Corvallis, OR 97339				*/
/*			(503)753-9051					*/
/*									*/
/*	This program is hereby placed in the public domain.  In		*/
/*	contrast to other claims of "public domain", this means no	*/
/*	copyright is claimed and you may do anything you like with PP,	*/
/*	including selling it!  As a gesture of courtesy, please retain	*/
/*	the authorship information in the source code and		*/
/*	documentation.							*/
/*									*/
/*	Functions contained in this module:				*/
/*									*/
/*		cur_user	Select current disk-user (CPM).		*/
/*		do_line		Issue #line output directive.		*/
/*		doinclude	Process #include directive.		*/
/*		doline		Process #line directive.		*/
/*		gchbuf		Get char from file buffer.		*/
/*		gchfile		Get new buffer from file and return ch.	*/
/*		gchpb		Get char from pushback buffers.		*/
/*		getchn		Get char from buffer, ignoring "\\n"	*/
/*		inc_open	Open file for inclusion.		*/
/*		init_path	Initialize Ipath information.		*/
/*		popfile		Pop to next lower file level.		*/
/*		readline	Read and macro-process a line of text.	*/
/*		scaneol		Scan to eol, leaving EOL as next token.	*/
/*		set_user	Select a disk-user (CPM).		*/
/*		trigraph	Handle ANSI trigraph translations.	*/
/*									*/
/************************************************************************/

#include "pp.h"

#if	HOST == ISQC
/************************************************************************/
/*									*/
/*	cur_user							*/
/*									*/
/*	Restore disk/user to that at program start.			*/
/*									*/
/************************************************************************/
cur_user()
    {
    bdos(BDOS_SELDISK, Orig_disk);
    bdos(BDOS_USER, Orig_user);
    }
#endif	/* HOST == ISQC */

/************************************************************************/
/*									*/
/*	do_line								*/
/*									*/
/*	Write #line directive to output file, synching to current	*/
/*	input line number and file name.				*/
/*									*/
/************************************************************************/

/*
 *	The following definition of NL_CHAR is used to minimize the file
 *	storage space required for the output file PP writes.  If the external
 *	file format stores "\n" as a single character (as on the UNIX systems),
 *	then NL_CHAR is set to 1.  Otherwise it is set to 2 to reflect the CR/
 *	LF used for external files.  NLCHAR is only used by this function to
 *	optimize the tradeoff between emitting extra "#line" directives (if
 *	enabled), or to output extra "\n" characters instead.
 */
#if	(HOST == ISUNIX) OR (HOST == ISBSD) OR (HOST == ISMPW) OR (HOST == ISXENIX)
#define	NL_CHAR	1
#else	/* ISUNIX || ISBSD || ISMPW || ISXENIX */
#define	NL_CHAR	2
#endif	/* ISUNIX || ISBSD || ISMPW || ISXENIX */

do_line(at_bol)
  char at_bol;			/* TRUE if already at beginning of line */
    {
    register int n;
    char filen[FILENAMESIZE + 1], buf[TOKENSIZE + 1];

    n = Tokenline - Outline;	/* Difference in line #s */

    sprintf(filen, " \"%s\"", Filestack[Tokenfile]->f_name);
    sprintf(buf, "%s#%s %d%s\n",
	    !at_bol ? "\n" : "",	/* Emit a \n if needed */
	    Lineopt == LINE_EXP ? "line" : "",
#if	HOST == ISQC
	    Tokenline-1,		/* QC bug */
#else	/* HOST != ISQC */
	    Tokenline,
#endif	/* HOST == ISQC */
	    Do_name ? filen : "");

#if	(TARGET == ISQC) OR (TARGET == ISTC) OR (TARGET == ISQCX)
    if((!Do_name && (n >= 0) && (n < (strlen(buf)/NL_CHAR + 1))) || Do_asm)
#else	/* ! ((TARGET == ISQC) OR (TARGET == ISTC) OR (TARGET == ISQCX)) */
    if(!Do_name && (n >= 0) && (n < (strlen(buf)/NL_CHAR + 1)))
#endif	/* (TARGET == ISQC) OR (TARGET == ISTC) OR (TARGET == ISQCX) */
	{
	while(n-- > 0)
	    putc('\n', Output);	/* Write newlines to synch */
	}
    else
	{
	fprintf(Output, buf);
	}

    Outline = Tokenline;		/* Make them the same */
    Do_name = FALSE;
    }

/************************************************************************/
/*									*/
/*	doinclude							*/
/*									*/
/*	Process #include file commands.					*/
/*									*/
/*	The file name is parsed and the file is opened.  A stack of	*/
/*	current files and file names is pushed, with the new one	*/
/*	being placed upon the top of the other, currently opened,	*/
/*	files.  Upon return, the next character read is from the	*/
/*	newly opened file.						*/
/*									*/
/*									*/
/************************************************************************/
doinclude()
    {
    extern int getchn();
    extern char *readline();

    char incfile[FILENAMESIZE + 1];	/* Include file name */
#if	HOST != ISQC
    char filename[FILENAMESIZE + 1];	/* unix file name build area */
    register char **ip;
#endif	/* HOST != ISQC */
    register int c, d;
    register char *incf;
    register int ok;
#if	HOST == ISQC
    register int user, disk;
    register char *p, **ip;
#endif	/* HOST == ISQC */
    char buf[TOKENSIZE];

#if	DEBUG
if (Debug) printf("doinclude: include level:%d\n", Filelevel);
#endif	/* DEBUG */
    if (Filelevel >= FILESTACKSIZE)
	{
	non_fatal("Include file stack overflow","");
	return;
	}
/*
 *	Read line and process any macros encountered.
 */
    pbcstr(readline(buf, TOKENSIZE, GT_ANGLE));

    while(istype((c = getchn()), C_W))
	;

    if (c == EOF)
	end_of_file();
    else
    if (c == '<')
	d = '>';
    else
    if (c == '"')
	d = '"';
    else
	{
	non_fatal("Bad include argument","");
	if(c == '\n')
	    pushback('\n');
	return;
	}

    for (incf = incfile; c != EOF && c != '\n';)
	{
	if ((c = getchn()) == d)
	    break;

	if (incf >= &incfile[FILENAMESIZE])
	    {
	    non_fatal("Include file name too long","");
	    return;
	    }
	else
	    *incf++ = c;
	}

    while((c != '\n') && (c != EOF)) c = getchn();	/* Slurp trailing \n */

    if (incf != incfile)
	*incf = '\0';
    else
	{
	non_fatal("Illegal file name","");
	return;
	}

    if(Lineopt) do_line(TRUE);	/* Catch up before include file switch */

#if	DEBUG
if (Debug) printf("doinclude: <%s>\n", incfile);
#endif	/* DEBUG */

    if(Verbose)
	printf("*** Include %s\n", incfile);

    ok = FALSE;

#if	HOST == ISQC		/* The following stuff works only under CP/M */

    /* Determine the type of include: if d == '<' then search the	*/
    /* path indicated by the -i parameter (or the default.)  If the	*/
    /* delimiter is '"' then include the current directory as well.	*/

    if(d == '"')
	ok = inc_open(incfile, -1, 0);	/* The current location */

    /* Try all path options in -i list (including DFLT_PATH stuff)	*/
    for(ip = &Ipath[0]; (*ip != NULL) && !ok; ip++)
	{
	p = *ip;

	while(istype(*p & 0xFF, C_W))
	    ++p;		/* Skip white space */

	disk = *p++;		/* Get drive letter */

	if(isupper(disk))
	    disk = tolower(disk);

	if(disk == '*')
	    user = -1;		/* Use the default */
	else
	if(disk >= 'a' && disk <= 'p')
	    {
	    disk -= 'a';		/* Drive */

	    if(!isdigit(*p))
		user = -1;		/* Default */
	    else
		{
		user = atoi(p);
		if((user < 0) || (user > 31))
		    {
		    non_fatal("Illegal user number","");
		    return;
		    }
		}
	    }
	else
	    {
	    non_fatal("Illegal disk drive specifier","");
	    return;
	    }

	ok = inc_open(incfile, user, disk);
	}

#else	/* HOST != ISQC */

    if(d == '"')
	{
	/* Look in current directory */
	strcpy(filename, Filestack[Filelevel]->f_name);	/* Original name */
	if(strrchr(filename, SLASHCHAR))
	    strcpy(strrchr(filename, SLASHCHAR)+1, incfile);
	else
	    strcpy(filename, incfile);

	ok = inc_open(filename);
	}

    /* Look through all paths for an existance of the file	*/

    for(ip = &Ipath[0]; *ip != NULL && !ok; ip++)
	{
	strcpy(filename, *ip);		/* Copy path name		*/
	strcat(filename, SLASHSTR);	/* Append / for directory	*/
	strcat(filename, incfile);	/* Append local file name	*/
	ok = inc_open(filename);	/* Attempt an open		*/
	}

#endif	/* (HOST != ISQC) */

#if	HOST == ISQC
    cur_user();			/* Restore current user/disk */
#endif	/* HOST == ISQC */

    if(!ok) non_fatal("Unable to open include file: ",incfile);
    pushback('\n');
    Lastnl = TRUE;	/* Let token scanner see first things on line */
    }

/************************************************************************/
/*									*/
/*	doline								*/
/*									*/
/*	Set current line number and file identifier to that specified.	*/
/*									*/
/************************************************************************/
doline()
    {
    extern int getchn();
    extern int getnstoken();
    extern char *readline();

    char *p;
    int c, l;
    char buf[TOKENSIZE];
/*
 *	Read line and process any macros encountered.
 */
    pbcstr(readline(buf, TOKENSIZE, GT_STR));

    while(istype((c = getchn()), C_W))
	;

    if(istype(c, C_D))
	{
	for(l = 0; istype(c, C_D); c = getchn())
	    l = l * 10 + c - '0';

	Line = l - 1;			/* Set line number */

	pushback(c);
	c = getnstoken(GT_STR);

	if((c != '\n') && (c != EOF))
	    {
	    if (c == '"')
		{
		p = strrchr(Token, '"');	/* Find ending " */

		if(p-Token > FILENAMESIZE)	/* Allow for first " */
		    p = &Token[FILENAMESIZE+1];

		*p = '\0';		/* Terminate it */
		strcpy(Filestack[Filelevel]->f_name, Token+1);
		Do_name = TRUE;	/* Need filename on "output" #line */
		}
	    else
		{
		pushback(c);
		c = '\0';
		}
	    }
	}
    else
	{
	pushback(c);
	c = '\0';
	}
    if(c == '\0')
	non_fatal("\"#line\" argument error","");

    while((c != '\n') && (c != EOF))
	c = getnstoken(GT_STR);		/* Slurp trailing \n */
    pushback('\n');
    }

/************************************************************************/
/*									*/
/*	gchbuf								*/
/*									*/
/*	Get char from input buffer.  If out of chars in buffer, read	*/
/*	another buffer.							*/
/*									*/
/************************************************************************/
int
gchbuf()
    {
    register int c;

    for(;;)
	{
	if(Lasteol)
	    {
	    Lasteol = FALSE;
	    Line++;
	    }

	if(!istype(c = (Bufc-- ? *Bufp++ : gchfile()), C_C))
	    break;	/* If no need to examine closely */

#ifdef	IGNORE_CR
	if(c == '\r')
	    continue;
#endif	/* IGNORE_CR */

	if(c == '\n')
	    Lasteol = TRUE;	/* Inc line number next time */
#ifdef	PP_SYSIO
#if	HOST == ISQC
	else
	if(c == ENDFILE)
	    {
	    Bufc = 0;
	    continue;			/* Try to get next file */
	    }
#endif	/* HOST == ISQC */
#endif	/* PP_SYSIO */
	break;
	}

    return(c);			/* And exit */
    }

/************************************************************************/
/*									*/
/*	gchfile								*/
/*									*/
/*	Get next buffer of chars from input file.  If end of file	*/
/*	found, pop file level and read from lower file.  If out of	*/
/*	files, return EOF.						*/
/*									*/
/************************************************************************/
int
gchfile()
	{
	extern int gchbuf();
	extern int popfile();
	extern int trigraph();

	register struct file *f;

	if(Filelevel < 0)
	    {
	    Bufc = 0;
	    return (EOF);
	    }
	else if((Filestack[Filelevel]->f_eof) && popfile())
	    return (A_trigraph ? trigraph() : gchbuf());
					/* (Possible recursion)		*/
	if(Filelevel < 0)
	    {
	    Bufc = 0;
	    return (EOF);
	    }

	f = Filestack[Filelevel];

#if	HOST == ISQC
	set_user();
#endif	/* HOST == ISQC */

	Bufp = f->f_buf;		/* Set buffer address */

#ifdef	PP_SYSIO
	if((Bufc = read(f->f_fd, Bufp, BUFFERSIZE)) == 0)
#else	/* !PP_SYSIO */
	if((Bufc = fread(Bufp, 1, BUFFERSIZE, f->f_file)) == 0)
#endif	/* PP_SYSIO */
	    {
	    f->f_eof = TRUE;
	    return ('\n');		/* Fake '\n' before EOF */
	    }

#if	HOST == ISQC
	cur_user();
#endif	/* HOST == ISQC */

	Bufc--;
	return(*Bufp++ & 0xFF);		/* Return the char */
	}

/************************************************************************/
/*									*/
/*	gchpb								*/
/*									*/
/*	Get character from push-back buffer.  If out of chars, return	*/
/*	next char routine to get chars from file.			*/
/*									*/
/************************************************************************/
int
gchpb()
    {
    extern int gchbuf();
    extern int trigraph();

    register int c;

    for(;;)
	{
	if(Pbbufp->pb_type == PB_CHAR)
	    c = (Pbbufp--)->pb_val.pb_char;	/* Pop the char */
	else
	if(Pbbufp->pb_type == PB_STRING)
	    {
	    /* Get next char from string */
	    if((c = *(Pbbufp->pb_val.pb_str++) & 0xFF) == '\0')
		{
/*
 *	End of the string.  Pop the stack to get the next pushback buffer
 *	which contains the starting address for the string.
 */
		Pbbufp--;
		free(Pbbufp->pb_val.pb_str);	/* Free up the memory */
		Pbbufp--;	/* Get next real pushback buffer */
		continue;	/* Try again to get some input */
		}
	    }
	else
	    {
	    /* Assume PB_TOS and pop to file input */
	    Nextch = A_trigraph ? trigraph : gchbuf;	/* Next char source */
	    return (A_trigraph ? trigraph() : gchbuf());
	    }
	return (c);
	}
/*lint -e533 lint doesn't understand about functions which never return */
    }
/*lint +e533 lint doesn't understand about functions which never return */

/************************************************************************/
/*									*/
/*	getchn								*/
/*									*/
/*	Get next char from nextch() routine, ignoring \ followed	*/
/*	by a newline character.						*/
/*									*/
/************************************************************************/
int
getchn()
    {
    int c, c2;

    for(;;)
	{
	if((c = nextch()) == '\\')
	    {
	    c2 = nextch();
	    if(c2 == '\n')
		continue;		/* Ignore the char and get another */

	    pushback(c2);		/* We'll get to this char later */
	    }
	return (c);
	}
/*lint -e533 lint doesn't understand about functions which never return */
    }
/*lint +e533 lint doesn't understand about functions which never return */

/************************************************************************/
/*									*/
/*	inc_open							*/
/*									*/
/*	Open an include file for the selected path.			*/
/*									*/
/************************************************************************/
#if	HOST == ISQC
int	inc_open(incfile, u, d)
  register char *incfile;
  int u, d;
#else	/* HOST != ISQC */
int	inc_open(incfile)
  char *incfile;
#endif	/* HOST == ISQC */
    {
#ifdef	PP_SYSIO
    extern int open();
#else	/* !PP_SYSIO */
    extern FILE *fopen();
#endif	/* PP_SYSIO */
    extern char *malloc();

    register int v;
    register struct file *f, *fold;

#if	HOST == ISQC

#if	DEBUG
if(Debug) printf("inc_open: %s on %c%d\n", incfile, d+'A',u);
#endif	/* DEBUG */

    if(u >= 0)
	{
	bdos(BDOS_USER, u);
	bdos(BDOS_SELDISK, d);
	}

#else	/* HOST != ISQC */

#if	DEBUG
if(Debug) printf("inc_open: %s\n", incfile);
#endif	/* DEBUG */

#endif	/* HOST == ISQC */

    f =
    /*NOSTRICT*/
    Filestack[Filelevel+1] =
	(struct file *) malloc(sizeof(struct file));

    if(f == NULL)
	out_of_memory();

#ifdef	PP_SYSIO
    if (v = ((f->f_fd = open(incfile, 0)) != -1))
#else	/* !PP_SYSIO */
    if (v = (int)((f->f_file = fopen(incfile, "r")) != NULL))
#endif	/* PP_SYSIO */
	{
	if(Filelevel >= 0)		/* Don't do if first time thru */
	    {
#if	DEBUG
if(Debug) printf("inc_open pushing: Bufc=%d, Bufp=%04x, Lasteol=%d, Line=%d\n",
		Bufc, Bufp, Lasteol, Line);
#endif	/* DEBUG */

	    fold = Filestack[Filelevel];
	    fold->f_bufp = Bufp;	/* Save current buffer pointer */
	    fold->f_bufc = Bufc;	/* Save current buffer count */
	    fold->f_lasteol = Lasteol;	/* Save last char */
	    fold->f_line = Line;	/* Save current line # */
	    }

	Filelevel++;
	strcpy(f->f_name, incfile);
	Line		= 1;		/* Initial line number		*/
	Bufc		= 0;		/* No chars in buffer		*/
	f->f_eof	= 		/* Not at eof yet		*/
	Lasteol		= FALSE;	/* Last char was not EOL	*/
#if	HOST == ISQC
	f->f_disk = d;
	f->f_user = u;
#endif	/* HOST == ISQC */
	}
    else
	/*NOSTRICT*/
	free((char *)f);		/* Return the memory used */

#if	HOST == ISQC
    if(u >= 0)
	cur_user();			/* Restore current user/disk */
#endif	/* HOST == ISQC */

    if(v)
	{
	Do_name = TRUE;
	return (TRUE);
	}
    else
	return (FALSE);
    }

/************************************************************************/
/*									*/
/*	init_path							*/
/*									*/
/*	Initialize path searching information.  Try to open a file	*/
/*	defined as PATHFILE on A0:.  If available, read in the path	*/
/*	default information, otherwise use CPM_PATH as the default.	*/
/*									*/
/*	If initializing for non-CPM, initialize Ipath to whatever is	*/
/*	specified by the PPINC environment variable or DFLT_PATH if no	*/
/*	matching environment variable is found.				*/
/*									*/
/************************************************************************/
init_path()
  {
#if	HOST == ISQC
    extern FILE *fopen();
    extern char *malloc();

    register FILE *pf;
    register int inum;
    char pb[TOKENSIZE];

    bdos(BDOS_USER, 0); bdos(BDOS_SELDISK, 0);	/* Select A0 */

    if(pf = fopen(PATHFILE, "r"))
	{
	/* Found the file -- read lines and use as paths */
	for(inum = Ipcnt; inum < NIPATHS; inum++)
	    {
	    if(fgets(pb, TOKENSIZE, pf) != NULL)
		{
		if((/*NOSTRICT*/
		    Ipath[inum] = malloc((unsigned)(strlen(pb) + 1))) == NULL)
			out_of_memory();
		else
		    strcpy(Ipath[inum], pb);	/* Copy the default path */
		}
	    else
		{
		if(inum == Ipcnt)
		    {
		    /* Didn't find any -- give informative */
		    warning("Bad format on include path file: ",PATHFILE);
		    Ipath[inum++] = DFLT_PATH;	/* Use default path */
		    }
		break;
		}
	    }

	Ipcnt = inum;			/* Keep counter correct	*/

	if(fclose(pf) == EOF)
	    non_fatal("Unable to close include path file: ",PATHFILE);
	}
    else
	Ipath[Ipcnt++] = DFLT_PATH;	/* The default path list */

    cur_user();				/* Restore user/disk defaults */
#endif	/* HOST == ISQC */

#if	HOST != ISQC
    extern char *getenv();
    extern char *malloc();

    char *cptr1,*cptr2;
    char pb[TOKENSIZE];

    /* See if there is an environment variable for the default search path */

    if((cptr1 = getenv(ENV_PATH)) == NULL)
	cptr1 = strcpy(pb,DFLT_PATH);	/* Nope, use default path */
    else
	cptr1 = strcpy(pb,cptr1);

/*lint -e530 lint isn't smart enough to see that "cptr2" is initialized */
    for(; (*cptr1 != '\0') && (Ipcnt < NIPATHS); cptr1 = cptr2)
/*lint +e530 lint isn't smart enough to see that "cptr2" is initialized */
	{
	if((cptr2 = strchr(cptr1,PATHPUNC)) != NULL)
	    *cptr2++ = '\0';
	else
	    cptr2 = "\0";

	if((Ipath[Ipcnt] = malloc((unsigned) (strlen(cptr1) + 1))) == NULL)
	    out_of_memory();
	else
	    strcpy(Ipath[Ipcnt++],cptr1);
	 }

#endif	/* HOST != ISQC */
    }

/************************************************************************/
/*									*/
/*	popfile								*/
/*									*/
/*	Pop file stack to previous file and return FALSE if no more.	*/
/*									*/
/************************************************************************/
int
popfile()
    {
    register struct file *f;

#if	HOST == ISQC
    set_user();
#endif	/* HOST == ISQC */

#ifdef	PP_SYSIO
    if(close((f = Filestack[Filelevel])->f_fd) == -1)
#else	/* !PP_SYSIO */
    if(fclose((f = Filestack[Filelevel])->f_file) == EOF)
#endif	/* PP_SYSIO */
	non_fatal("Unable to close input/include file: ",f->f_name);

    /*NOSTRICT*/
    free((char *)f);		/* Free the entry */

#if	HOST == ISQC
    cur_user();
#endif	/* HOST == ISQC */

    if(Filelevel-- == 0)
	return (FALSE);		/* At bottom level, real EOF */

    f = Filestack[Filelevel];

    if(Verbose)
	printf("*** Resume  %s\n", f->f_name);

    Do_name = TRUE;		/* Next time do_line called, name it*/
    Bufc = f->f_bufc;
    Bufp = f->f_bufp;
    Lasteol = f->f_lasteol;
    Line = f->f_line;

#if	DEBUG
if(Debug) printf("popfile: Bufc=%d, Bufp=%04x, Lasteol=%d, Line=%d\n",
		Bufc, Bufp, Lasteol, Line);
#endif	/* DEBUG */

    return(TRUE);		/* All is ok -- return success */
    }

/************************************************************************/
/*									*/
/*	readline							*/
/*									*/
/*	Read and edit a line into buffer with macro expansion.		*/
/*									*/
/************************************************************************/
char *
readline(buf, bufsize, flags)
  register char *buf;
  register int bufsize;
  register int	flags;
    {
    extern char *addstr();
    extern char	*docall();
    extern int gettoken();
    extern struct symtab *lookup();

    static char rbo[] = "Read buffer overflow";

    register char *bufp;
    struct symtab *sy;
    register int t;

    for(bufp = buf; (t = gettoken(flags)) != '\n'; )
	{
	if(t == EOF)
	    end_of_file();
	if((t == LETTER) && ((sy = lookup(Token, NULL)) != NULL) &&
					(sy->disable != TRUE))
	    {
	    bufp = docall(sy, bufp, &buf[bufsize - 1]);
	    }
	else
	    bufp = addstr(bufp, &buf[bufsize - 1], rbo, Token);
	}

    pushback('\n');
    *bufp = '\0';

    for(bufp = buf; istype(*bufp & 0xFF, C_W); ++bufp)
	;	/* Skip leading blanks */

    return(bufp);
    }

/************************************************************************/
/*									*/
/*	scaneol								*/
/*									*/
/*	Scan for end of current line.					*/
/*									*/
/************************************************************************/
scaneol()
    {
    extern int gettoken();

    register int t;

    while((t = gettoken(GT_STR)) != '\n')
	{
	if(t == EOF)
	    return;	/* Absorb chars */
	}
    pushback('\n');	/* So the newline is seen */
    }

#if	HOST == ISQC
/************************************************************************/
/*									*/
/*	set_user							*/
/*									*/
/*	Set disk/user to that of current input file.			*/
/*									*/
/************************************************************************/
set_user()
    {
    /* Don't change if < 0 */
    if(Filestack[Filelevel]->f_user >= 0)
	{
	bdos(BDOS_SELDISK, Filestack[Filelevel]->f_disk);
	bdos(BDOS_USER, Filestack[Filelevel]->f_user);
	}
    }
#endif	/* HOST == ISQC */

/************************************************************************/
/*									*/
/*	trigraph							*/
/*									*/
/*	Handles ANSI trigraph substitutions (alternate to "gchbuf").	*/
/*									*/
/************************************************************************/
int
trigraph()
    {
    extern int gchbuf();

    int c;
    int q_count;

    if((c = gchbuf()) == '?')
	{
	if((c = gchbuf()) == '?')
	    {
	    q_count = 0;		/* Extra '?' encountered */
	    for(;;)
		{
		switch((int) (c = gchbuf()))
		    {
		    case '=':
				c = '#';
				break;
		    case '(':
				c = '[';
				break;
		    case '/':
				c = '\\';
				break;
		    case ')':
				c = ']';
				break;
		    case '\'':
				c = '^';
				break;
		    case '<':
				c = '{';
				break;
		    case '!':
				c = '|';
				break;
		    case '>':
				c = '}';
				break;
		    case '-':
				c = '~';
				break;
		    case '?':
				q_count++;
				continue;
		    default:
				pushback(c);
				pushback('?');
				c = '?';
				break;
		    }
/*
 *	Handle stuff like "????????????#" correctly.
 */
		if(q_count > 0)
		    {
		    pushback(c);
		    while(q_count-- > 1)
			pushback('?');
		    c = '?';
		    }
		break;
		}
	    }
	else
	    {
	    pushback(c);
	    c = '?';
	    }
	}
    return (c);
    }
