/************************************************************************/
/*									*/
/*	Program:	cioext.c					*/
/*									*/
/*	This module is a simple example of a server extension package	*/
/*	for use with the Logical Systems CIO "C" I/O handler.  The	*/
/*	intended purpose of this version is to provide a basis for	*/
/*	building other (user specific), versions.  This example is	*/
/*	written with Microsoft's MS-DOS and "C" (V5.1) in mind.  Other	*/
/*	operating system and compiler environments will force changes	*/
/*	both here and possibly in the CIO server ("cio.c").		*/
/*									*/
/*	Server extension functions provided by this example:		*/
/*									*/
/*		0	"segread()"					*/
/*		1	"intdos()"					*/
/*									*/
/*	Note that this program MUST be compiled with the SMALL code and	*/
/*	data model to function correctly with "cio.c".			*/
/*									*/
/*	Written by Kirk Bailey and Gary Oliver.  Copyright 1988 by	*/
/*	Logical Systems.  All rights reserved.				*/
/*									*/
/************************************************************************/

#include	<dos.h>
#include	"cioext.h"		/* Extension macro definitions */

/************************************************************************/
/*									*/
/*			Function Type Definitions			*/
/*									*/
/************************************************************************/

extern	int	_cioext();
extern	int	do_intdos();
extern	int	do_segread();

/************************************************************************/
/*									*/
/*	Routine:	_cioext						*/
/*									*/
/*	Routine called by CIO "main" to handle server extensions.	*/
/*									*/
/*	Parameters:							*/
/*		"code"	The server extension function #.		*/
/*		"sbuf"	A pointer to the buffer sent by the client.	*/
/*		"sbuflen" The length of "sbuf" in bytes.		*/
/*		"rbuf"	A pointer to the return buffer for the client.	*/
/*		"rbuflen" The expected length of "rbuf" in bytes.	*/
/*									*/
/*	Returns:	The desired return code for the client		*/
/*			"_cioext" call to give.				*/
/*									*/
/************************************************************************/

int
_cioext(code,sbuf,sbuflen,rbuf,rbuflen)
	int			code;
	void			*sbuf;
	int			sbuflen;
	void			*rbuf;
	int			rbuflen;
	{
	switch((int) code)		/* Desired function code */
		{
		case	CXF_INTDOS:
				return (do_intdos(sbuf,rbuf));

		case	CXF_SEGREAD:
				return (do_segread(rbuf));

		default:
				return (-1);	/* Not implemented error */
		}
/*lint -e533 thinks nothing is returned */
	}
/*lint +e533 thinks nothing is returned */

/************************************************************************/
/*									*/
/*	Routine:	do_intdos					*/
/*									*/
/*	Helper function for "do_int" to handle a "intdos()" extended	*/
/*	server command.							*/
/*									*/
/*	Parameters:							*/
/*		"input"	A pointer to the input buffer from the client.	*/
/*		"output"A pointer to the reply buffer to the client.	*/
/*									*/
/*	Returns:	Error status.					*/
/*									*/
/************************************************************************/

int
do_intdos(input,output)
	void			*input;
	void			*output;
	{
	return (intdos((union REGS *) input,(union REGS *) output));
	}

/************************************************************************/
/*									*/
/*	Routine:	do_segread					*/
/*									*/
/*	Helper function for "do_int" to handle a "segread()" extended	*/
/*	server command.							*/
/*									*/
/*	Parameters:							*/
/*		"output"A pointer to the reply buffer to the client.	*/
/*									*/
/*	Returns:	None (dummy 0).					*/
/*									*/
/************************************************************************/

int
do_segread(output)
	void			*output;
	{
	segread((struct SREGS *) output);
	return (0);			/* No return value */
	}
