/************************************************************************/
/*									*/
/*	File:	tasm3.c							*/
/*									*/
/*	Transputer Assembler (TASM), source file 3.  This file contains	*/
/*	the constant expression evaluation code.  This code is loosely	*/
/*	based on code written and modified by Gary Oliver and Dave	*/
/*	Regan (among others).						*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character classification macros */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tasmdef.h"		/* Assembler definitions */
#include	"tasmtyp.h"		/* Assembler data declarations */
#include	"tasmext.h"		/* Assembler externals */

/************************************************************************/
/*									*/
/*	Routine:	eval						*/
/*									*/
/*	Evaluate a "C" style constant expression using the normal	*/
/*	recursive descent technique.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval()
	{
	SLONG			val;

	val = eval_ternary();		/* Evaluate expression */
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_ternary					*/
/*									*/
/*	Handle ?: operator.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_ternary()
	{
	SLONG			val;
	SLONG			val_false;
	SLONG			val_true;

	val = eval_lor();		/* Look for expression */
	if(! match("?"))
		return (val);		/* ? part of ?: */
/*
 *	We know we have a ?: operator, evaluate it.
 */
	val_true = eval_ternary();	/* Get true-side expression */
	if(! match(":"))
		expr_err = TRUE;	/* Signal we had an error */
	val_false = eval_ternary();	/* Get false-side expression */
	if(val)
		return (val_true);	/* Return true side expression */
	else
		return (val_false);	/* Return false side expression */
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_lor					*/
/*									*/
/*	Handle || operator.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_lor()
	{
	SLONG			val;

	for(val = eval_land(); match("||"); )
		{
		if(eval_land())
			val = TRUE;
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_land					*/
/*									*/
/*	Handle && operator.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_land()
	{
	SLONG			val;

	for(val = eval_bor(); match("&&"); )
		{
		if(! eval_bor())
			val = FALSE;
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_bor					*/
/*									*/
/*	Handle | operator.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_bor()
	{
	SLONG			val;

	for(val = eval_bxor(); (! look("||")) && match("|"); )
		val |= eval_bxor();
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_bxor					*/
/*									*/
/*	Handle ^ operator.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_bxor()
	{
	SLONG			val;

	for(val = eval_band(); match("^"); )
		val ^= eval_band();
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_band					*/
/*									*/
/*	Handle & operator.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_band()
	{
	SLONG			val;

	for(val = eval_eq(); (! look("&&")) && match("&"); )
		val &= eval_eq();
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_eq						*/
/*									*/
/*	Handle == and != operators.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_eq()
	{
	SLONG			val;

	for(val = eval_rel(); ; )
		{
		if(match("=="))
			val = (val == eval_rel());
		else if(match("!="))
			val = (val != eval_rel());
		else
			break;
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_rel					*/
/*									*/
/*	Handle <, <=, > and >= operators.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_rel()
	{
	SLONG			val;

	for(val = eval_sh(); ; )
		{
		if(match("<="))
			val = (val <= eval_sh());
		else if(match(">="))
			val = (val >= eval_sh());
		else if(match("<"))
			val = (val < eval_sh());
		else if(match(">"))
			val = (val > eval_sh());
		else
			break;
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_sh						*/
/*									*/
/*	Handle << and >> operators.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_sh()
	{
	SLONG			temp;
	SLONG			val;

	for(val = eval_sum(); ; )
		{
		if(match("<<"))
			{
			temp = eval_sum();	/* Kludge for PCC bug */
			val <<= temp;
			}
		else if(match(">>"))
			{
			temp = eval_sum();	/* Kludge for PCC bug */
			val >>= temp;
			}
		else
			break;
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_sum					*/
/*									*/
/*	Handle binary + and - operators.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_sum()
	{
	SLONG			val;

	for(val = eval_mdr(); ; )
		{
		if(match("+"))
			val += eval_mdr();
		else if(match("-"))
			val -= eval_mdr();
		else
			break;
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_mdr					*/
/*									*/
/*	Handle *, / and % operators.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_mdr()
	{
	SLONG			temp;
	SLONG			val;

	for(val = eval_ucom(); ; )
		{
		if(match("*"))
			val *= eval_ucom();
		else if(match("/"))
			{
			temp = eval_ucom();
			if(temp == 0)
				{
				expr_err = TRUE;
				val = 0;
				}
			else
				val /= temp;
			}
		else if(match("%"))
			{
			temp = eval_ucom();
			if(temp == 0)
				{
				expr_err = TRUE;
				val = 0;
				}
			else
				val %= temp;
			}
		else
			break;
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_ucom					*/
/*									*/
/*	Handle unary ~ operator.  Since all unary routines have the	*/
/*	same precedence and associativity this routine has been chosen	*/
/*	by fiat to be the "top" level routine which is called		*/
/*	recursively after a unary operator is recognized.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_ucom()
	{
/*lint -e502 lint doesn't like complementing signed integral types */
	if(match("~"))
		return (~eval_ucom());
/*lint +e502 lint doesn't like complementing signed integral types */
	else
		return (eval_unot());
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_unot					*/
/*									*/
/*	Handle unary ! operator.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_unot()
	{
	if(match("!"))
		return (! eval_ucom());
	else
		return (eval_usum());
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_usum					*/
/*									*/
/*	Handle unary + and - operators.  The plus is not required for	*/
/*	ordinary "C" constant expressions but is allowed here as a	*/
/*	seperator between the optional symbol name and the constant	*/
/*	expression.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_usum()
	{
	if(match("+"))
		return (eval_ucom());
	else if(match("-"))
		return (-eval_ucom());
	else
		return (eval_value());
	}

/************************************************************************/
/*									*/
/*	Routine:	eval_value					*/
/*									*/
/*	Handle numbers, character constants or nested expressions.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The value of the expression.			*/
/*									*/
/************************************************************************/

SLONG
eval_value()
	{
	register	int	c;
	register	UINT	count;
	STRING			s;
	UINT			st_index;
	SLONG			val;
/*
 *	First check for nested expressions.
 */
	if(match("("))
		{
		val = eval_ternary();	/* Evaluate it */
		if(! match(")"))
			expr_err = TRUE;
		}
/*
 *	Check for symbolic constants.
 */
	else if(symbol(*parse_ptr))
		{
		for(s = parse_ptr + 1; (symbol(*s) || isdigit(*s)); s++)
			;
		c = *s;			/* Save spot which will be zeroed */
		*s = '\0';		/* Temporary termination */
		st_index = lookup(parse_ptr);	/* Get symbol # */
		*s = c;			/* Restore old contents */
		parse_ptr = s;		/* Point past symbol */
		if(st[st_index].type & (S_CONSTANT | S_VALUE))
			val = st[st_index].min_value;
		else
			expr_err = TRUE;	/* Not a "constant" symbol */
		}
/*
 *	Check for character constants.
 */
	else if(match("'"))
		{
		if((val = *parse_ptr++) == '\\')	/* Test for escape */
			{
			switch((char) (val = *parse_ptr++))
				{
				case 'a':	val = '\a';	break;
				case 'b':	val = '\b';	break;
				case 'f':	val = '\f';	break;
#if	H_OPSYS == MPW
				case '\012':	val = '\012';	break;
#else
				case 'n':	val = '\n';	break;
#endif	/* H_OPSYS == MPW */
				case 'r':	val = '\r';	break;
				case 't':	val = '\t';	break;
				case 'v':	val = '\v';	break;
				case '\\':	val = '\\';	break;
				case '\'':	val = '\'';	break;
				case '\"':	val = '\"';	break;
				case '?':	val = '?';	break;

/*	Octal escape */		case '0': case '1': case '2': case '3':
				case '4': case '5': case '6': case '7':
					for(count = 2, val -= '0';
						(isoct(c = *parse_ptr) &&
						(count <= 3)); count++,
							parse_ptr++)
						{
						val = (val << 3) + c - '0';
						}
					break;

/*	Hex escape */		case 'x':
					if(! ishex(*parse_ptr))
						{
						expr_err = TRUE;
						break;
						}
					val = hexbin(*parse_ptr++);
					for(count = 2;
						(ishex(c = *parse_ptr) &&
						(count <= 3)); count++,
							parse_ptr++)
						{
						val = (val << 4) + hexbin(c);
						}
					break;

				default:
					expr_err = TRUE;
					break;
				}
			}
		if(*parse_ptr++ != '\'')
			expr_err = TRUE;
		}
/*
 *	Otherwise it must be a number of some base.
 */
	else				
		{
		skip_space();		/* Skip whitespace */
		c = *parse_ptr++;
		if(isdigit(c))		/* Is it a number? */
			{
			val = 0;
/*
 *	See if it is a octal or hex # or if it is a decimal #.
 */
			if(c == '0')
				{
				c = *parse_ptr++;
/*
 *	Its either octal or hex, figure out which.
 */
				if((c == 'x') || (c == 'X'))
					{
/*
 *	Its a hexadecimal #.
 */
					c = *parse_ptr++;
					while(ishex(c))
						{
						val = (val << 4) + hexbin(c);
						c = *parse_ptr++;
						}
					parse_ptr--;
					}
				else
/*
 *	Its an octal #.
 */
					{
					while(isoct(c))
						{
						val = (val << 3) + c - '0';
						c = *parse_ptr++;
						}
					parse_ptr--;
					}
				}
			else
/*
 *	Its a decimal #.
 */
				{
				val = (val * 10) + c - '0';
				c = *parse_ptr++;
				while(isdigit(c))
					{
					val = (val * 10) + c - '0';
					c = *parse_ptr++;
					}
				parse_ptr--;
				}
			}
		else
/*
 *	Wasn't any # we could recognize!
 */
			{
			expr_err = TRUE;
			}
		}
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	hexbin						*/
/*									*/
/*	Convert a hexadecimal digit to binary.				*/
/*									*/
/*	Parameters:							*/
/*		"c"	The digit to convert.				*/
/*									*/
/*	Returns:	The resulting binary value.			*/
/*									*/
/************************************************************************/

SLONG
hexbin(c)
	int			c;
	{
	if(isdigit(c))
		return ((SLONG) (c - '0'));
	else
		return ((SLONG) ((isupper(c) ? tolower(c) : c) - 'a' + 10));
	}

/************************************************************************/
/*									*/
/*	Routine:	ishex						*/
/*									*/
/*	Test a digit to see if it could be part of a hexadecimal #.	*/
/*									*/
/*	Parameters:							*/
/*		"c"	The digit to test.				*/
/*									*/
/*	Returns:	TRUE if so, FALSE if not.			*/
/*									*/
/************************************************************************/

BOOL
ishex(c)
	int			c;
	{
	return (isdigit(c) ||
		((c >= 'a') && (c <= 'f')) || ((c >= 'A') && (c <= 'F')));
	}

/************************************************************************/
/*									*/
/*	Routine:	isoct						*/
/*									*/
/*	Test a digit to see if it could be part of a octal #.		*/
/*									*/
/*	Parameters:							*/
/*		"c"	The digit to test.				*/
/*									*/
/*	Returns:	TRUE if so, FALSE if not.			*/
/*									*/
/************************************************************************/

BOOL
isoct(c)
	int			c;
	{
	return ((c >= '0') && (c <= '7'));
	}

/************************************************************************/
/*									*/
/*	Routine:	look						*/
/*									*/
/*	Used by the expression evaluation routines to check and see if	*/
/*	the next characters pointed at by "parse_ptr" match the string	*/
/*	passed to "look".  If so TRUE is returned but "parse_ptr" is	*/
/*	not affected.  Preceeding whitespace is skipped ("parse_ptr"),	*/
/*	prior to performing the match.					*/
/*									*/
/*	Parameters:							*/
/*		"str"	A pointer to the string being compared.		*/
/*									*/
/*	Returns:	TRUE if successfully matched.			*/
/*									*/
/************************************************************************/

BOOL
look(str)
	STRING			str;
	{
	register	STRING	temp;

	skip_space();
	temp = parse_ptr;
	while(*str != '\0')
		{
		if(*str++ != *temp++)
			return (FALSE);
		}
	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	match						*/
/*									*/
/*	Used by the expression evaluation routines to check and see if	*/
/*	the next characters pointed at by "parse_ptr" match the string	*/
/*	passed to "match".  If so "parse_ptr" is updated to point	*/
/*	beyond the matched item.  Preceeding whitespace is skipped	*/
/*	("parse_ptr"), prior to performing the match.			*/
/*									*/
/*	Parameters:							*/
/*		"str"	A pointer to the string being compared.		*/
/*									*/
/*	Returns:	TRUE if successfully matched.			*/
/*									*/
/************************************************************************/

BOOL
match(str)
	STRING			str;
	{
	register	STRING	temp;

	skip_space();			/* Skip any whitespace */
	temp = parse_ptr;
	while(*str != '\0')
		{
		if(*str++ != *temp++)
			return (FALSE);
		}
	parse_ptr = temp;		/* Point past matched string */
	return (TRUE);
	}
