/************************************************************************/
/*									*/
/*	File:	tasm6.c							*/
/*									*/
/*	Transputer Assembler (TASM), source file 6.  This file contains	*/
/*	the floating point conversion code needed for the ".real32" and	*/
/*	".real64" pseudo-opcodes.					*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1987-1989 by Logical Systems.		*/
/*									*/
/*				08/10/87				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character class include file */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tasmdef.h"		/* Assembler definitions */
#include	"tasmtyp.h"		/* Assembler data declarations */
#include	"tasmext.h"		/* Assembler externals */

/************************************************************************/
/*									*/
/*	Routine:	atoieee						*/
/*									*/
/*	Convert a string to an IEEE format single precision floating	*/
/*	point number.  The return value is the binary floating point	*/
/*	representation of the string but is declared "long" since the	*/
/*	internal representation of floating point numbers is NOT very	*/
/*	portable.							*/
/*									*/
/*	Parameters:	"result" -	A pointer to where to store the	*/
/*					resulting 16 or 32 bit number.	*/
/*			"full" -	TRUE if full double precision	*/
/*					result needed, FALSE for single	*/
/*					precision.			*/
/*									*/
/*	Returns:	ERRORS if an error was encountered, NOERRORS if	*/
/*			OK.						*/
/*									*/
/************************************************************************/

#define	MAXEXP		500		/* Maximum single +/- exponent range */

BOOL
atoieee(result,full)
	SLONG			result[];
	BOOL			full;
	{
	register	int	c;
	register	int	exp;
	int			esign;
	int			expmod;
	ULONG			fract[2];
	int			fsign;
	register	int	itemp;
	ULONG			ltemp[2];

	fract[0] = fract[1] = exp = 0;	/* Reset accumulators */

	skip_space();
/*
 *	Process sign indication for number (if any).
 */
	if((c = *parse_ptr++) == '-')
		{
		fsign = -1;
		c = *parse_ptr++;
		}
	else if(c == '+')
		{
		fsign = 1;
		c = *parse_ptr++;
		}
	else
		fsign = 1;
/*
 *	Process fraction up to the (optional), decimal point.
 */
	while(isdigit(c))
		{
		if(fract[1] < ((MAXULONG - 9) / 10))
			{		/* Multiply "fract" by 10 */
			ltemp[0] = fract[0];
			ltemp[1] = fract[1];
			llshift(fract);
			llshift(fract);
			ladd(fract,ltemp);
			llshift(fract);
			ltemp[0] = c - '0';
			ltemp[1] = 0;
			ladd(fract,ltemp);	/* Add digit */
			}
		else
			exp++;		/* At precision limit, kick exponent */
		c = *parse_ptr++;
		}
/*
 *	See if we got a decimal point, if so continue accumulating fraction.
 */
	if(c == '.')
		{
		for(c = *parse_ptr++; isdigit(c); c = *parse_ptr++)
			{
			if(fract[1] < ((MAXULONG - 9) / 10))
				{	/* Multiply "fract" by 10 */
				ltemp[0] = fract[0];
				ltemp[1] = fract[1];
				llshift(fract);
				llshift(fract);
				ladd(fract,ltemp);
				llshift(fract);
				ltemp[0] = c - '0';
				ltemp[1] = 0;
				ladd(fract,ltemp);	/* Add digit */
				exp--;
				}
			}
		}
/*
 *	See if we have an exponent and process exponent sign.
 */
	if((c == 'E') || (c == 'e'))
		{
		expmod = exp;		/* Remember exponent modifier */
		exp = 0;		/* Reset new exponent accumulator */
		if((c = *parse_ptr++) == '-')	/* Check for exponent sign */
			{
			esign = -1;
			c = *parse_ptr++;
			}
		else if(c == '+')
			{
			esign = 1;
			c = *parse_ptr++;
			}
		else
			esign = 1;
/*
 *	Process exponent value if one is present.
 */
		while(isdigit(c))
			{
			if(exp < ((MAXINT - 9) / 10))
				exp = (exp * 10) + c - '0';	/* Add digit */
			c = *parse_ptr++;
			}
/*
 *	Combine exponent modifier and explicit exponent (if any), into the
 *	final value of the base 10 exponent.  Clip the resulting value so
 *	that the following routines won't loop forever if a bogus value is
 *	being parsed.
 */
 		if(esign < 0)
			exp = expmod - exp;
		else
			exp += expmod;
		}
/*
 *	We have parsed number.  Backup "parse_ptr" to whatever stopped us and
 *	convert number into appropriate floating point format.
 */
	parse_ptr--;
	if((fract[0] == 0) && (fract[1] == 0))	/* If number was zero */
		{
		if(full)
			{
			result[0] = fract[0];
			result[1] = fract[1];
			}
		else
			result[0] = fract[1];
		return (NOERRORS);
		}
	if(exp > MAXEXP)
		return (ERRORS);
	else if(exp < -(MAXEXP + 1))
		return (ERRORS);
/*
 *	Now we convert the fraction times base 10 exponent to a fraction times
 *	base 2 exponent.
 */
	itemp = exp;			/* Remember exponent base 10 */
	exp = 0;			/* Reset exponent base 2 */
	if(itemp > 0)			/* If zero, not much to do */
		{
		while(itemp--)		/* Count down base 10 exponent */
			{
/*
 *	Make room in the fraction for the multiplication by 10 by
 *	shifting it right, rounding the result, and incrementing the base 2
 *	exponent.
 */
			while(fract[1] >= (MAXULONG / 5))
				{
				ltemp[0] = fract[0] & 1;
				ltemp[1] = 0;
				lrshift(fract);
				ladd(fract,ltemp);	/* Round */
				exp++;		/* Increment binary exponent */
				}
			ltemp[0] = fract[0];
			ltemp[1] = fract[1];
			llshift(fract);
			llshift(fract);
			ladd(fract,ltemp);
			exp++;			/* *10 = (*5) * (*2) */
			}
		}
	else if(itemp < 0)
		{
		while(itemp++)		/* Count up base 10 exponent */
			{
/*
 *	Shift the fraction as far left as possible prior to dividing
 *	it by 10 to minimize the roundoff error.
 */
			while(fract[1] <= (MAXULONG / 2))
				{
				llshift(fract);
				exp--;
				}
			ldiv5(fract);
			exp--;		/* Last /2 factor to get /10 */
			}
		}
/*
 *	Okay, the fraction is the final fraction to the appropriate base 2
 *	exponent, and we know the desired sign of the exponent.  Now cram it
 *	into the 32 or 64 bit IEEE formats as appropriate.
 */
	if(full)
		{
		while(fract[1] >= 0x00200000)
			{
			ltemp[0] = fract[0] & 1;
			ltemp[1] = 0;
			lrshift(fract);
			ladd(fract,ltemp);	/* Round */
			exp++;
			}
		while(fract[1] < 0x00100000)
			{
			llshift(fract);
			exp--;
			}
/*
 *	Unlike the earlier rough exponent clipping, this time we make sure
 *	the number can be represented (if not, we make it fit anyway).
 */
		if((exp += (1023 + 52)) >= 2047)
			return (ERRORS);
		else if(exp < 0)
			return (ERRORS);
		ltemp[1] = exp;
		fract[1] = (fract[1] & 0x000FFFFF) | (ltemp[1] << 20);
		if(fsign < 0) fract[1] |= 0x80000000;
		result[0] = fract[0];
		result[1] = fract[1];
		}
	else
		{
		while(fract[1] >= 0x01000000)
			{
			ltemp[0] = fract[0] & 1;
			ltemp[1] = 0;
			lrshift(fract);
			ladd(fract,ltemp);	/* Round */
			exp++;
			}
		while(fract[1] < 0x00800000)
			{
			llshift(fract);
			exp--;
			}
/*
 *	Since we are only returning the top word as our result mantissa, round
 *	it based on the low order word.
 */
		if(fract[0] & 0x80000000)
			fract[1]++;
		if(fract[1] >= 0x01000000)	/* Handle carry */
			{
			lrshift(fract);
			exp++;
			}
/*
 *	Unlike the earlier rough exponent clipping, this time we make sure
 *	the number can be represented (if not, we make it fit anyway).
 */
		if((exp += (127 + 55)) >= 255)
			return (ERRORS);
		else if(exp < 0)
			return (ERRORS);
		ltemp[1] = exp;
		fract[1] = (fract[1] & 0x007FFFFF) | (ltemp[1] << 23);
		if(fsign < 0)
			fract[1] |= 0x80000000;
		result[0] = fract[1];
		}
	return (NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	ladd						*/
/*									*/
/*	A helping function for "atoieee" which treats an array of two	*/
/*	ULONG integers as one 64 bit word and adds on 64 bit word to	*/
/*	the other.							*/
/*									*/
/*	Parameters:	"result"-	A pointer to the array of two	*/
/*					ULONG integers which is both	*/
/*					one of the input operands and	*/
/*					the place to store the result.	*/
/*			"value"	-	The other input operand.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
ladd(result,value)
	ULONG			result[];
	ULONG			value[];
	{
	ULONG			carry;
	ULONG			temp;

	carry = 0;
	temp = result[0] + value[0];

	if(result[0] & 0x80000000)
		{
		if((value[0] & 0x80000000) || ((temp & 0x80000000) == 0))
			carry = 1;
		}
	else
		{
		if((value[0] & 0x80000000) && ((temp & 0x80000000) == 0))
			 carry = 1;
		}
	result[0] = temp;
	result[1] = result[1] + value[1] + carry;
	}

/************************************************************************/
/*									*/
/*	Routine:	llshift						*/
/*									*/
/*	A helping function for "atoieee" which treats an array of two	*/
/*	ULONG integers as one 64 bit word and shifts the combination	*/
/*	left one bit.							*/
/*									*/
/*	Parameters:	"value"	-	A pointer to the array of two	*/
/*					ULONG integers.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
llshift(value)
	ULONG			value[];
	{
	if(value[0] & 0x80000000)
		value[1] = (value[1] << 1) | 1;
	else
		value[1] <<= 1;
	value[0] <<= 1;
	}

/************************************************************************/
/*									*/
/*	Routine:	lrshift						*/
/*									*/
/*	A helping function for "atoieee" which treats an array of two	*/
/*	ULONG integers as one 64 bit word and shifts the combination	*/
/*	right one bit.							*/
/*									*/
/*	Parameters:	"value"	-	A pointer to the array of two	*/
/*					ULONG integers.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
lrshift(value)
	ULONG			value[];
	{
	if(value[1] & 1)
		value[0] = (value[0] >> 1) | 0x80000000;
	else
		value[0] = (value[0] >> 1) & 0x7FFFFFFF;
	value[1] = (value[1] >> 1) & 0x7FFFFFFF;
	}

/************************************************************************/
/*									*/
/*	Routine:	ldiv5						*/
/*									*/
/*	A helping function for "atoieee" which treats an array of two	*/
/*	ULONG integers as one 64 bit word and divides the word by 5.	*/
/*	The technique used is a variation on Newton-Raphson iteration.	*/
/*	Note that the result is ROUNDED, not TRUNCATED!			*/
/*									*/
/*	Parameters:	"value"	-	A pointer to the array of two	*/
/*					ULONG integers.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
ldiv5(value)
	ULONG			value[];
	{
	ULONG			result[2];
	register	int	shift;
	ULONG			temp1[2];
	ULONG			temp2[2];

	result[0] = 0;			/* Result accumulator */
	result[1] = value[1] / 5;

	temp1[0] = value[0];		/* Dividend for this pass */
	temp1[1] = value[1] % 5;

	temp2[1] = 0;

	FOREVER
		{
/*
 *	Now do the general iteration case.
 */
		for(shift = 0; temp1[1] != 0; shift++)
			lrshift(temp1);
		temp2[0] = temp1[0] / 5;
		if(temp2[0] == 0)
			{
/*
 *	Round result.
 */
			if(temp1[0] % 5 > (5 / 2))
				{
				temp1[0] = 1;
				ladd(result,temp1);
				}
			break;
			}
		temp1[0] = temp2[0];
		while(shift-- != 0)
			llshift(temp1);
		ladd(result,temp1);	/* Update result accumulator */
		temp1[0] = result[0];
		temp1[1] = result[1];
		llshift(temp1);		/* Compute (current_result*5) */
		llshift(temp1);
		ladd(temp1,result);
		temp1[0] = ~temp1[0];	/* Compute -(current_result*5) */
		temp1[1] = ~temp1[1];
		temp2[0] = 1;
		ladd(temp1,temp2);
		ladd(temp1,value);	/* Compute #-(current_result*5) */
		}
/*
 *	All done, return result.
 */
	value[0] = result[0];
	value[1] = result[1];
	}
