/************************************************************************/
/*									*/
/*	File:	tasmext.h						*/
/*									*/
/*	External declarations for the Transputer Assembler (TASM).	*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

/*
 *	I/O definitions.
 */

extern	char	ifnbuf[];		/* Input filename buffer */
extern	char	lifnbuf[];		/* Last filename in a "#line" stmt */
extern	STREAM	in_fp;			/* Input file pointer */
extern	char	ofnbuf[];		/* Output filename buffer */
extern	STREAM	out_fp;			/* Output file pointer */
extern	char	listbuf[];		/* Listing filename buffer */
extern	STREAM	list_fp;		/* Listing file pointer */
extern	char	temp1buf[];		/* Temp file #1 filename buffer */
extern	STREAM	temp1_fp;		/* Temp file #1 file pointer */
extern	char	temp2buf[];		/* Temp file #2 filename buffer */
extern	STREAM	temp2_fp;		/* Temp file #2 file pointer */

extern	STREAM	read_temp;		/* Temp file being read */
extern	STREAM	write_temp;		/* Temp file being written */

/*
 *	General definitions.
 */

extern	UCHAR	acc_dta[];		/* T_DATA pack accumulator */
extern	int	acc_dtasize;		/* T_DATA pack accumulated size */
extern	SLONG	acc_storesize;		/* T_STORAGE accumulated size */
extern	SLONG	aligned[];		/* Loc. counter for last ".align" */
extern	BOOL	eject;			/* TRUE if formfeed on next file */
extern	BOOL	compress;		/* TRUE if to pack T_DATA/T_STORAGE */
extern	SLONG	con_expr;		/* Expression value from operand */
extern	UINT	cpu_type;		/* CPU specific opcodes bitmask */
extern	BOOL	emulate;		/* TRUE if instruction emulation ON */
extern	UINT	errors;			/* Total # of errors encountered */
extern	BOOL	expr_err;		/* TRUE if an expression error found */
extern	BOOL	had_amper;		/* TRUE if operand symbol was "@" */
extern	BOOL	had_diff;		/* TRUE if operand was "rel-rel" */
extern	BOOL	had_dollar;		/* TRUE if operand had "$" prefix */
extern	BOOL	had_expr;		/* TRUE if operand contained expr */
extern	BOOL	had_symbol;		/* TRUE if operand contained symbol */
extern	UINT	max_i_size;		/* Max. instr. size in bytes */
extern	SLONG	max_loc[];		/* Current max location counter */
extern	SLONG	max_loc_adj[];		/* "max_loc" adj for fwd ref */
extern	SLONG	min_loc[];		/* Current min location counter */
extern	SLONG	min_loc_adj[];		/* "min_loc" adj for fwd ref */
extern	UINT	module;			/* Current module # */
extern	UINT	number;			/* Current input line # */
extern	int	optimize;		/* Optimization level */
extern	char	parse_buf[];		/* Parsing buffer */
extern	STRING	parse_ptr;		/* Parsing buffer pointer */
extern	BOOL	rel_only;		/* TRUE if making relocatable code */
extern	BOOL	retf_active;		/* TRUE if holding a ".retf" pseudo */
extern	UINT	retf_number;		/* Line # ".retf" appeared on */
extern	SLONG	retf_value;		/* Value of ".retf" operand field */

#if	H_OPSYS == MPW
extern	struct	symtab	*st;		/* Symbol table */
#else	/* H_OPSYS == MPW */
extern	struct	symtab	st[];		/* Symbol table */
#endif	/* H_OPSYS == MPW */

extern	UINT	st_num1,st_num2;	/* ST entry #'s for operand symbols */
extern	BOOL	translator;		/* TRUE to source is translated */
extern	BOOL	verbose;		/* Default: make output verbose */
extern	UINT	word_size;		/* Processor word size in bytes */

/*
 *	Function definitions.
 */

extern	VOID	addr();
extern	SLONG	adj_max_loc();
extern	SLONG	adj_min_loc();
extern	VOID	align();
extern	VOID	all();
extern	VOID	assemble();
extern	VOID	banner();
extern	UINT	bind_constant();
extern	VOID	build_i();
extern	BOOL	check_def();
extern	VOID	copy();
extern	VOID	copy_out();
extern	VOID	data();
extern	VOID	db();
extern	VOID	dbnz();
extern	VOID	do_addr();
extern	BOOL	do_diff();
extern	VOID	do_em();
extern	VOID	do_extern();
extern	VOID	do_float();
extern	BOOL	do_i_addr();
extern	BOOL	do_i_diff();
extern	BOOL	do_i_wdiff();
extern	VOID	do_module();
extern	VOID	do_noem();
extern	VOID	do_norel();
extern	VOID	do_public();
extern	VOID	do_rel();
extern	VOID	do_retf();
extern	VOID	do_set();
extern	VOID	do_sym();
extern	VOID	do_val();
extern	BOOL	dor_i_ref();
extern	BOOL	dor_ref();
extern	VOID	do_set();
extern	VOID	ds();
extern	VOID	dw();
extern	SLONG	eval_ternary();
extern	SLONG	eval_lor();
extern	SLONG	eval_land();
extern	SLONG	eval_bor();
extern	SLONG	eval_bxor();
extern	SLONG	eval_band();
extern	SLONG	eval_eq();
extern	SLONG	eval_rel();
extern	SLONG	eval_sh();
extern	SLONG	eval_sum();
extern	SLONG	eval_mdr();
extern	SLONG	eval_ucom();
extern	SLONG	eval_unot();
extern	SLONG	eval_usum();
extern	SLONG	eval_value();
extern	VOID	exit();
extern	VOID	fatal();
extern	VOID	fatali();
extern	VOID	filename();
extern	BOOL	findit();
extern	VOID	flush_retf();
extern	VOID	fpu_i();
extern	VOID	grow_ref();
extern	VOID	grow_diff();
extern	VOID	grow_wdiff();
extern	SLONG	hexbin();
extern	BOOL	ishex();
extern	BOOL	isoct();
extern	UINT	i_size();
extern	VOID	ladd();
extern	VOID	ldiv5();
extern	VOID	list_data();
extern	VOID	list_hex();
extern	VOID	list_lnum();
extern	VOID	list_pass();
extern	VOID	list_put();
extern	VOID	list_range();
extern	VOID	list_src();
extern	VOID	list_title();
extern	VOID	list_write();
extern	VOID	llshift();
extern	VOID	lrshift();
extern	VOID	loc_pad();
extern	BOOL	look();
extern	UINT	lookup();
extern	VOID	main();
extern	VOID	makefn();
extern	int	nc1();
extern	int	nextc();
extern	BOOL	note_extern();
extern	BOOL	note_public();
extern	VOID	noti();
extern	VOID	op_i();
extern	VOID	opr_i();
extern	VOID	out_rec();
extern	VOID	out_write();
extern	VOID	out_1b();
extern	VOID	out_2b();
extern	VOID	out_4b();
extern	BOOL	parse();
extern	VOID	pass0();
extern	VOID	pass2();
extern	VOID	pass3();
extern	VOID	pfix_i();
extern	VOID	post_eof();
extern	VOID	post_error();
extern	VOID	pp_line();
extern	VOID	put_1b();
extern	VOID	put_2b();
extern	VOID	put_4b();
extern	VOID	rel_sym();
extern	VOID	skip_space();
extern	VOID	str_out();
extern	VOID	string_copy();
extern	VOID	swap_temp();
extern	BOOL	symbol();
extern	VOID	t212();
extern	VOID	t414();
extern	VOID	t800();
extern	VOID	terminate();
extern	VOID	the_end();
extern	UINT	updat_length();
extern	VOID	updat_loc();
extern	VOID	usagerr();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

extern	BOOL	atoieee();
extern	SLONG	eval();
extern	int	get_1b();
extern	int	get_2b();
extern	SLONG	get_4b();
extern	BOOL	match();
extern	BOOL	operand();
extern	BOOL	pass1();
extern	BOOL	str_size();

/*lint -fvr */
