/************************************************************************/
/*									*/
/*	File:	tasmtyp.h						*/
/*									*/
/*	Data type declarations for the Transputer Assembler (TASM).	*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1987 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

struct	opcode				/* Instruction code length entry */
	{
	STRING	name;			/* Instruction/pseudo-op name ptr */
	UCHAR	supported;		/* Which CPU's support this opcode */
	SLONG	parameter;		/* Generic "decode" parameter */
	VOID	(*decode)();		/* Pointer to opcode decoding func */
	};

#define	ALL		0x01		/* Supported by all CPU's */
#define	T212		0x02		/* Supported by T212/T222/T225 */
#define	T414		0x04		/* Supported by T414/T425 */
#define	T800		0x08		/* Supported by T800/T801/T805 */

/*
 *	The bit masks for the "supported" field in opcode table entries.
 */

struct	symtab				/* Symbol table */
	{
	UCHAR	type;			/* Symbol type */
	UCHAR	module;			/* Definition module # */
	SLONG	max_value;		/* Maximum definition location */
	SLONG	min_value;		/* Minimum definition location */
	union
		{
		STRING	name;		/* Point to ASCII symbol name */
		UINT	number;		/* Or relative symbol # */
		} id;
	};

/*
 *	The bit masks for the "type" field in symbol table entries.
 */

#define	S_DEFINED	0x01		/* Set if symbol has been defined */
#define	S_PUBLIC	0x02		/* Set if symbol is "public" */
#define	S_EXTERN	0x04		/* Set if symbol is "extern" */
#define	S_KEEP		0x08		/* Set if the linker needs this sym */
#define	S_CONSTANT	0x10		/* Set if symbol is constant value */
#define	S_VALUE		0x20		/* Set if symbol is a ".val" result */
