/************************************************************************/
/*									*/
/*	File:	tlib.c							*/
/*									*/
/*	Transputer Librarian (TLIB).					*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				12/20/89				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character type functions */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */

#if	H_OPSYS == VMS
#include	<types.h>		/* Types needed for time/date */
#include	<stat.h>		/* Structures needed for "fstat" */
#else	/* !VMS */
#if	H_OPSYS != MPW
#include	<sys/types.h>		/* Types needed for time/date */
#include	<sys/stat.h>		/* Structures needed for "fstat" */
#endif	/* !MPW && !VMS */
#endif	/* VMS */


#if	(H_OPSYS!=BSD4_2)&&(H_OPSYS!=MPW)&&(H_OPSYS!=XENIX)&&(H_OPSYS!=VMS)
#include	<sys/utime.h>		/* Structures needed for "utime" */
#endif	/* !BSD4_2 && !MPW && !XENIX && !VMS */

#if	H_OPSYS == BSD4_2
#include	<sys/file.h>
#define	SEEK_SET	L_SET
#define	SEEK_CUR	L_INCR
#define	SEEK_END	L_XTND
struct	utimbuf
	{
	time_t	actime;
	time_t	modtime;
	};
#endif	/* BSD4_2 */

#if	H_OPSYS == MPW
#include	<FCntl.h>		/* Holds F_RENAME */
#define	SEEK_SET	0
#define	SEEK_CUR	1
#define	SEEK_END	2
#endif	/* MPW */

#if	H_OPSYS == XENIX
#define	SEEK_SET	0
#define	SEEK_CUR	1
#define SEEK_END	2
#endif	/* XENIX */

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	I/O definitions.
 */

char	libbuf[FNSIZE + 1] =	{0};	/* Library filename buffer */
STREAM	lib_fp		=	NULL;	/* Library file pointer */
char	iobuf[FNSIZE + 1] =	{0};	/* I/O filename buffer */
STREAM	io_fp		=	NULL;	/* I/O file pointer */
char	tempbuf[FNSIZE + 1] =	{0};	/* Temp filename buffer */
STREAM	temp_fp		=	NULL;	/* Temp file pointer */

/*
 *	General definitions.
 */

#define	FNODE_ALLOC	100		/* # file_node's "malloc"d at once */

#if	H_OPSYS == MPW
#define	MAX_BUCKETS	2047		/* # of hash buckets for symbols */
#else	/* MPW */
#define	MAX_BUCKETS	4095		/* # of hash buckets for symbols */
#endif	/* MPW */

#define	MAX_DEF		1000		/* Max. # of symbol defs/file */
#define	MAX_FILES	1000		/* Max. # of files in a library */

/*
 *	Filename info stuff (we need to remember the names of all the files in
 *	a library to avoid duplicates).  This structure is also used during
 *	topological sorts to represent the file dependency structure.
 */

struct	file_node			/* File dependency link node */
	{
	short			fnum;	/* Dependent file number */
	struct	file_node	*link;	/* Next dependent file node */
	};

struct	ref_node			/* External reference link node */
	{
	STRING			name;	/* External symbol name */
	struct	file_node	*def_list;	/* Def. list for symbol */
	struct	file_node	*ref_list;	/* Reference list for symbol */
	struct	ref_node	*link;	/* Hash bucket forward pointer */
	};

struct	file_header			/* Filename info table */
	{
	STRING			name;	/* Pointer to filename */
	short			value;	/* Value/tag byte */
	short			next;	/* Next file # to "append" to lib */
	struct	file_node	*link;	/* Dependency node list */
	};

char	brf[] = "Bad record format on file: ";	/* Error message */
UINT	cpu_type =	TYPE_UNKNOWN;	/* Processor type not yet selected */
struct	file_header	fnames[MAX_FILES + 1] = {0};	/* File info */
char	fnbuf[FNSIZE + 1] =	{0};	/* General filename buffer */
struct	file_node	*fnode_list = NULL;	/* "file_node" free list */
UINT	fnsize		=	0;	/* Size of name in "fnbuf" */
struct	ref_node	*hash_table[MAX_BUCKETS] = {0};	/* Sym tbl */
int	last_file	=	0;	/* Most recent file # to be appended */
int	now_file	=	0;	/* File being checked */
BOOL	not_present	=	FALSE;	/* TRUE if file to update not there */
UINT	num_files	=	0;	/* # of files in library */
BOOL	verbose		=	VERBOSE;/* Default: make output verbose */

/*
 *	The following variable and definitions represent what TLIB has been
 *	instructed to do.
 */

#define	APPEND		0		/* Append files to a library */
#define	BUILD		1		/* Sort files and build library */
#define	COPY		2		/* Make copies of files in a library */
#define	DELETE		3		/* Delete file in a library */
#define	EXTRACT		4		/* Extract files from library */
#define	INSERT		5		/* Insert files into a library */
#define	LIST		6		/* List the library contents */
#define	SPLIT		7		/* "split" a library into files */
#define	UPDATE		8		/* Update files within library */

UINT	job		=	LIST;	/* Default to listing the library */

/*
 *	T_SYMBOL record stuff.
 */

UINT	num_syms	=	0;	/* # of symbols in "symbols" */
STRING	symbols[MAX_DEF] =	{0};	/* Symbol table for sorting defs */
BOOL	symdef		=	0;	/* TRUE if "symbuf" was definition */
char	symbuf[SYMNMESIZE + 1] = {0};	/* General symbol buffer */

/************************************************************************/
/*									*/
/*			Function Type Declarations			*/
/*									*/
/************************************************************************/

VOID	append_files();
VOID	banner();
VOID	build_lib();
VOID	coalesce();
VOID	copy_files();
VOID	create_lib();

#if	H_OPSYS != MPW
STRING	ctime();
#endif	/* !MPW */

VOID	delete_files();
VOID	exit();
VOID	extract_files();
VOID	fatal();
int	filecmp();
VOID	free_fnode();
VOID	free_rnode();
VOID	get_fnames();
struct	file_node	*get_fnode();
long	get_mtime();
struct	ref_node	*get_rnode();
VOID	insert_fnode();
int	inl_1b();
VOID	insert_files();
VOID	io_seek();
SLONG	io_tell();
VOID	lib_seek();
SLONG	lib_tell();
VOID	list_files();
struct	ref_node	*lookup();
VOID	main();
VOID	makefn();
VOID	order_files();
VOID	out_1b();
VOID	outl_1b();
VOID	outl_2b();
VOID	outt_1b();
VOID	outt_2b();
VOID	outt_4b();
VOID	process_defs();
VOID	qsort();
VOID	read_files();
VOID	read_io_sym();
VOID	rd_lib_fn();
VOID	rd_lib_sym();
VOID	set_mtime();
VOID	split();
VOID	tag_files();
VOID	temp_seek();
SLONG	temp_tell();
VOID	terminate();
VOID	update_files();
VOID	update_rnode();
VOID	usagerr();
VOID	warning();
VOID	visit_node();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

BOOL	add_fname();
int	in_1b();
SLONG	inl_4b();
SLONG	search();

/*lint -fvr */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Opens the library filename, deciphers any command line		*/
/*	switches, and starts doing whatever was requested.		*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
main(argc,argv)
	int			argc;
	char			*argv[];
	{
	extern	char		*strrchr();

	STRING			cos = "Conflicting options selected";
	UINT			files;
	register	STRING	ptr;
	int			targc;
	char			**targv;
/*
 *	Check validity of library filename.
 */
	if(argc == 1)
		banner();		/* Tell user who we are */
	if((argc -= 2) < 0)
		usagerr();		/* Needs at least a library file */
	ptr = *++argv;			/* Should be library file */
	if(*ptr == '-')
		usagerr();
	makefn(ptr,libbuf,TLNK_LIB_EXT);	/* Expand into full filename */
/*
 *	Process any command line switches (remember if filenames were found).
 */
	targc = argc;			/* Remember where we were */
	targv = argv;
	files = 0;			/* # of remaining filenames */
	while(targc--)			/* Process rest of command line */
		{
		ptr = *++targv;
		if(*ptr == '-')
			{
			while (*++ptr)
				{
				switch((int) *ptr)
					{
/* Append files to the library */	case 'A':
					case 'a':
						if(job != LIST)
							fatal(cos,"");
						job = APPEND;
						break;

/* Sort files and build library */	case 'B':
					case 'b':
						if(job != LIST)
							fatal(cos,"");
						job = BUILD;
						break;

/* Copy files out of the library */	case 'C':
					case 'c':
						if(job != LIST)
							fatal(cos,"");
						job = COPY;
						break;

/* Delete files in the library */	case 'D':
					case 'd':
						if(job != LIST)
							fatal(cos,"");
						job = DELETE;
						break;

/* Extract file copies */		case 'E':
					case 'e':
						if(job != LIST)
							fatal(cos,"");
						job = EXTRACT;
						break;

/* Insert files into a library */	case 'I':
					case 'i':
						if(job != LIST)
							fatal(cos,"");
						job = INSERT;
						break;

/* Split a library into files */	case 'S':
					case 's':
						if(job != LIST)
							fatal(cos,"");
						job = SPLIT;
						break;

/* Update files within a library */	case 'U':
					case 'u':
						if(job != LIST)
							fatal(cos,"");
						job = UPDATE;
						break;

/* Turn OFF verbose */			case 'V':
					case 'v':
						verbose = !verbose;
						break;

					default:
						usagerr();
					}
				}
			}
		else
			files++;
		}
/*
 *	Print sign-on messages if appropriate.
 */
	if(verbose)
		banner();
/*
 *	Verify that the various portability items are configured OK.
 */

/*lint -e506 lint isn't very found of constant valued booleans */

	if(sizeof(SLONG) < 4)
		fatal("The size of SLONG is not correctly configured","");

/*lint +e506 lint isn't very fond of constant valued booleans */

	if(files >= MAX_FILES)
		fatal("Too many files in library: ",libbuf);
/*
 *	Open the library file if needed.
 */
#if	CRLF
	if((lib_fp = (STREAM) fopen(libbuf,"rb")) == NULL)
#else
	if((lib_fp = (STREAM) fopen(libbuf,"r")) == NULL)
#endif
		{
		if(job != BUILD)
			{
			warning("Unable to open library file, creating: ",
				libbuf);
			create_lib();
#if	CRLF
			if((lib_fp = (STREAM) fopen(libbuf,"rb")) == NULL)
#else
			if((lib_fp = (STREAM) fopen(libbuf,"r")) == NULL)
#endif
				{
				fatal("Unable to open created library file: ",
					libbuf);
				}
			}
		}
	else
		{
		if(job == BUILD)
			fatal("Library file name is already in use: ",libbuf);
		}
	if(job != BUILD)
		{
		if(inl_1b() != T_LIB_FILE)
			{
			fatal("Specified library file isn't in library format",
				"");
			}
		cpu_type = inl_1b();		/* Get library CPU type */
		}
/*
 *	Open a temporary file to hold the new library if changes are going to
 *	be made.
 */
	if((job != COPY) && (job != LIST) && (job != SPLIT))
		{
		strcpy(tempbuf,libbuf);		/* Add input FN preamble */
		strcpy(strrchr(tempbuf,'.'),".1");

#if	CRLF
		if((temp_fp = (STREAM) fopen(tempbuf,"wb")) == NULL)
#else
		if((temp_fp = (STREAM) fopen(tempbuf,"w")) == NULL)
#endif
			fatal("Unable to open library temp file: ",tempbuf);
		}
/*
 *	Everything is ready to go, now is the time to do something!
 */
	switch(job)
		{
		case	APPEND:		/* Append files to library */
			if(files == 0)
				usagerr();
			append_files(argc,argv);
			break;

		case	BUILD:		/* Sort input files and make library */
			if(files == 0)
				usagerr();
			build_lib(argc,argv);
			break;

		case	COPY:		/* Copy files from the library */
			if(files == 0)
				usagerr();
			copy_files(argc,argv);
			break;

		case	DELETE:		/* Delete files from the library */
			if(files == 0)
				usagerr();
			delete_files(argc,argv);
			break;

		case	EXTRACT:	/* Extract files from library */
			if(files == 0)
				usagerr();
			extract_files(argc,argv);
			break;

		case	INSERT:		/* Insert files into library */
			if(files < 2)
				usagerr();
			insert_files(argc,argv);
			break;

		case	LIST:		/* List library contents */
			list_files();
			break;

		case	SPLIT:		/* "split" library into files */
			split();
			break;

		case	UPDATE:		/* "update" files in library */
			if(files == 0)
				usagerr();
			update_files(argc,argv);
			break;
		}
	terminate(NOERRORS);		/* Head for the barn */
	}

/************************************************************************/
/*									*/
/*	Routine:	add_fname					*/
/*									*/
/*	Add a file name to the "fnames" list after checking to see	*/
/*	that it is not already on the list (in the library).  If the	*/
/*	filename IS on the list the return value reflects that fact	*/
/*	assuming the appropriate setting for the "partial" parameter.	*/
/*									*/
/*	Parameters:							*/
/*		s -	A pointer to the filename to add.		*/
/*		partial-TRUE if the filename is allowed to be already	*/
/*			present in the library (with that fact noted in	*/
/*			the return value.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

BOOL
add_fname(s,partial)
	STRING			s;
	BOOL			partial;
	{
	extern	char		*malloc();

	register	UINT	i;
	STRING			sptr;

	for(i = 0; i < num_files; i++)
		{
		if(filecmp(fnames[i].name,s) == 0)
			{
			if(partial)
				return (TRUE);
			else
				fatal("Library already contains file: ",s);
			}
		}
	if(num_files >= MAX_FILES)
		fatal("Too many files in library: ",libbuf);
	if((sptr = (STRING) malloc((UINT) (strlen(s) + 1))) == NULL)
		fatal("Insufficient filename string memory","");
	strcpy(sptr,s);			/* Save a copy of the filename */
	fnames[num_files++].name = sptr;
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	append_files					*/
/*									*/
/*	Append files to the library.					*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
append_files(argc,argv)
	int			argc;
	char			*argv[];
	{
	int			file_num;
	SLONG			file_size;
	register	STRING	ptr;
	register	STRING	s;
	register	int	temp;
	SLONG			was_here;

	if(job == APPEND)		/* If just normal APPEND */
		{
		get_fnames();		/* Get names of existing files */
/*
 *	Make sure we know where the end of the library is.  Copy the library
 *	onto the temp file minus the T_EOF record, which we will add later.
 */
		lib_seek((SLONG) -3,SEEK_END);	/* T_EOF? */
		file_size = lib_tell();	/* # of bytes to copy from library */
		if(inl_1b() != T_EOF)
			fatal("Unable to find end of library file: ",libbuf);
		rewind(lib_fp);		/* Rewind library to beginning */
		while(file_size--)
			outt_1b((int) inl_1b());	/* Copy lib */
		}

	if(job == BUILD)
		file_num = fnames[0].next;

	while((job == BUILD) ? file_num : argc--)	/* Process files */
		{
		ptr = *++argv;
		if((job == BUILD) || (*ptr != '-'))
			{
/*
 *	Figure out input file name and make sure it is distinct within library.
 *	If so, add it to the list of files within the library.  If it isn't
 *	distinct within the library either ignore it or quit depending on
 *	whether we are updating a library or really appending to one.
 */
			if(job == BUILD)
				{
				strcpy(iobuf,fnames[file_num].name);
				file_num = fnames[file_num].next;
				}
			else
				{
				makefn(ptr,iobuf,TLNK_INPUT_EXT);
				if(add_fname(iobuf,job == UPDATE))
					continue;
				}
/*
 *	Open input file to append to library.
 */
#if	CRLF
			if((io_fp = (STREAM) fopen(iobuf,"rb")) == NULL)
#else
			if((io_fp = (STREAM) fopen(iobuf,"r")) == NULL)
#endif
				fatal("Unable to open input file: ",iobuf);
			if(verbose)
				{
			printf("Appending file '%s' to library '%s'\n",
					iobuf,libbuf);
				}
			if(in_1b() != T_REL_FILE)
				fatal(brf,iobuf);
/*
 *	Ensure that the CPU type of the input file and temp file match.
 */
			temp = in_1b();	/* Get input file CPU type */
			if(cpu_type == TYPE_UNKNOWN)
				{
				cpu_type = temp;	/* New type */
				was_here = temp_tell();
				temp_seek((SLONG) 1,SEEK_SET);
				outt_1b((int) temp);	/* Set lib CPU type */
				temp_seek((SLONG) was_here, SEEK_SET);
				}
			else
				{
				if(temp != TYPE_UNKNOWN)
					{
					if(temp != cpu_type)
						{
			fatal("Unable to mix code for different CPU types","");
						}
					}
				}
/*
 *	Write the T_FILENAME header record for the new file.
 */
			outt_1b((int) T_FILENAME);
			outt_4b((SLONG) 0);	/* Dummy old&new line #'s */
			outt_1b((int) strlen(iobuf));	/* Filename length */
			s = iobuf;
			while(*s)
				outt_1b((int) *s++);
/*
 *	Read the T_SYMBOL records from the input file, note which are public,
 *	sort those and put in the library file after the T_FILENAME record.
 */
			process_defs();
/*
 *	Figure out how big input file is.
 */
			io_seek((SLONG) -3,SEEK_END);	/* Where T_EOF is? */
			if(in_1b() != T_EOF)
				{
				fatal("Unable to find end of input file: ",
					iobuf);
				}
			file_size = io_tell() + (SLONG) 2;	/* File size */
			outt_1b((int) T_SIZE);
			outt_4b((SLONG) get_mtime(io_fp,iobuf));
			outt_4b((SLONG) file_size);
/*
 *	Read input file into library file.
 */
			rewind(io_fp);	/* Reset input file */
			while(file_size--)	/* Append file */
				outt_1b((int) in_1b());
/*
 *	Close input file and go on.
 */
			if(fclose(io_fp) == EOF)
				fatal("Unable to close input file: ",iobuf);
			}
		}
	outt_1b((int) T_EOF);		/* Add end of library marker */
	outt_2b((int) 0);		/* Dummy line # */
	}

/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display sign-on message to user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
banner()
	{
	printf("%s, %s, %s\n\n",LIB_NAME,VERSION,COPYRIGHT);
	}

/************************************************************************/
/*									*/
/*	Routine:	build_lib					*/
/*									*/
/*	Build a library by doing a topological sort of the input files	*/
/*	to ensure that file dependencies are satisfied (and thus one	*/
/*	pass over the library will be sufficient).  If a dependency	*/
/*	cycle exists let the user know.					*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
build_lib(argc,argv)
	int			argc;
	char			*argv[];
	{
/*
 *	Read all the input files and gather the symbol/file dependency
 *	information.
 */
	read_files(argc,argv);
/*
 *	Convert symbol dependency graph to a file dependency graph with edges
 *	pointing backwards (ie. the file containing a symbol definition is set
 *	to depend on the files which reference it).
 */
	coalesce();
/*
 *	Do a reverse topological sort of the file dependency information to
 *	produce a ordering suitable for a one pass linker pass (or let the user
 *	know that a "dependency cycle" exists.
 */
	for(now_file = 1; now_file <= num_files; now_file++)
		{
		if(fnames[now_file].value == 0)
			visit_node(now_file);
		}
/*
 *	Looks OK, now write initial record onto temp file.
 */
	outt_1b((int) T_LIB_FILE);
	outt_1b((int) TYPE_UNKNOWN);
/*
 *	We have the correct file ordering, now append the files to the library
 *	in the correct order.
 */
	append_files(argc,argv);
	}

/************************************************************************/
/*									*/
/*	Routine:	coalesce					*/
/*									*/
/*	Convert from symbol dependency information (stored in		*/
/*	"hash_table"), to file dependency information (stored in	*/
/*	"fnames").  Note that we are actually building the dependency	*/
/*	graph in reverse since we want to have the file which contains	*/
/*	the definition of the symbol to depend on the file which has	*/
/*	the reference of the symbol.  This is so since we want to later	*/
/*	do a reverse topological sort (easier than doing a normal order	*/
/*	one).								*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
coalesce()
	{
	register struct	file_node *def_ptr;
	int			i;
	register struct	ref_node *ptr;
	register struct	file_node *ref_ptr;

	for(i = 0; i < MAX_BUCKETS; i++)
		{
		while((ptr = hash_table[i]) != NULL)
			{
			while((def_ptr = ptr->def_list) != NULL)
				{
				for(ref_ptr = ptr->ref_list; ref_ptr != NULL;
					ref_ptr = ref_ptr->link)
					{
					insert_fnode(def_ptr->fnum,
						ref_ptr->fnum);
					}
				ptr->def_list = def_ptr->link;
				free_fnode(def_ptr);
				}
			while((ref_ptr = ptr->ref_list) != NULL)
				{
				ptr->ref_list = ref_ptr->link;
				free_fnode(ref_ptr);
				}
			hash_table[i] = ptr->link;
			free_rnode(ptr);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	copy_files					*/
/*									*/
/*	Make individual copies of files from the library.		*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy_files(argc,argv)
	int			argc;
	char			*argv[];
	{
	SLONG			file_size;
	register	UINT	i;
	SLONG			start_point;
	register	int	temp;
	long			time;

	tag_files(argc,argv,FALSE);	/* Figure which files will be done */
	rewind(lib_fp);			/* Rewind library file */
	start_point = 2;		/* Point at first T_FILENAME */
	for(i = 0; i < num_files; i++)
		{
		if(fnames[i].value)
			{
			search((SLONG) start_point,fnames[i].name);
			if(inl_1b() != T_FILENAME)
				fatal(brf,libbuf);
			rd_lib_fn();	/* Get output filename */
			strcpy(iobuf,fnbuf);	/* Output file name to use */
/*
 *	Open prospective output filename.
 */
#if	CRLF
			if((io_fp = (STREAM) fopen(iobuf,"wb")) == NULL)
#else
			if((io_fp = (STREAM) fopen(iobuf,"w")) == NULL)
#endif
				fatal("Unable to open output file: ",iobuf);
			if(verbose)
				{
		printf("Creating a copy of file '%s' from library '%s'\n",
					iobuf,libbuf);
				}
/*
 *	Strip off T_SYMBOL definition header records.
 */
			while((temp = inl_1b()) == T_SYMBOL)
				rd_lib_sym();
/*
 *	Get size for new file.
 */
			if(temp != T_SIZE)
				fatal(brf,libbuf);
			time = inl_4b();	/* Get modification time */
			file_size = inl_4b();
/*
 *	Write file out.
 */
			while(file_size--)
				out_1b((int) inl_1b());
/*
 *	Close output file and go on.
 */
			start_point = lib_tell();
			if(fclose(io_fp) == EOF)
				fatal("Unable to close output file: ",iobuf);
			set_mtime(iobuf,time);	/* Set modification time */
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	create_lib					*/
/*									*/
/*	Create a skeleton (empty) library.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
create_lib()
	{
#if	CRLF
	if((lib_fp = (STREAM) fopen(libbuf,"wb")) == NULL)
#else
	if((lib_fp = (STREAM) fopen(libbuf,"w")) == NULL)
#endif
		fatal("Unable to open/create library file: ",libbuf);
/*
 *	Put the two record overhead in for the library format.
 */
	outl_1b((int) T_LIB_FILE);
	outl_1b((int) TYPE_UNKNOWN);
	outl_1b((int) T_EOF);
	outl_2b((int) 0);
/*
 *	Close the file under the current R/W attributes and return.
 */
	if(fclose(lib_fp) == EOF)
		fatal("Unable to close library file: ",libbuf);
	lib_fp = NULL;
	}

/************************************************************************/
/*									*/
/*	Routine:	delete_files					*/
/*									*/
/*	Remove files from the library.					*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
delete_files(argc,argv)
	int			argc;
	char			*argv[];
	{
	SLONG			file_size;
	register	UINT	i;
	SLONG			start_point;
	register	int	temp;

	tag_files(argc,argv,FALSE);	/* Figure which files will be done */
	rewind(lib_fp);			/* Rewind library file */
	outt_1b((int) inl_1b());	/* Copy T_LIB record */
	outt_1b((int) inl_1b());
	start_point = lib_tell();	/* Point to first T_FILENAME record */
	for(i = 0; i < num_files; i++)
		{
		if(fnames[i].value)
			{
			file_size = search(start_point,fnames[i].name) -
				start_point;
			lib_seek((SLONG) start_point,SEEK_SET);
/* Copy up to here */	while(file_size--)
				outt_1b((int) inl_1b());
			if(inl_1b() != T_FILENAME)
				fatal(brf,libbuf);
/* Skip filename */	rd_lib_fn();
			if(verbose)
				{
			printf("Removing file '%s' from library '%s'\n",
					iobuf,libbuf);
				}
/* Skip T_SYMBOL's */	while((temp = inl_1b()) == T_SYMBOL)
				rd_lib_sym();
/* Actual file size */	if(temp != T_SIZE)
				fatal(brf,libbuf);
			inl_4b();	/* Toss modification time/date */
/* Skip file */		lib_seek((SLONG) inl_4b(),SEEK_CUR);
			start_point = lib_tell();
			}
		}
/*
 *	Copy rest of library file to temp file.
 */
	lib_seek((SLONG) -3,SEEK_END);	/* T_EOF? */
	if(inl_1b() != T_EOF)
		fatal("Unable to find end of library file: ",libbuf);
	file_size = lib_tell() - start_point + 2;
	lib_seek((SLONG) start_point,SEEK_SET);
	while(file_size--)
		outt_1b((int) inl_1b());
	}

/************************************************************************/
/*									*/
/*	Routine:	extract_files					*/
/*									*/
/*	Remove files from the library and turn them back into		*/
/*	individual files.						*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
extract_files(argc,argv)
	int			argc;
	char			*argv[];
	{
	SLONG			file_size;
	register	UINT	i;
	SLONG			start_point;
	register	int	temp;
	long			time;

	tag_files(argc,argv,FALSE);	/* Figure which files will be done */
	rewind(lib_fp);			/* Rewind library file */
	outt_1b((int) inl_1b());	/* Copy T_LIB record */
	outt_1b((int) inl_1b());
	start_point = lib_tell();	/* Point to first T_FILENAME record */
	for(i = 0; i < num_files; i++)
		{
		if(fnames[i].value)
			{
			file_size = search(start_point,fnames[i].name) -
				start_point;
			lib_seek((SLONG) start_point,SEEK_SET);
			while(file_size--)
				outt_1b((int) inl_1b());
			if(inl_1b() != T_FILENAME)
				fatal(brf,libbuf);
			rd_lib_fn();	/* Get output filename */
			strcpy(iobuf,fnbuf);	/* Output file name to use */
/*
 *	Open prospective output filename.
 */
#if	CRLF
			if((io_fp = (STREAM) fopen(iobuf,"wb")) == NULL)
#else
			if((io_fp = (STREAM) fopen(iobuf,"w")) == NULL)
#endif
				fatal("Unable to open output file: ",iobuf);
			if(verbose)
				{
			printf("Extracting file '%s' from library '%s'\n",
					iobuf,libbuf);
				}
/*
 *	Strip off T_SYMBOL definition header records.
 */
			while((temp = inl_1b()) == T_SYMBOL)
				rd_lib_sym();
/*
 *	Get size for new file.
 */
			if(temp != T_SIZE)
				fatal(brf,libbuf);
			time = inl_4b();	/* Get modification time */
			file_size = inl_4b();
/*
 *	Write file out.
 */
			while(file_size--)
				out_1b((int) inl_1b());
/*
 *	Close output file and go on.
 */
			start_point = lib_tell();
			if(fclose(io_fp) == EOF)
				fatal("Unable to close output file: ",iobuf);
			set_mtime(iobuf,time);	/* Set modification time */
			}
		}
/*
 *	Copy rest of library file to temp file.
 */
	lib_seek((SLONG) -3,SEEK_END);	/* T_EOF? */
	if(inl_1b() != T_EOF)
		fatal("Unable to find end of library file: ",libbuf);
	file_size = lib_tell() - start_point + 2;
	lib_seek((SLONG) start_point,SEEK_SET);
	while(file_size--)
		outt_1b((int) inl_1b());
	}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		sptr1 -	A pointer to the first of the error message.	*/
/*		sptr2 -	A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	printf("FATAL: %s%s\n",sptr1,sptr2);	/* Print message */
	terminate(ERRORS);		/* Bye... */
	}

/************************************************************************/
/*									*/
/*	Routine:	filecmp						*/
/*									*/
/*	Compares two filenames using the case convention which has been	*/
/*	configured.							*/
/*									*/
/*	Parameters:							*/
/*		ptr1 -	A pointer to the first filename.		*/
/*		ptr2 -	A pointer to the second filename.		*/
/*									*/
/*	Returns:	The result of the comparison ala "strcmp".	*/
/*									*/
/************************************************************************/

int
filecmp(ptr1,ptr2)
	register	STRING	ptr1;
	register	STRING	ptr2;
	{
#if	!NOFILECASE
	return (strcmp(ptr1,ptr2));
#else
	while((*ptr1 != '\0') && (toupper(*ptr1) == toupper(*ptr2)))
		{
		ptr1++;
		ptr2++;
		}
	if((*ptr1 == '\0') && (*ptr2 == '\0'))
		return (0);
	else if(*ptr1 == '\0')
		return (-1);
	else if(*ptr2 == '\0')
		return (1);
	else
		return(toupper(*ptr1) - toupper(*ptr2));
#endif	/* !NOFILECASE */
	}

/************************************************************************/
/*									*/
/*	Routine:	free_fnode					*/
/*									*/
/*	Return a "file_node" structure to the "fnode_list" (free list).	*/
/*									*/
/*	Parameters:							*/
/*		ptr -	A pointer to the node to return.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
free_fnode(ptr)
	struct	file_node	*ptr;
	{
	ptr->link = fnode_list;
	fnode_list = ptr;
	}

/************************************************************************/
/*									*/
/*	Routine:	free_rnode					*/
/*									*/
/*	Return a "ref_node" structure to the "rnode_list" (free list).	*/
/*									*/
/*	Parameters:							*/
/*		ptr -	A pointer to the node to return.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
free_rnode(ptr)
	struct	ref_node	*ptr;
	{
	if(ptr->name != NULL)
		free(ptr->name);	/* Free name */
	free((STRING) ptr);		/* Free reference node */
	}

/************************************************************************/
/*									*/
/*	Routine:	get_fnames					*/
/*									*/
/*	Read the library file and record the names of all the files	*/
/*	it contains.  This is needed to prevent duplicate files from	*/
/*	being added.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
get_fnames()
	{
	extern	char		*malloc();

	register	STRING	s;
	register	int	temp;

	num_files = 0;			/* Keep track of how many we have */
	while((temp = inl_1b()) == T_FILENAME)
		{
		rd_lib_fn();
		if(num_files >= MAX_FILES)
			fatal("Too many files in library: ",libbuf);
		if((s = (STRING) malloc((UINT) (fnsize + 1))) == NULL)
			fatal("Insufficient filename string memory","");
		strcpy(s,fnbuf);	/* Save a copy of the filename */
		fnames[num_files++].name = s;
		while((temp = inl_1b()) == T_SYMBOL)
			rd_lib_sym();
		if(temp != T_SIZE)
			fatal(brf,libbuf);
		inl_4b();		/* Toss modification time/date */
		lib_seek((SLONG) inl_4b(),SEEK_CUR);
		}
	if(temp != T_EOF)
		fatal(brf,libbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_fnode					*/
/*									*/
/*	Get a "file_node" data structure via "malloc" (note that they	*/
/*	are allocated by block to minimize pointer overhead).		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	A pointer to the "file_node" allocated.		*/
/*									*/
/************************************************************************/

struct	file_node	*
get_fnode()
	{
	extern	char		*malloc();

	int			i;
	struct	file_node	*rval;

	if(fnode_list == NULL)		/* See if any on the free list */
		{
		if((rval = (struct file_node *)
			malloc(FNODE_ALLOC * sizeof(struct file_node))) ==
				NULL)
			{
			fatal("Insufficient file dependency node memory","");
			}
		for(i = 0; i < FNODE_ALLOC; i++)
			{
			rval->link = fnode_list;
			fnode_list = rval++;
			}
		}
	rval = fnode_list;
	fnode_list = rval->link;
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_mtime					*/
/*									*/
/*	Get last modification time/date for specified stream.		*/
/*									*/
/*	Parameters:							*/
/*		fp -	File pointer.					*/
/*		fname -	Filename in case of errors.			*/
/*									*/
/*	Returns:	The modification date in the form of a long	*/
/*			holding the # of seconds since 00:00:00,	*/
/*			January 1,1970 (GMT).				*/
/*									*/
/************************************************************************/

long
get_mtime(fp,fname)
	STREAM			fp;
	STRING			fname;
	{
#if	H_OPSYS == MPW
	return (0L);
#else	/* MPW */
	struct	stat		stat_buffer;

	if(fstat(fileno(fp),&stat_buffer) != 0)
		fatal("Unable to get modification time for file: ",fname);
	return (stat_buffer.st_mtime);
#endif	/* MPW */
	}

/************************************************************************/
/*									*/
/*	Routine:	get_rnode					*/
/*									*/
/*	Get a "ref_node" data structure via "malloc".			*/
/*									*/
/*	Parameters:							*/
/*		ptr -	A pointer to the node to return.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

struct	ref_node	*
get_rnode()
	{
	extern	char		*malloc();

	struct	ref_node	*rval;

	if((rval = (struct ref_node *) malloc(sizeof(struct ref_node))) ==
		NULL)
		{
		fatal("Insufficient symbol reference/definition memory","");
		}
	rval->name = NULL;
	rval->def_list = rval->ref_list = NULL;
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	in_1b						*/
/*									*/
/*	Read a byte from the current I/O file.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The byte value read.				*/
/*									*/
/************************************************************************/

int
in_1b()
	{
	int			c;

	if((c = fgetc(io_fp)) == EOF)
		{
		if(ferror(io_fp) || feof(io_fp))
			fatal("Error reading input file: ",iobuf);
		}
	return (c & 0xFF);
	}

/************************************************************************/
/*									*/
/*	Routine:	inl_1b,inl_4b					*/
/*									*/
/*	Read 1 or 4 bytes from the library file.			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The 1 or 4 byte value read.			*/
/*									*/
/************************************************************************/

int
inl_1b()
	{
	int			c;

	if((c = fgetc(lib_fp)) == EOF)
		{
		if(ferror(lib_fp) || feof(lib_fp))
			fatal("Error reading library file: ",libbuf);
		}
	return (c & 0xFF);
	}

SLONG
inl_4b()
	{
	SLONG			rval;

	rval = (SLONG) inl_1b();
	rval |= (((SLONG) inl_1b()) << 8);
	rval |= (((SLONG) inl_1b()) << 16);
	rval |= (((SLONG) inl_1b()) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	insert_files					*/
/*									*/
/*	Insert files into a library.					*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
insert_files(argc,argv)
	int			argc;
	char			*argv[];
	{
	SLONG			file_size;
	BOOL			insert_file;
	char			insert_name[FNSIZE + 1];
	SLONG			insert_point;
	register	STRING	s;
	register	int	temp;
	SLONG			was_here;

	get_fnames();			/* Get names of existing files */

	insert_file = TRUE;		/* FALSE if not insert filename */
	while(argc--)			/* Process filenames to insert */
		{
		s = *++argv;
		if(*s != '-')
			{
/*
 *	If this is the first filename encountered, it specifies the insertion
 *	point.  Skip forward in the library until we find this point.
 */
			if(insert_file)	/* If filename to insert before */
				{
				insert_file = FALSE;
				makefn(s,insert_name,TLNK_INPUT_EXT);
				insert_point = search((SLONG) 2,insert_name);
/*
 *	Weve found the point to insert at, copy the library up to this point
 *	to the temp file.
 */
				rewind(lib_fp);
				file_size = insert_point;
				while(file_size--)
					outt_1b((int) inl_1b());
				continue;
				}
/*
 *	Since this isn't the first filename, it must be the name of a file to
 *	insert at the current point.
 */
			makefn(s,iobuf,TLNK_INPUT_EXT);
			(VOID) add_fname(iobuf,FALSE);	/* Add name to list */
#if	CRLF
			if((io_fp = (STREAM) fopen(iobuf,"rb")) == NULL)
#else
			if((io_fp = (STREAM) fopen(iobuf,"r")) == NULL)
#endif
				fatal("Unable to open input file: ",iobuf);
			if(verbose)
				{
			printf("Inserting file '%s' into library '%s'\n",
					iobuf,libbuf);
				}
			if(in_1b() != T_REL_FILE)
				fatal(brf,iobuf);
/*
 *	Ensure that the CPU type of the input file and the temp file match.
 */
			temp = in_1b();	/* Get input file CPU type */
			if(cpu_type == TYPE_UNKNOWN)
				{
				cpu_type = temp;	/* New type */
				was_here = temp_tell();
				temp_seek((SLONG) 1,SEEK_SET);
				outt_1b((int) temp);	/* Set lib CPU type */
				temp_seek((SLONG) was_here, SEEK_SET);
				}
			else
				{
				if(temp != TYPE_UNKNOWN)
					{
					if(temp != cpu_type)
						{
			fatal("Unable to mix code for different CPU types","");
						}
					}
				}
/*
 *	Put T_FILENAME record for input file into temp file.
 */
			outt_1b((int) T_FILENAME);
			outt_4b((SLONG) 0);	/* Dummy old & new line #'s */
			outt_1b((int) strlen(iobuf));	/* Length */
			s = iobuf;
			while(*s)
				outt_1b((int) *s++);
/*
 *	Read T_SYMBOL records from the input file, note which are public,
 *	sort those and put in the library temp file.
 */
			process_defs();
/*
 *	Figure out how big input file is.
 */
			io_seek((SLONG) -3,SEEK_END);	/* Where T_EOF is? */
			if(in_1b() != T_EOF)
				{
				fatal("Unable to find end of input file: ",
					iobuf);
				}
			file_size = io_tell() + (SLONG) 2;
			outt_1b((int) T_SIZE);
			outt_4b((SLONG) get_mtime(io_fp,iobuf));
			outt_4b((SLONG) file_size);
/*
 *	Read input file into library file.
 */
			rewind(io_fp);	/* Reset input file */
			while(file_size--)
				outt_1b((int) in_1b());
/*
 *	Close input file and go on.
 */
			if(fclose(io_fp) == EOF)
				fatal("Unable to close input file: ",iobuf);
			}
		}
/*
 *	Append rest of original library file to library temp file.
 */
	lib_seek((SLONG) -3,SEEK_END);	/* T_EOF? */
	if(inl_1b() != T_EOF)
		fatal("Unable to find end of library file: ",libbuf);
	file_size = lib_tell() - insert_point + 2;	/* # of bytes left */
	lib_seek((SLONG) insert_point,SEEK_SET);
	while(file_size--)
		outt_1b((int) inl_1b());	/* Copy rest of lib */
	}

/************************************************************************/
/*									*/
/*	Routine:	insert_fnode					*/
/*									*/
/*	Called by "coalesce" to add a element to the file dependency	*/
/*	graph.								*/
/*									*/
/*	Parameters:							*/
/*		def -	The file number which defined the symbol.	*/
/*		ref -	The file number which referenced the symbol.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
insert_fnode(def,ref)
	short			def;
	short			ref;
	{
	struct	file_node	*ptr;

	for(ptr = fnames[def].link; ptr != NULL; ptr = ptr->link)
		{
		if(ptr->fnum == ref)
			break;
		}
	if(ptr == NULL)
		{
		ptr = get_fnode();
		ptr->fnum = ref;
		ptr->link = fnames[def].link;
		fnames[def].link = ptr;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	io_seek						*/
/*									*/
/*	Seek the input file pointer.					*/
/*									*/
/*	Parameters:							*/
/*		distance-A SLONG representing the distance to seek.	*/
/*		origin-	A int which specifies where the seek begins.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
io_seek(distance,origin)
	SLONG			distance;
	int			origin;
	{
/*
 *	The following "rewind" should be unnecessary, but a bug in Microsoft's
 *	"fseek" for the compact memory model (V4.0) seems to require it.  The
 *	bug seems to happen on "fseek" operations which move towards the
 *	beginning of the file from the current position, but hasn't been nailed
 *	down completely...
 */
	if(origin == SEEK_SET)
		rewind(io_fp);
	if(fseek(io_fp,distance,origin) != 0)
		fatal("Error seeking input file: ",iobuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	io_tell						*/
/*									*/
/*	Return the current position of the input file pointer.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

SLONG
io_tell()
	{
	SLONG			rval;

	if((rval = ftell(io_fp)) == (SLONG) -1)
		fatal("Error determining where input file pointer is: ",iobuf);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	lib_seek					*/
/*									*/
/*	Seek the library file pointer.					*/
/*									*/
/*	Parameters:							*/
/*		distance-A SLONG representing the distance to seek.	*/
/*		origin-	A int which specifies where the seek begins.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
lib_seek(distance,origin)
	SLONG			distance;
	int			origin;
	{
/*
 *	The following "rewind" should be unnecessary, but a bug in Microsoft's
 *	"fseek" for the compact memory model (V4.0) seems to require it.  The
 *	bug seems to happen on "fseek" operations which move towards the
 *	beginning of the file from the current position, but hasn't been nailed
 *	down completely...
 */
	if(origin == SEEK_SET)
		rewind(lib_fp);
	if(fseek(lib_fp,distance,origin) != 0)
		fatal("Error seeking library file: ",libbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	lib_tell					*/
/*									*/
/*	Return the current position of the library file pointer.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

SLONG
lib_tell()
	{
	SLONG			rval;

	if((rval = ftell(lib_fp)) == (SLONG) -1)
		{
		fatal("Error determining where library file pointer is: ",
			iobuf);
		}
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	list_files					*/
/*									*/
/*	Explicitly list the files in a library.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
list_files()
	{
	SLONG			big_temp;
	UINT			lib_count;
	UINT			sym_count;
	register	int	temp;
	long			time;

	sym_count = lib_count = 0;

	printf("Files within library '%s':\n",libbuf);
	if(! verbose)
		putchar('\n');

	while((temp = inl_1b()) == T_FILENAME)
		{
		lib_count++;
		rd_lib_fn();		/* Read T_FILENAME record */
		if(verbose)
			{
	printf("\nFile '%s' contains the following symbol definitions:\n",
			fnbuf);
			}
		while((temp = inl_1b()) == T_SYMBOL)
			{
			sym_count++;
			rd_lib_sym();
			if(verbose)
				printf("\t%s\n",symbuf);
			}
		if(temp != T_SIZE)
			fatal(brf,libbuf);
		time = inl_4b();
		big_temp = inl_4b();
#if	H_OPSYS == MPW
		printf("File '%s', size in bytes: %ld\n",fnbuf,big_temp);
#else	/* MPW */
		printf("File '%s', size in bytes: %ld, last modified: %s",
			fnbuf,big_temp,ctime(&time));
#endif	/* MPW */
		lib_seek((SLONG) big_temp, SEEK_CUR);
		}
	if(temp != T_EOF)
		fatal(brf,libbuf);

	if(lib_count)
		{
printf("\n%u file(s), containing %u symbol definition(s), in library '%s'\n",
			lib_count,sym_count,libbuf);
		}
	else
		printf("\nNo files in library\n");
	}

/************************************************************************/
/*									*/
/*	Routine:	lookup						*/
/*									*/
/*	Check if a symbol is in the hashed symbol table, if not add it.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	A pointer to the existing/resulting "ref_node"	*/
/*			in a hash table bucket chain.			*/
/*									*/
/************************************************************************/

struct	ref_node	*
lookup()
	{
	extern	char		*malloc();

	register	UINT	num;
	register struct ref_node *ptr;
	register	STRING	tptr;

	tptr = symbuf;			/* Point to symbol */
	num = 0;			/* Zero hash sum */
	while(*tptr)
		num = (num * 16) + *tptr++;	/* Compute hash */
	num %= MAX_BUCKETS;		/* Mask to size of hash table */
/*
 *	We now have the hash table bucket #, scan down the bucket link until
 *	we either find the symbol or run out of "ref_node"s.
 */
	for(ptr = hash_table[num]; ptr != NULL; ptr = ptr->link)
		{
		if(strcmp(symbuf,ptr->name) == 0)
			return (ptr);
		}
/*
 *	Well, no symbol match.  Now we get a virgin "ref_node" and add it to
 *	the bucket link list, filling in the symbol name.
 */
	ptr = get_rnode();
	ptr->link = hash_table[num];
	hash_table[num] = ptr;
	if((tptr = (STRING) malloc((UINT) (strlen(symbuf) + 1))) == NULL)
		fatal("Insufficient symbol string memory","");
	strcpy(tptr,symbuf);
	ptr->name = tptr;
	return (ptr);
	}

/************************************************************************/
/*									*/
/*	Routine:	makefn						*/
/*									*/
/*	Expand a skeleton filename by adding a default extension if	*/
/*	none is supplied.						*/
/*									*/
/*	Parameters:							*/
/*		sptr -	A pointer to the skeleton filename.		*/
/*		bptr -	A pointer to the buffer the result goes in.	*/
/*		ext -	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
makefn(sptr,bptr,ext)
	STRING			sptr;
	STRING			bptr;
	STRING			ext;
	{
	extern	char		*strchr();
	extern	char		*strrchr();

	STRING			s;

	if(strlen(sptr) > FNSIZE)
		usagerr();
	strcpy(bptr,sptr);
	if((s = strrchr(bptr,'.')) != NULL)
		{
		if((strchr(s,'/') == NULL) && (strchr(s,'\\') == NULL))
			return;
		}
	if(strlen(bptr) > (FNSIZE - 4))
		usagerr();
	strcat(bptr,ext);
	}

/************************************************************************/
/*									*/
/*	Routine:	out_1b						*/
/*									*/
/*	Write a byte to the I/O file.					*/
/*									*/
/*	Parameters:							*/
/*		value -	The byte value to write.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
out_1b(value)
	int			value;
	{
	if(putc(value,io_fp) == EOF)
		{
		if(ferror(io_fp))
			fatal("Error writing output file: ",iobuf);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	outl_1b,outl_2b					*/
/*									*/
/*	Write 1 or 2 bytes to the "old" library file.  These functions	*/
/*	are only used when a skeleton library file is being built by	*/
/*	"create_lib".							*/
/*									*/
/*	Parameters:							*/
/*		value -	The 1 or 2 byte value to write.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
outl_1b(value)
	int			value;
	{
	if(putc(value,lib_fp) == EOF)
		{
		if(ferror(lib_fp))
			fatal("Error writing library file: ",libbuf);
		}
	}

VOID
outl_2b(value)
	int			value;
	{
	outl_1b(value);
	outl_1b(value >> 8);
	}

/************************************************************************/
/*									*/
/*	Routine:	outt_1b,outt_2b,outt_4b				*/
/*									*/
/*	Write 1,2 or 4 bytes to the library temp file.			*/
/*									*/
/*	Parameters:							*/
/*		value -	The 1,2 or 4 byte value to write.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
outt_1b(value)
	int			value;
	{
	if(putc(value,temp_fp) == EOF)
		{
		if(ferror(temp_fp))
			fatal("Error writing temp file: ",tempbuf);
		}
	}

VOID
outt_2b(value)
	int			value;
	{
	outt_1b(value);
	outt_1b(value >> 8);
	}

VOID
outt_4b(value)
	SLONG			value;
	{
	outt_1b((int) value);
	outt_1b((int) (value >>= 8));
	outt_1b((int) (value >>= 8));
	outt_1b((int) (value >> 8));
	}

/************************************************************************/
/*									*/
/*	Routine:	process_defs					*/
/*									*/
/*	Reads the T_SYMBOL records from the current input file, records	*/
/*	those which are public, and writes them to the temp file in	*/
/*	sorted order.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

/*
 *	First a helping function for the sort routine call.
 */

static	int
symcmp(first,second)
	STRING			*first;
	STRING			*second;
	{
	return (strcmp(*first,*second));
	}

VOID
process_defs()
	{
	extern	char		*malloc();

	register	STRING	s;
	register	UINT	i;

	num_syms = 0;			/* # of defined symbols */
	while(in_1b() == T_SYMBOL)
		{
		read_io_sym();
		if(symdef)
			{
			if(num_syms >= MAX_DEF)
				{
				fatal("Too many defined symbols in file: ",
					iobuf);
				}
			if((s = (STRING)
				malloc((UINT) (strlen(symbuf) + 1))) == NULL)
				{
				fatal("Insufficient symbol string memory","");
				}
			strcpy(s,symbuf);
			symbols[num_syms++] = s;
			}
		}
	if(num_syms)
		{
		qsort((STRING) symbols,num_syms,(UINT) sizeof(STRING),symcmp);
		for(i = 0; i < num_syms; i++)
			{
			outt_1b((int) T_SYMBOL);
			outt_1b((int) TRUE);	/* All defined I hope! */
			outt_1b((int) strlen(symbols[i]));
			for(s = symbols[i]; *s; )
				outt_1b((int) *s++);
			free(symbols[i]);	/* Return free storage */
			}
		}
	else
		warning("No external symbols defined in file: ",iobuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	read_io_sym					*/
/*									*/
/*	Read T_SYMBOL record in input file into "symbuf" (everything	*/
/*	but the T_SYMBOL header byte which has already been skipped).	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
read_io_sym()
	{
	register	UINT	temp;
	register	UINT	length;

	symdef = in_1b();		/* Read symbol type */
	length = in_1b();		/* Get symbol length */
	for(temp = 0; temp < length; temp++)
		symbuf[temp] = in_1b();
	symbuf[temp] = '\0';
	}

/************************************************************************/
/*									*/
/*	Routine:	read_files					*/
/*									*/
/*	Read the library input files and build a symbol dependency	*/
/*	graph.								*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
read_files(argc,argv)
	int			argc;
	char			*argv[];
	{
	extern	char		*malloc();

	BOOL			defined;
	register	int	length;
	register	STRING	ptr;
	register	STRING	s;
	register	int	temp;

	while(argc--)			/* Process filenames for library */
		{
		ptr = *++argv;
		if(*ptr != '-')
			{
			makefn(ptr,iobuf,TLNK_INPUT_EXT);
			if((s = (STRING)
				malloc((UINT) (strlen(iobuf) + 1))) == NULL)
				{
				fatal("Insufficient filename string memory",
					"");
				}
			strcpy(s,iobuf);
			fnames[++num_files].name = s;
/*
 *	Open input file to put in library.
 */
#if	CRLF
			if((io_fp = (STREAM) fopen(iobuf,"rb")) == NULL)
#else
			if((io_fp = (STREAM) fopen(iobuf,"r")) == NULL)
#endif
				fatal("Unable to open input file: ",iobuf);
			if(in_1b() != T_REL_FILE)
				fatal(brf,iobuf);
			(VOID) in_1b();	/* Skip CPU type field */
/*
 *	Read all T_SYMBOL definitions from the file and add to the hash table
 *	as appropriate for defined or referenced symbols.
 */
			while(in_1b() == T_SYMBOL)
				{
				defined = in_1b();	/* Definition? */
				length = in_1b();	/* Get symbol length */
				for(temp = 0; temp < length; temp++)
					symbuf[temp] = in_1b();
				symbuf[temp] = '\0';
				update_rnode(defined);
				}
/*
 *	Close input file and go on.
 */
			if(fclose(io_fp) == EOF)
				fatal("Unable to close input file: ",iobuf);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	rd_lib_fn					*/
/*									*/
/*	Read a T_FILENAME record in the library file and store the	*/
/*	filename in the "fnbuf" buffer.  Note that the T_FILENAME	*/
/*	header byte has been already skipped.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
rd_lib_fn()
	{
	register	UINT	temp;

	inl_4b();			/* Toss old & new line #'s */
	fnsize = inl_1b();		/* Get filename length */
	for(temp = 0; temp < fnsize; temp++)
		fnbuf[temp] = inl_1b();
	fnbuf[temp] = '\0';
	}

/************************************************************************/
/*									*/
/*	Routine:	rd_lib_sym					*/
/*									*/
/*	Read T_SYMBOL record in library into "symbuf" (everything but	*/
/*	the T_SYMBOL header byte which has already been skipped).	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
rd_lib_sym()
	{
	register	UINT	length;
	register	UINT	temp;

	symdef = inl_1b();		/* Read symbol type */
	length = inl_1b();		/* Get symbol length */
	for(temp = 0; temp < length; temp++)
		symbuf[temp] = inl_1b();
	symbuf[temp] = '\0';
	}

/************************************************************************/
/*									*/
/*	Routine:	rename						*/
/*									*/
/*	Simulate "rename" library routine missing under XENIX.		*/
/*									*/
/*	Parameters:							*/
/*		from -	Old name pointer.				*/
/*		to -	New name pointer.				*/
/*									*/
/*	Returns:	Non-zero if error detected.			*/
/*									*/
/************************************************************************/

#if	H_OPSYS == XENIX
int
rename(from,to)
	STRING			from;
	STRING			to;
	{
	char			sysstr[512];

	sprintf(sysstr,"mv %s %s",from,to);
	return (system(sysstr));
	}
#endif	/* XENIX */

/************************************************************************/
/*									*/
/*	Routine:	search						*/
/*									*/
/*	Search the library for the specified file, return a pointer to	*/
/*	where it starts in the library.  Leave the library file pointer	*/
/*	at the spot where the file starts.				*/
/*									*/
/*	Parameters:							*/
/*		position-The starting position in the library file.	*/
/*		s -	A pointer to the filename to look for.		*/
/*									*/
/*	Returns:	The library file position where it begins.	*/
/*									*/
/************************************************************************/

SLONG
search(position,s)
	SLONG			position;
	STRING			s;
	{
	BOOL			not_found;
	SLONG			rval;
	register	int	temp;

	lib_seek(position,SEEK_SET);	/* Point at T_FILENAME/T_EOF record */

	not_found = TRUE;
	while((temp = inl_1b()) == T_FILENAME)
		{
		rval = lib_tell() - 1;
		rd_lib_fn();
		if(filecmp(fnbuf,s) == 0)
			{
			not_found = FALSE;
			break;
			}
		while((temp = inl_1b()) == T_SYMBOL)
			rd_lib_sym();
		if(temp != T_SIZE)
			fatal(brf,libbuf);
		inl_4b();		/* Toss modification time/date */
		lib_seek((SLONG) inl_4b(),SEEK_CUR);
		}
	if(not_found)
		{
		if(temp != T_EOF)
			fatal(brf,libbuf);
		else
			fatal("Library doesn't contain file: ",s);
		}
	lib_seek((SLONG) rval,SEEK_SET);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	set_mtime					*/
/*									*/
/*	Set last modification time/date for specified filename.		*/
/*									*/
/*	Parameters:							*/
/*		fname -	The name of the file to set.			*/
/*		time -	The time set the file to.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
set_mtime(fname,time)
	STRING			fname;
	long			time;
	{
#if	(H_OPSYS != MPW) && (H_OPSYS != XENIX) && (H_OPSYS != VMS)
	struct	utimbuf		times;

	times.actime = time;		/* Set access time field */
	times.modtime = time;		/* Set modification time field */

	if(utime(fname,&times) != 0)
		fatal("Unable to set modification time for file: ",fname);
#endif	/* !MPW && !XENIX && !VMS */
	}

/************************************************************************/
/*									*/
/*	Routine:	split						*/
/*									*/
/*	Split a library into its constituent files and remove the	*/
/*	library file.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
split()
	{
	register	int	temp;
	UINT			lib_count;
	SLONG			file_size;
	long			time;

	lib_count = 0;
	if(verbose)
		printf("Splitting library '%s'into:\n\n",libbuf);

	while((temp = inl_1b()) == T_FILENAME)
		{
		lib_count++;
		rd_lib_fn();		/* Read T_FILENAME record */
		strcpy(iobuf,fnbuf);	/* Output file name to use */
/*
 *	Open prospective output filename.
 */
#if	CRLF
		if((io_fp = (STREAM) fopen(iobuf,"wb")) == NULL)
#else
		if((io_fp = (STREAM) fopen(iobuf,"w")) == NULL)
#endif
			fatal("Unable to open output file: ",iobuf);
		if(verbose)
			printf("Creating file '%s'\n",iobuf);
/*
 *	Strip off T_SYMBOL definition header records.
 */
		while((temp = inl_1b()) == T_SYMBOL)
			rd_lib_sym();
/*
 *	Get size for new file.
 */
		if(temp != T_SIZE)
			fatal(brf,libbuf);
		time = inl_4b();	/* Get modification time */
		file_size = inl_4b();
/*
 *	Write file out.
 */
		while(file_size--)
			out_1b((int) inl_1b());
/*
 *	Close output file and go on.
 */
		if(fclose(io_fp) == EOF)
			fatal("Unable to close output file: ",iobuf);
		set_mtime(iobuf,time);	/* Set modification time */
		}
	if(temp != T_EOF)
		fatal(brf,libbuf);
/*
 *	Give a summary to the user about what was done.
 */
	if(verbose)
		{
		if(lib_count)
			{
		printf("\n%u file(s) created from splitting library '%s'\n",
				lib_count,libbuf);
			}
		else
			printf("\nNo files created\n");
		}
/*
 *	Close and remove library file.
 */
	if(fclose(lib_fp) == EOF)
		warning("Unable to close library file: ",libbuf);
	else if(unlink(libbuf) == -1)
		warning("Unable to remove library file: ",libbuf);
	lib_fp = NULL;
	}

/************************************************************************/
/*									*/
/*	Routine:	tag_files					*/
/*									*/
/*	Annotate the "fname" array to show which library files will be	*/
/*	affected by whatever command routine calls "tag_files".		*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*		partial-A BOOL which is TRUE if a file listed on the	*/
/*			command line is not required to be present.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
tag_files(argc,argv,partial)
	int			argc;
	char			*argv[];
	BOOL			partial;
	{
	register	UINT	i;
	register	STRING	s;

	get_fnames();			/* Fill "fnames" structure */
	while(argc--)			/* Check each command line argument */
		{
		s = *++argv;
		if(*s != '-')
			{
			makefn(s,iobuf,TLNK_INPUT_EXT);	/* Complete filename */
			for(i = 0; i < num_files; i++)	/* Check each file */
				{
				if(filecmp(fnames[i].name,iobuf) == 0)
					{
					fnames[i].value = TRUE;
					break;
					}
				}
			if(i == num_files)
				{
				if(partial)
					{
					not_present = TRUE;
			warning("Library doesn't contain file: ",iobuf);
					}
				else
					{
				fatal("Library doesn't contain file: ",iobuf);
					}
				}
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	temp_seek					*/
/*									*/
/*	Seek the library temp file pointer.				*/
/*									*/
/*	Parameters:							*/
/*		distance-A SLONG representing the distance to seek.	*/
/*		origin-	A int which specifies where the seek begins.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
temp_seek(distance,origin)
	SLONG			distance;
	int			origin;
	{
/*
 *	The following "rewind" should be unnecessary, but a bug in Microsoft's
 *	"fseek" for the compact memory model (V4.0) seems to require it.  The
 *	bug seems to happen on "fseek" operations which move towards the
 *	beginning of the file from the current position, but hasn't been nailed
 *	down completely...
 */
	if(origin == SEEK_SET)
		rewind(temp_fp);
	if(fseek(temp_fp,distance,origin) != 0)
		fatal("Error seeking temp file: ",tempbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	temp_tell					*/
/*									*/
/*	Return the current position of the temp file pointer.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

SLONG
temp_tell()
	{
	SLONG			rval;

	if((rval = ftell(temp_fp)) == (SLONG) -1)
		{
		fatal("Error determining where temp file pointer is: ",
			tempbuf);
		}
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	terminate					*/
/*									*/
/*	If an error was indicated erase the temp file (if one was	*/
/*	created).  If no error is indicated, and a temp file was used,	*/
/*	delete the original library file and rename the temp file with	*/
/*	the original library.						*/
/*									*/
/*	Parameters:							*/
/*		rval -	The desired "exit" value to return.		*/
/*									*/
/*	Returns:	None, and takes its time about it too.		*/
/*									*/
/************************************************************************/

VOID
terminate(rval)
	int			rval;
	{
	if(rval == ERRORS)		/* If had errors, remove temp file */
		{
		if(temp_fp != NULL)	/* Temp file used? */
			{
			if(fclose(temp_fp) == EOF)
				{
				warning("Unable to close temp file: ",
					tempbuf);
				}
			else if(unlink(tempbuf) == -1)
				{
				warning("Unable to remove temp file: ",
					tempbuf);
				}
			}
		}
	else				/* No errors detected */
		{
		if(temp_fp != NULL)	/* Temp file used? */
			{
			if(fclose(temp_fp) == EOF)
				{
				warning("Unable to close temp file: ",
					tempbuf);
				}
			else if(job == BUILD)
				{
#if	H_OPSYS == MPW
				if(faccess(tempbuf,F_RENAME,
						(long *) libbuf) != 0)
#else	/* MPW */
				if(rename(tempbuf,libbuf) != 0)
#endif	/* MPW */
					{
				warning("Unable to change temp file name: ",
					tempbuf);
					}
				}
			else
				{
				if(lib_fp != NULL)	/* Lib file open? */
					{
					if(fclose(lib_fp) == EOF)
						{
				warning("Unable to close library file: ",
							libbuf);
						}
					}
				if(unlink(libbuf) == -1)
					{
				warning("Unable to remove old library file: ",
						libbuf);
					}
#if	H_OPSYS == MPW
				else if(faccess(tempbuf,F_RENAME,
						(long *) libbuf) != 0)
#else	/* MPW */
				else if(rename(tempbuf,libbuf) != 0)
#endif	/* MPW */
					{
				warning("Unable to change temp file name: ",
						tempbuf);
					}
				}
			}
		}
	exit(rval);			/* We are history */
	}

/************************************************************************/
/*									*/
/*	Routine:	update_files					*/
/*									*/
/*	Replace files in the library with newer versions.		*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
update_files(argc,argv)
	int			argc;
	char			*argv[];
	{
	SLONG			file_size;
	register	UINT	i;
	register	STRING	s;
	SLONG			start_point;
	register	int	temp;
	SLONG			was_here;

	tag_files(argc,argv,TRUE);	/* Figure which files will be done */
	rewind(lib_fp);			/* Rewind library file */
	outt_1b((int) inl_1b());	/* Copy T_LIB record */
	outt_1b((int) inl_1b());
	start_point = lib_tell();	/* Point to first T_FILENAME record */
	for(i = 0; i < num_files; i++)
		{
		if(fnames[i].value)
			{
			file_size = search(start_point,fnames[i].name) -
				start_point;
			lib_seek((SLONG) start_point,SEEK_SET);
			while(file_size--)
				outt_1b((int) inl_1b());
			if(inl_1b() != T_FILENAME)
				fatal(brf,libbuf);
			rd_lib_fn();	/* Get update filename */
/* Skip T_SYMBOL rec */	while((temp = inl_1b()) == T_SYMBOL)
				rd_lib_sym();
			if(temp != T_SIZE)
				fatal(brf,libbuf);
			inl_4b();	/* Toss old modification time/date */
/* Skip old copy */	start_point = inl_4b();
			start_point += lib_tell();
			strcpy(iobuf,fnbuf);	/* Update file name to use */
/*
 *	Open prospective update filename.
 */
#if	CRLF
			if((io_fp = (STREAM) fopen(iobuf,"rb")) == NULL)
#else
			if((io_fp = (STREAM) fopen(iobuf,"r")) == NULL)
#endif
				fatal("Unable to open update file: ",iobuf);
			if(verbose)
				{
			printf("Updating file '%s' within library '%s'\n",
					iobuf,libbuf);
				}
			if(in_1b() != T_REL_FILE)
				fatal(brf,iobuf);
/*
 *	Ensure that the CPU type of the update file and temp file match.
 */
			temp = in_1b();	/* Get input file CPU type */
			if(cpu_type == TYPE_UNKNOWN)
				{
				cpu_type = temp;	/* New type */
				was_here = temp_tell();
				temp_seek((SLONG) 1,SEEK_SET);
				outt_1b((int) temp);	/* Set lib CPU type */
				temp_seek((SLONG) was_here, SEEK_SET);
				}
			else
				{
				if(temp != TYPE_UNKNOWN)
					{
					if(temp != cpu_type)
						{
			fatal("Unable to mix code for different CPU types","");
						}
					}
				}
/*
 *	Write the T_FILENAME header record for the updated file.
 */
			outt_1b((int) T_FILENAME);
			outt_4b((SLONG) 0);	/* Dummy old&new line #'s */
			outt_1b((int) fnsize);	/* Filename length */
			s = iobuf;
			while(*s)
				outt_1b((int) *s++);
/*
 *	Read the T_SYMBOL records from the update file, note which are public,
 *	sort those and put in the library file after the T_FILENAME record.
 */
			process_defs();
/*
 *	Figure out how big the update file is.
 */
			io_seek((SLONG) -3,SEEK_END);	/* Where T_EOF is? */
			if(in_1b() != T_EOF)
				{
				fatal("Unable to find end of update file: ",
					iobuf);
				}
			file_size = io_tell() + (SLONG) 2;	/* File size */
			outt_1b((int) T_SIZE);
			outt_4b((SLONG) get_mtime(io_fp,iobuf));
			outt_4b((SLONG) file_size);
/*
 *	Read update file into library file.
 */
			rewind(io_fp);	/* Reset input file */
			while(file_size--)	/* Append file */
				outt_1b((int) in_1b());
/*
 *	Close update file and go on.
 */
			if(fclose(io_fp) == EOF)
				fatal("Unable to close update file: ",iobuf);
			}
		}
/*
 *	Copy rest of library file to temp file.  If any files listed on the
 *	command line were not already present in the library, add them
 *	by calling "append_files" (we don't add the T_EOF record to the temp
 *	file in this case).
 */
	lib_seek((SLONG) -3,SEEK_END);	/* T_EOF? */
	if(not_present)
		file_size = lib_tell() - start_point;
	else
		file_size = lib_tell() - start_point + 3;
	if(inl_1b() != T_EOF)
		fatal("Unable to find end of library file: ",libbuf);
	lib_seek((SLONG) start_point,SEEK_SET);
	while(file_size--)
		outt_1b((int) inl_1b());
	if(not_present)
		append_files(argc,argv);
	}

/************************************************************************/
/*									*/
/*	Routine:	update_rnode					*/
/*									*/
/*	Adds a "file_node" to either the definition or reference list	*/
/*	for the current symbol's "ref_node".				*/
/*									*/
/*	Parameters:							*/
/*		defined-TRUE if the current file contains a definition	*/
/*			for the symbol.					*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
update_rnode(defined)
	BOOL			defined;
	{
	struct	file_node	*fptr;
	struct	ref_node	*rptr;

	rptr = lookup();		/* Find/get "ref_node" for symbol */
	fptr = get_fnode();		/* Get reference/definition node */
	fptr->fnum = num_files;		/* File # which we are doing */
	if(defined)
		{
		fptr->link = rptr->def_list;
		rptr->def_list = fptr;
		}
	else
		{
		fptr->link = rptr->ref_list;
		rptr->ref_list = fptr;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect librarian invocation to user with a suggestion	*/
/*	of how it should be done.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
usagerr()
	{
printf("FATAL: General usage is 'tlib library_filename [options]*'\n");
printf("Options syntax ('-v' to disable verbose):\n\n");
printf("Append files:	'-a [-v] [file_to_append]*'\n");
printf("Build library:	'-b [-v] [file_to_sort_and_append]*'\n");
printf("Copy-out files:	'-c [-v] [file_to_copy]*'\n");
printf("Delete files:	'-d [-v] [file_to_delete]*'\n");
printf("Extract files:	'-e [-v] [file_to_extract]*'\n");
printf("Insert files:	'-i [-v] file_to_insert_before [file_to_insert]*'\n");
printf("List library:	'[-v]'\n");
printf("Split library:	'-s [-v]'\n");
printf("Update files:	'-u [-v] [file_to_update]*'\n");
	terminate(ERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	visit_node					*/
/*									*/
/*	Helping function for "build_lib" reverse topological sort on	*/
/*	the file dependency graph.					*/
/*									*/
/*	Parameters:							*/
/*		i -	Node #.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
visit_node(i)
	register	int	i;
	{
	register struct	file_node *ptr;

	fnames[i].value = 1;		/* Make !0 and !1 so can cycle chk */

	for(ptr = fnames[i].link; ptr != NULL; ptr = ptr->link)
		{
		if(fnames[ptr->fnum].value == 0)
			visit_node(ptr->fnum);
		else if(fnames[ptr->fnum].value == 1)
			{
	fatal("External symbol dependency cycle detected involving file: ",
				fnames[ptr->fnum].name);
			}
		}
	fnames[last_file].next = i;	/* Keep track of order to append */
	last_file = i;
	fnames[i].value = -1;		/* Already visited */
	}

/************************************************************************/
/*									*/
/*	Routine:	warning						*/
/*									*/
/*	Print out a "WARNING" error message.				*/
/*									*/
/*	Parameters:							*/
/*		sptr1 -	A pointer to the first of the error message.	*/
/*		sptr2 -	A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
warning(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	printf("WARNING: %s%s\n",sptr1,sptr2);	/* Print message */
	}
