/************************************************************************/
/*									*/
/*	File:	tlnk1.c							*/
/*									*/
/*	Transputer Linker (TLNK), source file 1.  This file contains	*/
/*	the top level code for the linker and most of the  utility and	*/
/*	I/O functions.							*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character type functions */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tlnkdef.h"		/* Linker definitions */
#include	"tlnktyp.h"		/* Linker data declarations */
#include	"tlnkext.h"		/* Linker externals */

#if	H_OPSYS == BSD4_2
#include	<sys/file.h>
#define	SEEK_SET	L_SET
#define	SEEK_CUR	L_INCR
#define	SEEK_END	L_XTND
#endif	/* BSD4_2 */

#if	(H_OPSYS == MPW) || (H_OPSYS == XENIX)
#define	SEEK_SET	0
#define	SEEK_CUR	1
#define	SEEK_END	2
#endif	/* MPW || XENIX */

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	I/O definitions.
 */

char	ifnbuf[FNSIZE + 1] =	{0};	/* Input filename buffer */
STREAM	in_fp		=	NULL;	/* Input file pointer */
char	ofnbuf[FNSIZE + 1] =	{'\0'};	/* Output filename buffer */
STREAM	out_fp		=	NULL;	/* Output file pointer */
char	listbuf[FNSIZE + 1] =	{'\0'};	/* Listing filename buffer */
STREAM	list_fp		=	NULL;	/* Listing file pointer */
char	cmdfnbuf[FNSIZE + 1] =	{0};	/* Command file filename buffer */
STREAM	cmdf_fp		=	NULL;	/* Command file file pointer */

VSTREAM	mod_fp[NUMMODULES] =	{NULL};	/* Temp file ptrs for each "module" */
VSTREAM	temp_fp		=	NULL;	/* "extra" temp file pointer */
char	temp_path[FNSIZE + 1] = {'\0'};	/* Pathname for temp files */

#if	VMFILE
struct	vbuf	*vbfree	=	NULL;	/* Virtual memory buffer free list */
#endif	/* VMFILE */

#if	IOBUFSIZE
STRING	temp_fp_buf	=	NULL;	/* Stream buf ptr for "extra" file */
#endif

/*
 *	General definitions.
 */

UCHAR	acc_dta[ACCDATASIZE] =	{0};	/* T_DATA pack accumulator */
int	acc_dtasize	=	0;	/* T_DATA pack accumulated size */
SLONG	acc_storesize	=	0;	/* T_STORAGE accumulated size */
BOOL	aligned[NUMMODULES] =	{0};	/* TRUE if temp file is "aligned" */
UINT	base_symbol	=	0;	/* Local ST start for current file */
BOOL	byaddress	=	FALSE;	/* TRUE/sort symbol table by address */
BOOL	cfile		=	0;	/* TRUE if a command file being used */
UINT	cmd_number	=	0;	/* Command file line # */
BOOL	compress	=	FALSE;	/* TRUE if to pack T_DATA/T_STORAGE */
UINT	cpu_type	=TYPE_UNKNOWN;	/* Processor type not yet selected */
SLONG	entry_n		=	0;	/* Entrypoint address if constant */
UINT	entry_s		=	0;	/* Entrypoint LST entry # if symbol */
UINT	entry_type	=NO_ENTRY;	/* Type of entrypoint selected */
UINT	errors		=	0;	/* Total # of errors encountered */
BOOL	expr_err	=	0;	/* TRUE if an expression error found */
BOOL	first_ifile	=	0;	/* TRUE prior to first input file */

#if	H_OPSYS == MPW
struct	gst	*globals=	NULL;	/* Global symbol table */
struct	lst	*locals	=	NULL;	/* Local symbol table */
#else	/* H_OPSYS != MPW */
#if	HUGE_LOCALST
struct	gst globals[MAX_GLOBALS] = {0};	/* Global symbol table */
struct	lst huge *locals =	NULL;	/* Local symbol table */
#else	/* !HUGE_LOCALST */
struct	gst globals[MAX_GLOBALS] = {0};	/* Global symbol table */
struct	lst locals[MAX_SYMBOLS] = {0};	/* Local symbol table */
#endif	/* HUGE_LOCALST */
#endif	/* H_OPSYS == MPW */

BOOL	is_lib		=	0;	/* TRUE if reading from a library */
SLONG	load_addr	=	0;	/* Default load address */
SLONG	max_addr	=	0;	/* Maximum address (current module) */
SLONG	max_addr_adj	=	0;	/* Max addr fwd ref. adjustment */
UINT	max_i_size	=	MAX_I_SIZE_32B;	/* Max. instr. size in bytes */
SLONG	min_addr	=	0;	/* Minimum address (current module) */
SLONG	min_addr_adj	=	0;	/* Min addr fwd ref. adjustment */
UINT	module		=	0;	/* Current module # */
UINT	next_symbol	=	0;	/* Local ST end (+1 current file) */
UINT	number		=	0;	/* Source file input line # */
UINT	num_ext		=	0;	/* # of external symbols */
BOOL	old_file[NUMMODULES] =	{0};	/* TRUE if mod. doesn't need align. */
char	parse_buf[MAXLNSIZE + 1] = {0};	/* Input parsing buffer */
STRING	parse_ptr	=	NULL;	/* Parsing buffer pointer */
UINT	region		=	0;	/* Current code/data region tag */
SLONG	stack_addr	=	0;	/* Default stack address */
BOOL	static_link	=	FALSE;	/* TRUE for static link data binding */
UINT	sym_col		=	0;	/* Columns # for file symbols */
UINT	udef_list	=	0;	/* Undefined external list head */
UINT	word_size	=	WORD_SIZE_32B;	/* CPU word size in bytes */

/*
 *	The following variables are used by "read_file" and associated helping
 *	routines.
 */

char	fnbuf[FNSIZE + 1] =	{0};	/* Most recent given by T_FILENAME */
UINT	fnsize		=	0;	/* Length of filename in "fnbuf" */
UINT	new_ln		=	0;	/* Most recent given by T_FILENAME */
UINT	old_ln		=	0;	/* Most recent given by T_FILENAME */
BOOL	used_modules[NUMMODULES] = {0};	/* TRUE if filename current */

/*
 *	Local buffering control stuff.
 */

struct	lbuf	get_buf	=	{0};	/* "get_X" buffer */
struct	lbuf	put_buf	=	{0};	/* "put_X" buffer */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Opens the command and output files, allocates free memory and	*/
/*	starts the linking process.					*/
/*									*/
/*	Parameters:							*/
/*		"argc" An integer count of the command line arguments.	*/
/*		"argv" A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#if	H_OPSYS != VMS
VOID
#endif	/* H_OPSYS != VMS */
main(argc, argv)
	int			argc;
	char			*argv[];
	{
#if	HUGE_LOCALST
	extern	void	huge	*halloc();
	size_t			tempsize;
#endif	/* HUGE_LOCALST */
	extern	STRING		malloc();
/*
 *	Check validity of command filename and open it if present.
 */
	if(argc == 1)
		{
		banner();		/* Tell user who we are */
		cfile = FALSE;		/* No command file present */
		}
	else
		{
		if(argc > 2)
			{
			banner();	/* Tell user who we are */
			usagerr();	/* Needs at most a command file */
			}
		if(makefn(*++argv,cmdfnbuf,TLNK_CMD_EXT))	/* Expand */
			fatal("Command filename too long","");
#if	CRLF
		if((cmdf_fp = (STREAM) fopen(cmdfnbuf,"rt")) == NULL)
#else
		if((cmdf_fp = (STREAM) fopen(cmdfnbuf,"r")) == NULL)
#endif
			fatal("Unable to open command file: ",cmdfnbuf);
		cfile = TRUE;		/* We have the command file open */
		}
/*
 *	Verify that the various portability items are configured OK.
 */

/*lint -e506 lint isn't very found of constant valued booleans */

	if(sizeof(SLONG) < 4)
		fatal("The size of SLONG is not correctly configured","");

/*lint +e506 lint isn't very fond of constant valued booleans */

#if	H_OPSYS == MPW
/*
 *	If we are running under MPW we need to "calloc" the local and
 *	global symbol tables.
 */
	if((globals = (struct gst *) calloc(MAX_GLOBALS,sizeof(struct gst)))
		== NULL)
		{
		fatal("Insufficient memory for global symbol table","");
		}
	if((locals = (struct lst *) calloc(MAX_SYMBOLS,sizeof(struct lst)))
		== NULL)
		{
		fatal("Insufficient memory for local symbol table","");
		}
#else	/* H_OPSYS != MPW */
#if	HUGE_LOCALST
/*
 *	If we are running under MS-DOS with a local symbol table bigger than
 *	64K we need to use a "huge" pointer for it and dynamically allocate it.
 *	Note that the size of the structure must be rounded up to be a power of
 *	two if the total allocation will exceed 128K bytes.
 */
	tempsize = sizeof(struct lst);
/*lint -e506 lint isn't very found of constant valued booleans */
	if(((long) MAX_SYMBOLS * (long) sizeof(struct lst)) > 131072L)
/*lint +e506 lint isn't very found of constant valued booleans */
		{
		if((tempsize > 64) || (tempsize < 8))
			fatal("Local symbol table configuration error","");
		else if(tempsize > 32)
			tempsize = 64;
		else if(tempsize > 16)
			tempsize = 32;
		else if(tempsize > 8)
			tempsize = 16;
		}
	if((locals = (struct lst huge *) halloc((long) MAX_SYMBOLS,tempsize))
		== NULL)
		{
		fatal("Insufficient memory for local symbol table","");
		}
#endif	/* HUGE_LOCALST */
#endif	/* H_OPSYS == MPW */
/*
 *	Get the memory to be used for the input and temp file local buffers.
 */
	if((get_buf.start = (UCHAR *) malloc(LOCAL_BUF_SIZE)) == NULL)
		fatal("Insufficient local buffer memory","");
	get_buf.count = 0;
#if	!VMFILE
	if((put_buf.start = (UCHAR *) malloc(LOCAL_BUF_SIZE)) == NULL)
		fatal("Insufficient local buffer memory","");
	put_buf.fp = NULL;
	put_buf.ptr = put_buf.start;
	put_buf.count = 0;
#endif	/* !VMFILE */
/*
 *	We are ready to go, now is the time to read the operator commands
 *	and relocatable input files, assemble/link the code and produce a
 *	listing file if directed to do so.
 */
	parse();			/* Parse user commands */
	assemble();			/* Assemble/link the code */
	}

/************************************************************************/
/*									*/
/*	Routine:	assemble					*/
/*									*/
/*	Assemble, link, generate listing info and the linkage output	*/
/*	file.								*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
assemble()
	{
	register	UINT	temp;
	register	UINT	udef_count;
/*
 *	First add T_ALIGN (if needed), and T_EOF records to all the "module"
 *	temp files (which are full as a result of the "parse" routine).
 */
	for(module = 0; module < NUMMODULES; module++)
		{
		if(mod_fp[module] != NULL)	/* See if "module" in use */
			{
			put_flush();		/* Flush buffer */
			if(! aligned[module])	/* Need aligning? */
				{
				PUT_1B((int) T_ALIGN);	/* Record type */
				put_2b((int) 0);	/* Dummy line # */
				}
			PUT_1B((int) T_EOF);	/* Record type */
			put_2b((int) 0);	/* Dummy line # */
			}
		}
#if	DEBUG
	putchar('[');			/* Bracket pass indicators */
#endif
	get_buf.count = 0;		/* Initialize "get" local buffer */
	pass0();			/* Initialize symbol table addresses */
/*
 *	Do the iterative "growing" of PC relative immediate data until enough
 *	prefix instructions are generated to handle the data.  These operations
 *	are not performed if TLNK found any errors.  When finished we generate
 *	the output file as well as the listing file (if requested).
 */
	if(! errors)
		{
		FOREVER
			{
			if(pass1())
				break;	/* Quit if no better */
			}
		pass2();		/* Grow boundary cases */
		pass1();		/* Final pass */
		pass3();		/* Generate output & listing files */
		}
#if	DEBUG
	putchar(']');			/* Bracket pass indicators */
#endif
/*
 *	If undefined symbols exist, display a sorted list to standard output.
 */
	udef_count = 0;
	if(! sort_udef())		/* Any undefined symbols? */
		{
		printf("\nUndefined symbol(s):\n");
		udef_count = 0;		/* Keep track of # of undefined */
		for(temp = udef_list; temp != MAX_SYMBOLS;
			temp = globals[temp - (MAX_SYMBOLS + 1)].lst_num)
			{
			printf("%s\n",globals[temp - (MAX_SYMBOLS+1)].name);
			udef_count++;
			}
		}
/*
 *	Assuming we are being run interactively, add the usage statistics to
 *	standard output.
 */
	if(! cfile)
		{
		printf("\nExternal symbols: %u",num_ext);
		printf("\tExternal symbol table usage: %u%%\n",
			(UINT) (((SLONG) num_ext * (SLONG) 100) / (SLONG)
				MAX_GLOBALS));
		printf("Total symbols:    %u",next_symbol);
		printf("\tTotal symbol table usage:    %u%%\n",
			(UINT) (((SLONG) next_symbol * (SLONG) 100) / (SLONG)
				MAX_SYMBOLS));
		}
/*
 *	Assuming we have a listing file, add the usage statistics to it.
 */
	if(list_fp != NULL)
		{
		list_st();		/* Add symbol table dump */
		fprintf(list_fp,"\nExternal symbols: %u",num_ext);
		fprintf(list_fp,"\tExternal symbol table usage: %u%%\n",
			(UINT) (((SLONG) num_ext * (SLONG) 100) / (SLONG)
				MAX_GLOBALS));
		fprintf(list_fp,"Total symbols:    %u",next_symbol);
		fprintf(list_fp,"\tTotal symbol table usage:    %u%%\n",
			(UINT) (((SLONG) next_symbol * (SLONG) 100) / (SLONG)
				MAX_SYMBOLS));
		if(udef_count == 0)
			putc('\n',list_fp);
		else
			fprintf(list_fp,"\n%u undefined symbols\n",udef_count);
		if(errors)
			fprintf(list_fp,"%u errors detected\n",errors);
		else
			fprintf(list_fp,"No errors detected\n");
		}
/*
 *	Show # of undefined symbols.
 */
	if(udef_count == 0)
		{
		if(errors || (! cfile))
			putchar('\n');
		}
	else
		printf("\n%u undefined symbols\n",udef_count);
/*
 *	Show error status information if appropriate.
 */
	if(errors)
		printf("%u errors detected\n",errors);
	else if(! cfile)
		printf("No errors detected\n");
/*
 *	Terminate with correct return code.
 */
	if(errors || (udef_count > 0))
		terminate(ERRORS);
	else
		terminate(NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display sign-on message to user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
banner()
	{
	printf("%s, %s, %s\n\n",LNK_NAME,VERSION,COPYRIGHT);
	}

/************************************************************************/
/*									*/
/*	Routine:	copy_in						*/
/*									*/
/*	Copy a specified # of bytes from the input file to the current	*/
/*	"module" temp file.						*/
/*									*/
/*	Parameters:							*/
/*		"count"	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy_in(count)
	register	UINT	count;
	{
	while(count--)
		PUT_1B(IN_1B());
	}

/************************************************************************/
/*									*/
/*	Routine:	copy_out					*/
/*									*/
/*	Copy a specified # of bytes from the "extra" temp file to the	*/
/*	output file.							*/
/*									*/
/*	Parameters:							*/
/*		"count"	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy_out(count)
	register	UINT	count;
	{
	while(count--)
		out_1b(GET_1B());
	}

/************************************************************************/
/*									*/
/*	Routine:	copy_temp					*/
/*									*/
/*	Copy a specified # of bytes from the "extra" temp file to the	*/
/*	current "module" temp file.					*/
/*									*/
/*	Parameters:							*/
/*		"count"	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
copy_temp(count)
	register	UINT	count;
	{
	if((get_buf.count >= count) &&
#if	VMFILE
		((VBSIZE - put_buf.count) >= count)
#else	/* !VMFILE */
		((LOCAL_BUF_SIZE - put_buf.count) > count)
#endif	/* VMFILE */
		)
		{
		(VOID) memcpy((char *) put_buf.ptr,(char *) get_buf.ptr,count);
		get_buf.ptr += count;
		get_buf.count -= count;
		put_buf.ptr += count;
		put_buf.count += count;
		}
	else
		{
/*		while(count--)
			PUT_1B(GET_1B());
*/
		while(((int) count) > 0)
			{
			switch((int) count)
				{
				default:
					PUT_1B(GET_1B());
				case 3:
					PUT_1B(GET_1B());
				case 2:
					PUT_1B(GET_1B());
				case 1:
					PUT_1B(GET_1B());
				}
			count -= 4;
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	eval						*/
/*									*/
/*	Evaluate "C" style decimal, hexadecimal and octal numbers.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The numeric value.				*/
/*									*/
/************************************************************************/

SLONG
eval()
	{
	register	int	c;
	SLONG			val;

	while(isspace(*parse_ptr))
		parse_ptr++;	/* Skip whitespace */
	if(isdigit(c = *parse_ptr++))	/* Is it a number? */
		{
		val = 0;
/*
 *	See if it is a octal or hex # or if it is a decimal #.
 */
		if(c == '0')
			{
			c = *parse_ptr++;
/*
 *	Its either octal or hex, figure out which.
 */
			if((c == 'x') || (c == 'X'))
				{
/*
 *	Its a hexadecimal #.
 */
				while(ishex(c = *parse_ptr++))
					{
/*	Detect overflow		*/	if(val & 0xF0000000)
						{
						expr_err = TRUE;
						return (val);
						}
					val = (val << 4) + hexbin(c);
					}
				parse_ptr--;
				}
			else
/*
 *	Its an octal #.
 */
				{
				while(isoct(c))
					{
/*	Detect overflow		*/	if(val & 0xE0000000)
						{
						expr_err = TRUE;
						return (val);
						}
					val = (val << 3) + c - '0';
					c = *parse_ptr++;
					}
				parse_ptr--;
				}
			}
		else
/*
 *	Its a decimal #.
 */
			{
			val = (val * 10) + c - '0';
			while(isdigit(c = *parse_ptr++))
				{
				val = (val * 10) + c - '0';
				}
			parse_ptr--;
			}
		}
	else
/*
 *	Wasn't any # we could recognize!
 */
		{
		expr_err = TRUE;
		}
/*
 *	Return the result.
 */
	return (val);
	}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		"sptr1"	A pointer to the first of the error message.	*/
/*		"sptr2" A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	printf("FATAL: %s%s\n",sptr1,sptr2);	/* Print message */
	if(list_fp != NULL)
		fprintf(list_fp,"FATAL: %s%s\n",sptr1,sptr2);
	terminate(ERRORS);		/* Bye... */
	}

/************************************************************************/
/*									*/
/*	Routine:	fatali						*/
/*									*/
/*	Print out a internal "FATAL" error message and quit.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the error message.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatali(sptr)
	STRING			sptr;
	{
	fatal("TLNK internal error #",sptr);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_next,get_2b,get_4b				*/
/*									*/
/*	Read 1, 2 or 4 bytes from the current "extra" temp file.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The 1, 2 or 4 byte value read.			*/
/*									*/
/************************************************************************/

int
get_next()
	{
#if	VMFILE
	register struct	vbuf	*vb;

	if(((vb = temp_fp->current) == NULL) ||
		((temp_fp->current = vb->next) == NULL))
		{
		fatal("Unexpected end of vmfile","");
		}
	vb->next = vbfree;
	vbfree = vb;
	get_buf.ptr = temp_fp->current->vbuffer;
	get_buf.count = temp_fp->current->vcount;
#else	/* VMFILE */
	if((get_buf.count = fread((char *) (get_buf.ptr = get_buf.start),1,
		LOCAL_BUF_SIZE,temp_fp)) == 0)
		{
		if(temp_path[0] != '\0')
			fatal("Error reading temp file on path: ",temp_path);
		else
			{
			fatal("Error reading temp file on current directory",
				"");
			}
		}
#endif	/* VMFILE */
	return (GET_1B());
	}

int
get_2b()
	{
	int			rval;

	rval = GET_1B();
	return (rval | (GET_1B() << 8));
	}

SLONG
get_4b()
	{
	SLONG			rval;

	rval = (SLONG) GET_1B();
	rval |= (((SLONG) GET_1B()) << 8);
	rval |= (((SLONG) GET_1B()) << 16);
	return (rval | (((SLONG) GET_1B()) << 24));
	}

/************************************************************************/
/*									*/
/*	Routine:	get_fname					*/
/*									*/
/*	Get a file or pathname from the parse buffer.			*/
/*									*/
/*	Parameters:							*/
/*		"buf"	A pointer to where to put the result.		*/
/*									*/
/*	Returns:	TRUE if an error was detected.			*/
/*									*/
/************************************************************************/

BOOL
get_fname(buf)
	char			buf[];
	{
	register	UINT	i;
	char			tbuf[FNSIZE + 1];

	skip_space();			/* Skip whitespace */
	for(i = 0; i < FNSIZE; i++)
		{
		if(not_fname(*parse_ptr))
			break;		/* End of name */
		tbuf[i] = *parse_ptr++;
		}
	if((i == 0) || (i == FNSIZE))
		return (TRUE);
	tbuf[i] = '\0';
	strcpy(buf,tbuf);		/* Update with new value */
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_num						*/
/*									*/
/*	Use "eval" to compute the value of a numeric string and give	*/
/*	an error message if something isn't right.			*/
/*									*/
/*	Parameters:							*/
/*		"result"	A pointer to a signed long where the	*/
/*				"result" goes.				*/
/*									*/
/*	Returns:	TRUE if error found, FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
get_num(result)
	SLONG			*result;
	{
	SLONG			rval;

	skip_space();			/* Skip whitespace */
	expr_err = FALSE;		/* No errors to start with */
	rval = eval();			/* Get value */
	if(expr_err)
		{
		bad_command("Illegal numeric constant","");
		return (TRUE);
		}
	*result = rval;			/* Store result away */
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	hexbin						*/
/*									*/
/*	Convert a hexadecimal digit to binary.				*/
/*									*/
/*	Parameters:							*/
/*		"c"	The digit to convert.				*/
/*									*/
/*	Returns:	The resulting binary value.			*/
/*									*/
/************************************************************************/

SLONG
hexbin(c)
	int			c;
	{
	if(isdigit(c))
		return ((SLONG) (c - '0'));
	else
		return ((SLONG) ((isupper(c) ? tolower(c) : c) - 'a' + 10));
	}

/************************************************************************/
/*									*/
/*	Routine:	in_seek						*/
/*									*/
/*	Seek the input file the specified # of characters forward from	*/
/*	the current position.						*/
/*									*/
/*	Parameters:							*/
/*		"distance" -	How far to go forward.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
in_seek(distance)
	SLONG			distance;
	{
	if(fseek(in_fp,distance,SEEK_CUR) != 0)
		fatal("Error seeking file: ",ifnbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	in_next,in_1b,in_2b,in_4b			*/
/*									*/
/*	Read 1, 2 or 4 bytes from the current relocatable input file.	*/
/*	Note that the "in_next" routine is a helping function for the	*/
/*	IN_1B macro which uses the locally bufferred input stream when	*/
/*	reading from input files ("is_lib" is FALSE).			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The 1, 2 or 4 byte value read.			*/
/*									*/
/************************************************************************/

int
in_next()
	{
	if((get_buf.count = fread((char *) (get_buf.ptr = get_buf.start),1,
		LOCAL_BUF_SIZE,in_fp)) == 0)
		{
		fatal("Error reading input file: ",ifnbuf);
		}
	return (IN_1B());
	}

int
in_1b()					/* Used for reading from libraries */
	{
	int			c;

	if((c = fgetc(in_fp)) == EOF)
		{
		if(ferror(in_fp) || feof(in_fp))
			fatal("Error reading input file: ",ifnbuf);
		}
	return (c & 0xFF);
	}

int
in_2b()
	{
	int			rval;

	rval = IN_1B();
	return ((rval | (IN_1B() << 8)) & 0xFFFF);
	}

SLONG
in_4b()
	{
	SLONG			rval;

	rval = (SLONG) IN_1B();
	rval |= (((SLONG) IN_1B()) << 8);
	rval |= (((SLONG) IN_1B()) << 16);
	return (rval | (((SLONG) IN_1B()) << 24));
	}

/************************************************************************/
/*									*/
/*	Routine:	init_input					*/
/*									*/
/*	Initialize everything required before the input files are read.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
init_input()
	{
	register	UINT	i;

#if	!VMFILE
	char			buf1[FNSIZE + 1];
	char			buf2[FNSIZE + 1];
/*
 *	First, close and remove any existing temp files (this is only useful if
 *	we are running in interactive mode and the user has supplied a new
 *	set of input files after we've complained about the previous set).
 */
	if(temp_fp != NULL)		/* "extra" temp file opened? */
		{
		strcpy(buf1,temp_path);
		strcat(buf1,EXTRA_TEMP);
		if(fclose(temp_fp) == EOF)
			fatal("Unable to close temp file: ",buf1);
		else if(unlink(buf1) == -1)
			fatal("Unable to remove temp file: ",buf1);
		temp_fp = NULL;		/* Indicate its closed now! */
#if	IOBUFSIZE
		free(temp_fp_buf);	/* Free explicit buffer used */
#endif	/* IOBUFSIZE */
		put_buf.fp = NULL;	/* Initialize "put" local buffer */
		put_buf.ptr = put_buf.start;
		put_buf.count = 0;
		}
/*
 *	Now close any "module" temp files.
 */
	for(module = 0; module < NUMMODULES; module++)
		{
		if(mod_fp[module] != NULL)	/* temp for this module? */
			{
			strcpy(buf1,temp_path);
			strcat(buf1,TEMP_NAME);
			sprintf(buf2,"%u",module);
			strcat(buf1,buf2);
			if(fclose(mod_fp[module]) == EOF)
				fatal("Unable to close temp file: ",buf1);
			else if(unlink(buf1) == -1)
				fatal("Unable to remove temp file: ",buf1);
			mod_fp[module] = NULL;	/* Indicate its closed now! */
			}
		}
#endif	/* !VMFILE */
/*
 *	Now close and delete any existing listing file.
 */
	if(list_fp != NULL)
		{
		if(fclose(list_fp) == EOF)
			fatal("Unable to close listing file: ",listbuf);
		else if(unlink(listbuf) == -1)
			fatal("Unable to remove listing file: ",listbuf);
		}
/*
 *	Reset various other variables.
 */
	get_buf.count = 0;		/* Initialize "in" local buffer */
	for(module = 0; module < NUMMODULES; module++)
		used_modules[module] = aligned[module] = FALSE;
	first_ifile = TRUE;		/* Next input file will be first */
	number = 1;			/* Default source file line # */
	sym_col = 0;			/* Reset symbol name column */
/*
 *	Initialize the global symbol table entry # field in the local symbol
 *	table to be an element just outside the global symbol table.
 *	This value indicates that no global symbol corresponds to the local
 *	symbol table entry (initially anyway).  These fields are changed to be
 *	valid global symbol table entries when corresponding T_SYMBOL records
 *	are found in relocatable input files.
 */
	for(i = 0; i < MAX_SYMBOLS; i++)
		locals[i].gst_num = MAX_GLOBALS;
	next_symbol = base_symbol = 0;	/* Local symbol table counters */
/*
 *	Initialize the local symbol table entry # field in the global symbol
 *	table to be an element just outside the local symbol table.  This value
 *	holds the local symbol table entry # of the definition of the global
 *	symbol table entry (unknown to start with).  If a "name" has been
 *	associated with any global symbol table entries, then the space used is
 *	returned (using "free").
 */
	for(i = 0; i < MAX_GLOBALS; i++)
		{
		if(globals[i].name != NULL)
			{
			free(globals[i].name);
			globals[i].name = NULL;
			}
		globals[i].lst_num = MAX_SYMBOLS;
		}
	num_ext = 0;			/* No global/external symbols */
/*
 *	Initialize current module #.
 */
	module = 0;
	}

/************************************************************************/
/*									*/
/*	Routine:	ishex						*/
/*									*/
/*	Test a digit to see if it could be part of a hexadecimal #.	*/
/*									*/
/*	Parameters:							*/
/*		"c"	The digit to test.				*/
/*									*/
/*	Returns:	TRUE if so, FALSE if not.			*/
/*									*/
/************************************************************************/

BOOL
ishex(c)
	int			c;
	{
	return (isdigit(c) ||
		((c >= 'a') && (c <= 'f')) || ((c >= 'A') && (c <= 'F')));
	}

/************************************************************************/
/*									*/
/*	Routine:	isoct						*/
/*									*/
/*	Test a digit to see if it could be part of a octal #.		*/
/*									*/
/*	Parameters:							*/
/*		"c"	The digit to test.				*/
/*									*/
/*	Returns:	TRUE if so, FALSE if not.			*/
/*									*/
/************************************************************************/

BOOL
isoct(c)
	int			c;
	{
	return ((c >= '0') && (c <= '7'));
	}

/************************************************************************/
/*									*/
/*	Routine:	lbanner						*/
/*									*/
/*	Write message to list file user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
lbanner()
	{
	if(list_fp != NULL)
		fprintf(list_fp,"%s, %s, %s\n",LNK_NAME,VERSION,COPYRIGHT);
	}

/************************************************************************/
/*									*/
/*	Routine:	lookup						*/
/*									*/
/*	Check if a symbol is in the global symbol table, if not add it.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the start of the symbol.		*/
/*									*/
/*	Returns:	The entry # for the existing/resulting global	*/
/*			symbol table entry.				*/
/*									*/
/************************************************************************/

UINT
lookup(sptr)
	STRING			sptr;
	{
	extern	STRING		malloc();

	register	UINT	i;
	register	UINT	num;
	register struct	gst	*ptr;
	register	STRING	tptr;

	tptr = sptr;			/* Point to symbol */
	num = 0;			/* Zero hash sum */
	while(*tptr != '\0')
		num = (num * 16) + *tptr++;	/* Compute hash */
	num &= GLOBALS_MASK;		/* Mask to size of table */
/*
 *	We now have the symbol table entry number to start scanning at,
 *	but we need to convert that to the correct pointer, therefore:
 */
	ptr = globals + num;
/*
 *	Scan table for exact match or empty slot.
 */
	for(i = 0; i < MAX_GLOBALS; i++ , num++ , ptr++)
		{
		if(ptr >= &globals[MAX_GLOBALS])	/* Wrap around? */
			{
			num = 0;	/* If so */
			ptr = globals;
			}
		if(ptr->name != NULL)	/* Entry filled? */
			{
			if(! strncmp(sptr,ptr->name,SYMNMESIZE))
				return (num);	/* Exact match */
			continue;	/* Filled, try next */
			}
		num_ext++;		/* Another external symbol */
		i = strlen(sptr) + 1;	/* Figure out how much space we need */
		if((tptr = (STRING)
			malloc((i < SYMNMESIZE) ? i : SYMNMESIZE)) == NULL)
			{
		fatal("Insufficient global symbol table string memory","");
			}
		string_copy(tptr,sptr,SYMNMESIZE);
		ptr->name = tptr;	/* Remember string location */
		return (num);
		}
	fatal("Global symbol table full","");
/*lint -e533 lint doesn't understand about functions which never return */
	}
/*lint +e533 lint doesn't understand about functions which never return */

/************************************************************************/
/*									*/
/*	Routine:	maketdir					*/
/*									*/
/*	If the user has supplied a temp file directory, put it in	*/
/*	canonical form with a trailing delimiter.  If no explicit	*/
/*	directory has been supplied check the TEMP_PATH environment	*/
/*	variable for a suitable directory.  If the environment variable	*/
/*	doesn't exist use the current directory.			*/
/*									*/
/*	Parameters:	None						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
maketdir()
	{
	UINT			i;
	STRING			s;

	if(temp_path[0] == '\0')	/* Did user provide anything? */
		{			/* If not, is there a env. var.? */
		if((s = getenv(TEMP_PATH)) != NULL)
			{
			if(strlen(s) < FNSIZE)
				strcpy(temp_path,s);
			}
		}
/*
 *	Ensure temp directory has trailing delimiter.
 */
#if	H_OPSYS == MPW
	if(((i = strlen(temp_path)) > 0) && (temp_path[i - 1] != ':'))
		strcat(temp_path,":");
#else	/* !MPW */
	if(((i = strlen(temp_path)) > 0) && (temp_path[i - 1] != '/') &&
		(temp_path[i - 1] != '\\'))
		{
		strcat(temp_path,"/");
		}
#endif	/* MPW */
	}

/************************************************************************/
/*									*/
/*	Routine:	makefn						*/
/*									*/
/*	Expand a skeleton filename by adding a default extension if	*/
/*	none is supplied.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the skeleton filename.		*/
/*		"bptr"	A pointer to the buffer the result goes in.	*/
/*		"ext"	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	TRUE if filename too long, FALSE otherwise.	*/
/*									*/
/************************************************************************/

BOOL
makefn(sptr,bptr,ext)
	STRING			sptr;
	STRING			bptr;
	STRING			ext;
	{
	extern	STRING		strchr();
	extern	STRING		strrchr();

	STRING			s;

	if(strlen(sptr) > FNSIZE)
		return (TRUE);
	strcpy(bptr,sptr);
	if((s = strrchr(bptr,'.')) != NULL)
		{
		if((strchr(s,'/') == NULL) && (strchr(s,'\\') == NULL))
			return (FALSE);
		}
	if(strlen(bptr) > (FNSIZE - 4))
		return (TRUE);
	strcat(bptr,ext);
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	match						*/
/*									*/
/*	Used by the input parsing routines to check and see if the next	*/
/*	characters pointed at by "parse_ptr" match the string passed to	*/
/*	"match".  If so "parse_ptr" is updated to point beyond the	*/
/*	matched item.  Preceeding whitespace is skipped	("parse_ptr"),	*/
/*	prior to performing the match.  Note that the match is case	*/
/*	independent (the parameter to "match" should be upper case).	*/
/*									*/
/*	Parameters:							*/
/*		"str"	A pointer to the string being compared.		*/
/*									*/
/*	Returns:	TRUE if successfully matched.			*/
/*									*/
/************************************************************************/

BOOL
match(str)
	STRING			str;
	{
	register	STRING	temp;

	temp = parse_ptr;
	while(isspace(*temp))
		temp++;
	while(*str != '\0')
		{
		if(*str++ != (islower(*temp) ? toupper(*temp) : *temp))
			return (FALSE);
		else
			temp++;
		}
	parse_ptr = temp;		/* Point past matched string */
	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	match_first					*/
/*									*/
/*	Like "match", but also slurps any alphanumerics which		*/
/*	immediately follow the matched string.				*/
/*									*/
/*	Parameters:							*/
/*		"str"	A pointer to the string being compared.		*/
/*									*/
/*	Returns:	TRUE if successfully matched.			*/
/*									*/
/************************************************************************/

BOOL
match_first(str)
	STRING			str;
	{
	if(! match(str))
		return (FALSE);
	while(isalnum(*parse_ptr))
		parse_ptr++;
	return (TRUE);
	}

/************************************************************************/
/*									*/
/*	Routine:	non_fatal					*/
/*									*/
/*	Display a message to the user in response to a non-fatal error	*/
/*	detected while the input files are being read or assembled/	*/
/*	linked.								*/
/*									*/
/*	Parameters:							*/
/*		"s1" -	A pointer to the first string to display.	*/
/*		"s2" -	A pointer to the second string to display.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
non_fatal(s1,s2)
	STRING			s1;
	STRING			s2;
	{
	printf("<%s> @ %u: %s%s\n",fnbuf,number,s1,s2);
	if(list_fp != NULL)
		fprintf(list_fp,"<%s> @ %u: %s%s\n",fnbuf,number,s1,s2);
	errors++;			/* Remember we had an error */
	}

/************************************************************************/
/*									*/
/*	Routine:	not_fname					*/
/*									*/
/*	Return TRUE if the character parameter could not be part of a	*/
/*	filename.							*/
/*									*/
/*	Parameters:							*/
/*		"c" -	The character being tested.			*/
/*									*/
/*	Returns:	TRUE if not part of a legal filename.		*/
/*									*/
/************************************************************************/

BOOL
not_fname(c)
	int			c;
	{
	if(isspace(c) || (c == ',') || (c == ';') || (c == '\n') ||
		(c == '\0'))
		{
		return (TRUE);
		}
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	out_write					*/
/*									*/
/*	Write a specified # of characters to the output file.		*/
/*									*/
/*	Parameters:							*/
/*		"start"	Starting address of characters to write.	*/
/*		"size"	The number to write.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
out_write(start,size)
	STRING			start;
	int			size;
	{
	if(fwrite(start,1,(UINT) size,out_fp) != size)
		fatal("Error writing output file: ",ofnbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	out_1b,out_2b,out_4b				*/
/*									*/
/*	Write 1,2 or 4 bytes to the output file.			*/
/*									*/
/*	Parameters:							*/
/*		"value"	The 1,2 or 4 byte value to write.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
out_1b(value)
	int			value;
	{
	if(fputc(value,out_fp) == EOF)
		{
		if(ferror(out_fp))
			fatal("Error writing output file: ",ofnbuf);
		}
	}

VOID
out_2b(value)
	int			value;
	{
	out_1b(value);
	out_1b(value >> 8);
	}

VOID
out_4b(value)
	SLONG			value;
	{
	out_1b((int) value);
	out_1b((int) (value >>= 8));
	out_1b((int) (value >>= 8));
	out_1b((int) (value >> 8));
	}

/************************************************************************/
/*									*/
/*	Routines:	put_flush,put_2b,put_4b				*/
/*									*/
/*	Flush buffer and write 2 or 4 bytes to the current temp file	*/
/*	(one of those used to accumulate the code in a "module").	*/
/*									*/
/*	Parameters:							*/
/*		"value"	The 2 or 4 byte value to write.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
put_flush()
	{
#if	VMFILE
	if((put_buf.fp != NULL) && (put_buf.fp->current != NULL))
		put_buf.fp->current->vcount = put_buf.count;
	put_buf.fp = mod_fp[module];
	if(put_buf.fp->current == NULL) 
		put_buf.count = VBSIZE;
	else
		{
		put_buf.count = put_buf.fp->current->vcount;
		put_buf.ptr = put_buf.fp->current->vbuffer + put_buf.count;
		}
#else	/* VMFILE */
	if((put_buf.start != NULL) && (put_buf.fp != NULL) &&
		(put_buf.count != 0))
		{
		if(fwrite((char *) put_buf.start,1,(UINT) put_buf.count,
			put_buf.fp) != put_buf.count)
			{
			if(temp_path[0] != '\0')
				{
		fatal("Error writing temp file on path: ",temp_path);
				}
			else
				{
		fatal("Error writing temp file on current directory","");
				}
			}
		}
	put_buf.count = 0;
	put_buf.ptr = put_buf.start;
	put_buf.fp = mod_fp[module];
#endif	/* VMFILE */
	}

#if	VMFILE
VOID
alloc_vbuf()
	{
	extern	STRING		malloc();

	if(vbfree == NULL)
		{
		if((vbfree = (struct vbuf *) malloc(sizeof(struct vbuf))) ==
			NULL)
			{
			fatal("Out of vmfile memory", "");
			}
		vbfree->next = NULL;
		}
	if(put_buf.fp->current == NULL)
		put_buf.fp->first = vbfree;
	else
		{	
		put_buf.fp->current->next = vbfree;
		put_buf.fp->current->vcount = put_buf.count;
		}
	put_buf.fp->current = vbfree;
	vbfree->vcount = 0;
	put_buf.ptr = vbfree->vbuffer;
	put_buf.count = 0;
	vbfree = vbfree->next;
	put_buf.fp->current->next = NULL;
	}
#endif	/* VMFILE */

VOID
put_2b(value)
	int			value;
	{
	PUT_1B(value);
	PUT_1B(value >> 8);
	}

VOID
put_4b(value)
	SLONG			value;
	{
	PUT_1B((int) value);
	PUT_1B((int) (value >>= 8));
	PUT_1B((int) (value >>= 8));
	PUT_1B((int) (value >> 8));
	}

/************************************************************************/
/*									*/
/*	Routine:	skip_space					*/
/*									*/
/*	Skip whitespace in the parse buffer.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
skip_space()
	{
	while(isspace(*parse_ptr))
		parse_ptr++;
	}

/************************************************************************/
/*									*/
/*	Routine:	string_copy					*/
/*									*/
/*	Copy at most "n" characters from one string to another.  No	*/
/*	more characters are copied past the '\0'.  This is functionally	*/
/*	identical to "strncpy" without the zero padding and return	*/
/*	value.								*/
/*									*/
/*	Parameters:							*/
/*		"dest"	Points to the string destination.		*/
/*		"src"	Points to the string source.			*/
/*		"n"	The maximum # of characters to copy.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
string_copy(dest,src,n)
	STRING			dest;
	STRING			src;
	register	UINT	n;
	{
	while(n--)
		{
		if((*dest++ = *src++) == '\0')
			return;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	swap_temp					*/
/*									*/
/*	Exchange the current "module" temp file pointer and the "swap"	*/
/*	temp file pointer.  Both files are rewound.  This is done to	*/
/*	allow the old contents of a "module" file to be read and	*/
/*	processed by a "pass" which will produce the new version of the	*/
/*	temp file.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
swap_temp()
	{
	VSTREAM			swap;

#if	VMFILE
	register struct vbuf	*vb;

	while((vb = temp_fp->current) != NULL)
		{
		temp_fp->current = vb->next;
		vb->next = vbfree;
		vbfree = vb;
		}
	swap = temp_fp;                 /* Temporary for exchange */
	temp_fp = mod_fp[module];       /* Old module file */
	mod_fp[module] = swap;          /* New module file to fill */
	put_flush();                    /* Flush out any waiting output */
	temp_fp->current = temp_fp->first;
	get_buf.ptr = temp_fp->current->vbuffer;
	get_buf.count = temp_fp->current->vcount;
#else	/* VMFILE */
	swap = temp_fp;			/* Temporary for exchange */
	temp_fp = mod_fp[module];	/* Old module file */
	mod_fp[module] = swap;		/* New module file to fill */
	put_flush();			/* Flush out any waiting output */
	rewind(temp_fp);		/* Old module file */
	rewind(swap);			/* New module file to fill */
	get_buf.count = 0;		/* Flush out any waiting input */
#endif	/* VMFILE */
	}

/************************************************************************/
/*									*/
/*	Routine:	terminate					*/
/*									*/
/*	Erase temp files and return to the operating system.		*/
/*									*/
/*	Parameters:							*/
/*		"rval"	The desired "exit" value to return.		*/
/*									*/
/*	Returns:	None, and takes its time about it too.		*/
/*									*/
/************************************************************************/

VOID
terminate(rval)
	int			 rval;
	{
	char			buf1[FNSIZE + 1];
	char			buf2[FNSIZE + 1];
	char			buf3[FNSIZE + 1];

#if	DEBUG				/* Leave temp files during debugging */
	exit(rval);
#else	/* DEBUG */
#if	!VMFILE
/*
 *	Close all the temp files, starting with the "extra" one.  Note that we
 *	don't know which temp filename goes with which pointer since the
 *	assembly/linkage process swaps the "extra" temp file pointer for the
 *	others during normal operation.  Thus the error messages are a bit
 *	"bland".
 */
	if(temp_fp != NULL)		/* "extra" temp file opened? */
	    {
	    if(fclose(temp_fp) == EOF)
		{
		if(temp_path[0] != '\0')
		    {
		warning("Unable to close temp file on path: ",temp_path);
		    }
		else
		    {
		warning("Unable to close temp file on current directory","");
		    }
		}
	    }
/*
 *	Now close "module" temp files.
 */
	for(module = 0; module < NUMMODULES; module++)
	    {
	    if(mod_fp[module] != NULL)	/* See if temp file for this module */
		{
		if(fclose(mod_fp[module]) == EOF)
		    {
		    if(temp_path[0] != '\0')
			{
		warning("Unable to close temp file on path: ",temp_path);
			}
		    else
			{
		warning("Unable to close temp file on current directory","");
			}
		    }
		}
	    }
/*
 *	Now delete all the temp files.  The error messages here can be
 *	specific as to what file is involved.
 */
	strcpy(buf1,temp_path);
/*
 *	Delete the "extra" temp file.
 */
	if(temp_fp != NULL)		/* "extra" temp file opened? */
	    {
	    strcpy(buf2,buf1);
	    strcat(buf2,EXTRA_TEMP);
	    if(unlink(buf2) == -1)
		warning("Unable to remove temp file: ",buf2);
	    }
/*
 *	Now delete "module" temp files.
 */
	for(module = 0; module < NUMMODULES; module++)
	    {
	    if(mod_fp[module] != NULL)	/* See if temp file for this module */
		{
		strcpy(buf2,buf1);
		strcat(buf2,TEMP_NAME);
		sprintf(buf3,"%u",module);
		strcat(buf2,buf3);
		if(unlink(buf2) == -1)
		    warning("Unable to remove temp file: ",buf2);
		}
	    }
#endif	/* !VMFILE */
	exit(rval);			/* We are history */
#endif	/* DEBUG */
/*lint -e529 lint doesn't understand about #if requirements */
	}
/*lint +e529 lint doesn't understand about #if requirements */

/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect linker invocation to user with a suggestion	*/
/*	of how it should be done.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
usagerr()
	{
	fatal("Usage is 'tlnk [command_file]'","");
	}

/************************************************************************/
/*									*/
/*	Routine:	warning						*/
/*									*/
/*	Print out a "WARNING" error message.				*/
/*									*/
/*	Parameters:							*/
/*		"sptr1"	A pointer to the first of the error message.	*/
/*		"sptr2" A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
warning(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	printf("WARNING: %s%s\n",sptr1,sptr2);	/* Print message */
	if(list_fp != NULL)
		fprintf(list_fp,"WARNING: %s%s\n",sptr1,sptr2);
	}
