/************************************************************************/
/*									*/
/*	File:	tlnk4.c							*/
/*									*/
/*	Transputer Linker (TLNK), source file 4.  This file contains	*/
/*	the code which does most of the listing generation for TLNK.	*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character type functions */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tlnkdef.h"		/* Linker definitions */
#include	"tlnktyp.h"		/* Linker data declarations */
#include	"tlnkext.h"		/* Linker externals */

/************************************************************************/
/*									*/
/*	Routine:	list_st						*/
/*									*/
/*	List the global symbol table and other information.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

/*
 *	First a helping function for the sort routine call.
 */

static	int
symcmp(first,second)
	struct	gst		*first;
	struct	gst		*second;
	{
	SLONG			first_addr;
	SLONG			second_addr;

	if(byaddress)
		{
		if(first->lst_num >= MAX_SYMBOLS)	/* Was it defined? */
			first_addr = 0;		/* Nope */
		else
			first_addr = locals[first->lst_num].min_addr;
		if(second->lst_num >= MAX_SYMBOLS)	/* Was it defined? */
			second_addr = 0;	/* Nope */
		else
			second_addr = locals[second->lst_num].min_addr;
/*
 *	Note that the normal subtraction trick is not adequate here since the
 *	result will be of type SLONG and probably won't be accurately
 *	representable in a "int".
 */
		if((first_addr - second_addr) > (SLONG) 0)
			return (1);
		else if((first_addr - second_addr) < (SLONG) 0)
			return (-1);
		else
			return (0);
		}
	else
		return (strncmp(first->name,second->name,SYMNMESIZE));
	}

VOID
list_st()
	{
	register	UINT	columns;
	STRING			e_name;
	SLONG			e_value;
	register	UINT	i;
	register	UINT	j;
/*
 *	Get entrypoint info if needed before global symbol table gets mucked
 *	with.
 */
	if(entry_type != NUM_ENTRY)
		{
		e_name = globals[locals[entry_s].gst_num].name;
		e_value = get_min(entry_s);
		}
/*
 *	First coalesce the "hashed" global symbol table.
 */
	for(j = i = 0; i < MAX_GLOBALS; i++)
		{
		if((globals[i].name != NULL) && (i != j))
			{
			globals[j].name = globals[i].name;
			globals[j].lst_num = globals[i].lst_num;
			j++;
			}
		}
	if(j > 0)
		qsort((STRING) globals,j,sizeof(struct gst),symcmp);
/*
 *	Display for user.
 */
	if(sym_col != 0)		/* Cleanup trailing table stuff */
		{
		sym_col = 0;
		putc('\n',list_fp);
		}
	putc('\f',list_fp);
	lbanner();
	fprintf(list_fp,"\nSymbol table:\n\n");
	for(columns = i = 0; i < j; i++)
		{
		fprintf(list_fp,"%s\t",globals[i].name);
		if(globals[i].lst_num >= MAX_SYMBOLS)	/* Was it defined? */
			fprintf(list_fp,"**********");
		else
			{
			fprintf(list_fp,"0x%08lX",
				locals[globals[i].lst_num].min_addr);
			}
		if(++columns >= ST_LIST_COLS)
			{
			columns = 0;
			putc('\n',list_fp);
			}
		else
			putc('\t',list_fp);
		}
	if(columns != 0)
		putc('\n',list_fp);
/*
 *	Add info about the load address, stack address and entry point.
 */
	fprintf(list_fp,"\nLoad address:\t0x%08lX",load_addr);
	fprintf(list_fp,"\tStack address:\t0x%08lX\n",stack_addr);
	if(entry_type == NUM_ENTRY)
		fprintf(list_fp,"Entry point:\t0x%08lX\n",entry_n);
	else
		fprintf(list_fp,"Entry point [%s]:\t0x%08lX\n",e_name,e_value);
	}

/************************************************************************/
/*									*/
/*	Routine:	out_rec						*/
/*									*/
/*	Outputs the begining of a T_* record (the record type).  If	*/
/*	there is an accumulated T_DATA or T_STORAGE block, this is	*/
/*	output first.							*/
/*									*/
/*	Parameters:	T_* code.					*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
out_rec(op)
	UINT			op;
	{
	if(acc_dtasize > 0)
		{
		out_1b((int) T_DATA);
		out_2b((int) 0);
		out_2b((int) acc_dtasize);
		out_write((char *) acc_dta,acc_dtasize);
		acc_dtasize = 0;
		}
	if(acc_storesize > 0)
		{
		out_1b((int) T_STORAGE);
		out_2b((int) 0);
		out_4b((SLONG) acc_storesize);
		acc_storesize = 0;
		}
	if(op == T_RESERVED)
		return;
	out_1b((int) op);
	}

/************************************************************************/
/*									*/
/*	Routine:	pass3						*/
/*									*/
/*	Reads the final temp files and produces the linked output file	*/
/*	ready to be downloaded.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
pass3()
	{
	SLONG			big_temp;
	SLONG			local_bytes;
	SLONG			module_bytes;
	BOOL			new_load;
	BOOL			notdone;
	register	UINT	temp;

#if	DEBUG
	putchar('3');			/* Indicate doing pass 3 */
#endif
/*
 *	Output file identification record.
 */
	out_1b((int) T_LD_FILE);	/* Indicate linked/located */
	out_1b((int) cpu_type);		/* CPU type identifier */
/*
 *	Output default load address.
 */
	out_1b((int) T_LOAD);
	out_4b((SLONG) (min_addr = load_addr));
/*
 *	Output stack address.
 */
	out_1b((int) T_STACK);
	out_4b((SLONG) stack_addr);
/*
 *	Output entrypoint address.
 */
	out_1b((int) T_ENTRY);
	if(entry_type == SYM_ENTRY)
		out_4b((SLONG) get_min(entry_s));
	else
		out_4b((SLONG) entry_n);
/*
 *	Output title to listing file (if one is being generated).
 */
	if(list_fp != NULL)
		{
		if(sym_col != 0)	/* Cleanup trailing column stuff */
			{
			sym_col = 0;
			putc('\n',list_fp);
			}
		putc('\f',list_fp);
		lbanner();
		fprintf(list_fp,"\nModule load map:\n");
		}
/*
 *	Loop through all the records in all the module temp files.  Write those
 *	that are needed to the output file, toss the rest.
 */
	for(module = 0; module < NUMMODULES; module++)
		{
		if(mod_fp[module] != NULL)	/* If something in module */
			{
			if(list_fp != NULL)
				{
				fprintf(list_fp,"\nModule #%u:\n",module);
				new_load = TRUE;
				module_bytes = 0;	/* Bytes in module */
				}
			swap_temp();	/* Swap module & extra temp file */
			notdone = TRUE;	/* Until we need to find next module */
			while(notdone)
				{
				switch(temp = GET_1B())
					{
/* Normal data definition */		case T_DATA:
					if(list_fp != NULL)
						{
						if(new_load)
							{
							new_load = FALSE;
							local_bytes = 0;
		fprintf(list_fp,"\tStart:\t0x%08lX\t",min_addr);
							}
						}
					number = get_2b();
					temp = get_2b();
					min_addr += temp;
					module_bytes += temp;
					local_bytes += temp;
					if(compress)
						{
					if(acc_storesize)
						out_rec(T_RESERVED);
					while(temp > 0)
						{
						if(acc_dtasize >=
							ACCDATASIZE)
							{
							out_rec(T_RESERVED);
							}
						acc_dta[acc_dtasize++] =
							GET_1B();
						temp--;
						}
						}
					else
						{
						out_1b((int) T_DATA);
						out_2b((int) number);
						out_2b((int) temp);
						copy_out(temp);
						}
					break;

/* Label definition */			case T_DEF:
					get_4b();	/* Line #,symbol # */
					break;

/* Reserve space pseudo-op */		case T_STORAGE:
					if(list_fp != NULL)
						{
						if(new_load)
							{
							new_load = FALSE;
							local_bytes = 0;
		fprintf(list_fp,"\tStart:\t0x%08lX\t",min_addr);
							}
						}
					number = get_2b();	/* Line # */
					big_temp = get_4b();
					min_addr += big_temp;
					module_bytes += big_temp;
					local_bytes += big_temp;
					if(compress)
						{
						if(acc_dtasize > 0)
							out_rec(T_RESERVED);
						acc_storesize += big_temp;
						}
					else
						{
						out_1b((int) temp);
						out_2b((int) number);
						out_4b((SLONG) big_temp);
						}
					break;

/* Address reference in data */		case T_ADDR_DATA:
					if(list_fp != NULL)
						{
						if(new_load)
							{
							new_load = FALSE;
							local_bytes = 0;
		fprintf(list_fp,"\tStart:\t0x%08lX\t",min_addr);
							}
						}
					min_addr += (SLONG) word_size;
					module_bytes += (SLONG) word_size;
					local_bytes += (SLONG) word_size;
					out_rec(temp);
					out_2b((int) (number = get_2b()));
					out_2b((int) 0); /* Fake ST index # */
/* Get symbol table address value */	big_temp = get_min((UINT) get_2b());
/* Add address offset */		out_4b(big_temp + get_4b());
					break;

/* Filename record */			case T_FILENAME:
					if(compress)
						{
						(VOID) get_4b();/* Line #'s */
						temp = GET_1B();
						while(temp-- > 0)
							(VOID) GET_1B();
						}
					else
						{
						out_1b((int) temp);
						copy_out(4);	/* Line #'s */
						out_1b((int)
							(temp = GET_1B()));
						copy_out(temp);
						}
					break;

/* Non-symbolic debug data record */	case T_DEBUG_DATA:
					out_rec(temp);
					out_2b((int) (number = get_2b()));
					copy_out(4);	/* Value */
					out_2b((int) (temp = get_2b()));
					copy_out(temp);	/* Data */
					break;

/* Set load address instruction */	case T_LOAD:
					out_rec(temp);
					out_4b((SLONG) (big_temp = get_4b()));
					if(list_fp != NULL)
						{
						if((! new_load) &&
							(big_temp != min_addr))
							{
							new_load = TRUE;
		fprintf(list_fp,"End:\t0x%08lX\t",(min_addr - (SLONG) 1));
		fprintf(list_fp,"Size:\t0x%lX\n",local_bytes);
							}
						min_addr = big_temp;
						}
					break;

/* Found EOF record */			case T_EOF:
					if(acc_storesize || acc_dtasize)
						out_rec(T_RESERVED);
					notdone = FALSE;
					break;

					default:
					fatal("Corrupted temporary file","");
					}
				}
			if(list_fp)
				{
				if(module_bytes == (SLONG) 0)
					fprintf(list_fp,"Nothing in module\n");
				else
					{
					if(! new_load)
						{
		fprintf(list_fp,"End:\t0x%08lX\t",(min_addr - (SLONG) 1));
		fprintf(list_fp,"Size:\t0x%lX\n",local_bytes);
						}
		fprintf(list_fp,"Module #%u contained 0x%lX (%ld) bytes\n",
			module,module_bytes,module_bytes);
					}
				}
			}
		}
	out_1b((int) temp);		/* Write final EOF record */
	copy_out(2);			/* Line # */
	}
