/************************************************************************/
/*									*/
/*	File:	tlnkext.h						*/
/*									*/
/*	External declarations for the Transputer Linker (TLNK).		*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1987 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

/*
 *	I/O definitions.
 */

extern	char	ifnbuf[];		/* Input filename buffer */
extern	STREAM	in_fp;			/* Input file pointer */
extern	char	ofnbuf[];		/* Output filename buffer */
extern	STREAM	out_fp;			/* Output file pointer */
extern	char	listbuf[];		/* Listing filename buffer */
extern	STREAM	list_fp;		/* Listing file pointer */
extern	char	cmdfnbuf[];		/* Command file filename buffer */
extern	STREAM	cmdf_fp;		/* Command file file pointer */

extern	VSTREAM	mod_fp[];		/* Temp file ptrs for each "module" */
extern	VSTREAM	temp_fp;		/* "extra" temp file pointer */
extern	char	temp_path[];		/* Pathname for temp files */

#if	VMFILE
extern	struct	vbuf	*vbfree;	/* Virtual memory buffer free list */
#endif	/* VMFILE */

#if	IOBUFSIZE
extern	STRING	temp_fp_buf;		/* Stream buf ptr for "extra" file */
#endif

/*
 *	General definitions.
 */

extern	UCHAR	acc_dta[];		/* T_DATA pack accumulator */
extern	int	acc_dtasize;		/* T_DATA pack accumulated size */
extern	SLONG	acc_storesize;		/* T_STORAGE accumulated size */
extern	BOOL	aligned[];		/* TRUE if temp file is "aligned" */
extern	UINT	base_symbol;		/* Local ST start for current file */
extern	BOOL	byaddress;		/* TRUE/sort symbol table by address */
extern	BOOL	cfile;			/* TRUE if a command file being used */
extern	UINT	cmd_number;		/* Command file line # */
extern	BOOL	compress;		/* TRUE if to pack T_DATA/T_STORAGE */
extern	UINT	cpu_type;		/* Processor type not yet selected */
extern	SLONG	entry_n;		/* Entrypoint address if constant */
extern	UINT	entry_s;		/* Entrypoint LST entry # if symbol */
extern	UINT	entry_type;		/* Type of entrypoint selected */
extern	UINT	errors;			/* Total # of errors encountered */
extern	BOOL	expr_err;		/* TRUE if an expression error found */
extern	BOOL	first_ifile;		/* TRUE prior to first input file */

#if	H_OPSYS == MPW
extern	struct	gst	*globals;	/* Global symbol table */
extern	struct	lst	*locals;	/* Local symbol table */
#else	/* H_OPSYS != MPW */
#if	HUGE_LOCALST
extern	struct	gst	globals[];	/* Global symbol table */
extern	struct	lst	huge *locals;	/* Local symbol table */
#else	/* !HUGE_LOCALST */
extern	struct	gst	globals[];	/* Global symbol table */
extern	struct	lst	locals[];	/* Local symbol table */
#endif	/* HUGE_LOCALST */
#endif	/* H_OPSYS == MPW */

extern	BOOL	is_lib;			/* TRUE if reading from a library */
extern	SLONG	load_addr;		/* Default load address */
extern	SLONG	max_addr;		/* Maximum address (current module) */
extern	SLONG	max_addr_adj;		/* Max addr fwd ref. adjustment */
extern	UINT	max_i_size;		/* Max. instr. size in bytes */
extern	SLONG	min_addr;		/* Minimum address (current module) */
extern	SLONG	min_addr_adj;		/* Min addr fwd ref. adjustment */
extern	UINT	module;			/* Current module # */
extern	UINT	next_symbol;		/* Local ST end (+1 current file) */
extern	UINT	number;			/* Source file input line # */
extern	UINT	num_ext;		/* # of external symbols */
extern	BOOL	old_file[];		/* TRUE if mod. doesn't need align. */
extern	char	parse_buf[];		/* Input parsing buffer */
extern	STRING	parse_ptr;		/* Parsing buffer pointer */
extern	UINT	region;			/* Current code/data region tag */
extern	SLONG	stack_addr;		/* Default stack address */
extern	BOOL	static_link;		/* TRUE for static link data binding */
extern	UINT	sym_col;		/* Columns # for file symbols */
extern	UINT	udef_list;		/* Undefined external's list head */
extern	UINT	word_size;		/* Processor word size in bytes */

/*
 *	The following variables are used by "read_file" and associated helping
 *	routines.
 */

extern	char	fnbuf[];		/* Most recent given by T_FILENAME */
extern	UINT	fnsize;			/* Length of filename in "fnbuf" */
extern	UINT	new_ln;			/* Most recent given by T_FILENAME */
extern	UINT	old_ln;			/* Most recent given by T_FILENAME */
extern	BOOL	used_modules[];		/* TRUE if filename current */

/*
 *	Local buffering control stuff.
 */

extern	struct	lbuf	get_buf;	/* "get_X" buffer */
extern	struct	lbuf	put_buf;	/* "put_X" buffer */

/*
 *	Function definitions.
 */

extern	SLONG	adj_max_addr();
extern	SLONG	adj_min_addr();
extern	VOID	add_def_lst();
extern	VOID	add_lst();
extern	VOID	add_set_lst();
extern	UINT	add_symbol();

#if	VMFILE
extern	VOID	alloc_vbuf();
#endif	/* VMFILE */

extern	VOID	assemble();
extern	VOID	bad_command();
extern	BOOL	bad_order();
extern	VOID	banner();
extern	UINT	bind_constant();
extern	VOID	build_i();
extern	VOID	copy_in();
extern	VOID	copy_out();
extern	VOID	copy_temp();
extern	VOID	debug_addr();
extern	VOID	do_addr();
extern	VOID	do_diff();
extern	BOOL	do_i_addr();
extern	BOOL	do_i_diff();
extern	BOOL	do_i_wdiff();
extern	VOID	dor_ref();
extern	BOOL	dor_i_ref();
extern	VOID	dors_ref();
extern	BOOL	dors_i_ref();
extern	SLONG	eval();
extern	VOID	exit();
extern	VOID	fatal();
extern	VOID	fatali();
extern	STRING	getenv();
extern	BOOL	get_flags();
extern	SLONG	get_max();
extern	SLONG	get_min();
extern	int	get_next();
extern	UINT	get_region();
extern	SLONG	hexbin();
extern	UINT	i_size();
extern	VOID	in_seek();
extern	VOID	init_input();
extern	VOID	init_slb();
extern	VOID	interactive();
extern	BOOL	ishex();
extern	BOOL	isoct();
extern	VOID	lbanner();
extern	VOID	ldpi_addr();
extern	VOID	ldpi_build();
extern	UINT	ldpi_length();
extern	UINT	ldpi_size();
extern	BOOL	lib_file();
extern	VOID	list_st();
extern	UINT	lookup();

#if	H_OPSYS == VMS
extern	int	main();
#else	/* H_OPSYS == VMS */
extern	VOID	main();
#endif	/* H_OPSYS == VMS */

extern	VOID	maketdir();
extern	BOOL	match();
extern	BOOL	match_first();
extern	VOID	maybe_fatal();
extern	VOID	non_fatal();
extern	BOOL	not_fname();
extern	VOID	open_listing();
extern	VOID	open_output();
extern	VOID	open_temp();
extern	VOID	out_rec();
extern	VOID	out_write();
extern	VOID	out_1b();
extern	VOID	out_2b();
extern	VOID	out_4b();
extern	VOID	parse();
extern	VOID	pass0();
extern	VOID	pass2();
extern	VOID	pass3();
extern	VOID	put_flush();
extern	VOID	put_max();
extern	VOID	put_min();
extern	VOID	put_rec();
extern	VOID	put_region();
extern	VOID	put_2b();
extern	VOID	put_4b();
extern	VOID	qsort();
extern	BOOL	read_file();
extern	VOID	skip_space();
extern	BOOL	sort_udef();
extern	VOID	string_copy();
extern	VOID	swap_temp();
extern	VOID	terminate();
extern	VOID	update_addr();
extern	VOID	update_fname();
extern	UINT	update_length();
extern	VOID	usagerr();
extern	VOID	warning();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

extern	BOOL	addr_pad();
extern	int	get_2b();
extern	SLONG	get_4b();
extern	BOOL	get_entry();
extern	BOOL	get_fname();
extern	BOOL	get_num();
extern	int	in_next();
extern	int	in_1b();
extern	int	in_2b();
extern	SLONG	in_4b();
extern	BOOL	input_files();
extern	BOOL	library_files();
extern	BOOL	makefn();
extern	BOOL	pass1();

/*lint -fvr */
