/*
 *	TCX Compiler (10 of 11)
 *
 *	Copyright 1987-1988 by Quality Computer Systems and Logical Systems.
 *	All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "tcx.h"
#include "tcxdecl.h"

/*
 *	Initialization code.
 */
PUBLIC	VOID
init()
	{
	extern	VOID		initkeywords();

	if(Verbose)
		fprintf(STDERR,"%s%s\n%s\n\n",Signon,Version,Copyright);
	initkeywords();
	}


/*
 *	Get options from command line.
 */
PUBLIC	VOID
getoptions(argc,argv)
  char *argv[];
    {
    extern VOID		getfilename();
    extern VOID		segment();
    extern char		*strchr();
    extern char		*strncpy();
    extern char		*strrchr();
    extern VOID		usagerr();

    register char	*p;
    static int		rename = 0;

    SegMap[SEG_CODE]   = SEG_CODE;
    SegMap[SEG_UNINIT] = SEG_CODE;
    SegMap[SEG_INIT]   = SEG_CODE;

    if (argc < 2)
	usagerr();		/* Need at least one input file */
    Nextfil = argv + 1;
    while ((--argc != 0) && (**++argv != '-'))
	++Ninfils;

    if (Ninfils == 0)
	usagerr();	/* Must have an input file */
    getfilename(*Nextfil, Infil, "pp");	/* Default is PP files */
    Newfile = TRUE;
    while (argc--)
	{				/* Look for options */
	p = *argv++;
	if (*p == '-')
	    {
	    while (*++p)
		{
		switch((int) *p)
		    {
		case 'C':
		case 'c':
		    Source_info = FALSE;
		    continue;
#ifdef	DEBUG
		case 'D':
		case 'd':		/* Enable debugging */
		    Debug = TRUE;
		    continue;
#endif	/* DEBUG */
		case 'E':
		case 'e':		/* Don't delete output file if errs */
		    Errdel = FALSE;
		    continue;
		case 'F':
		case 'f':		/* Set floating point mode */
		    switch((int) *++p)
			{
		    case '0':
			Float_mode = FLOAT_PREANSI;
			break;
		    case '1':
			Float_mode = FLOAT_ANSI;
			break;
		    case '2':
			Float_mode = FLOAT_SINGLE;
			break;
		    default:
			usagerr();
			}
		    continue;
		case 'I':		/* Specify inter-pass external file */
		case 'i':
		    Inter_file = TRUE;
		    continue;
		case 'M':		/* Set segments	*/
		case 'm':
		    ++p;		/* Skip param char */
		    switch((int) *p)
			{
		    case 'C':
		    case 'c':	SegMap[SEG_CODE] = atoi(&p[1]);
				break;
		    case 'U':
		    case 'u':	SegMap[SEG_UNINIT] = atoi(&p[1]);
				break;
		    case 'I':
		    case 'i':	SegMap[SEG_INIT] = atoi(&p[1]);	
				break;
		    default:	usagerr();
				break;
			}
		    p = " ";		/* Ignore the rest of the arg */
		    continue;
		case 'O':
		case 'o':		/* Rename output */
		    rename = 1;
		    continue;
		case 'P':
		case 'p':		/* Set CPU type */
		    switch((int) *++p)
			{
		    case '0':
			Cpu_type = TALL;
			break;
		    case '2':
			Sizeof_int = WORD_SIZE_16B;
			Sizeof_long = WORD_SIZE_16B;
			Sizeof_ptr = WORD_SIZE_16B;
			Sizeof_short = WORD_SIZE_16B;
			if(*++p == '5')
				Cpu_type = T2EX;
			else
				{
				p--;
				Cpu_type = T212;
				}
			break;
		    case '4':
			if(*++p == '5')
				Cpu_type = T4EX;
			else
				{
				p--;
				Cpu_type = T414;
				}
			break;
		    case '8':
			if(*++p == '5')
				Cpu_type = T8EX;
			else
				{
				p--;
				Cpu_type = T800;
				}
			break;
		    case 'c':
		    case 'C':
			Cpu_schar = TRUE;	/* Chars are signed */
			break;
		    case 's':
		    case 'S':
			Cpu_srshift = TRUE;	/* >> is signed */
			break;
		    default:
			usagerr();
			}
		    continue;
		case 'Q':
		case 'q':		/* Disable selected optimizations */
		    switch((int) *++p)
			{
		    case '0':
			Peepopt = FALSE;
			break;
		    case '1':
			Builtin = FALSE;
			break;
		    case '2':
			FP_reorder = FALSE;
			break;
		    default:
			usagerr();
			}
		    continue;
		case 'R':		/* Force relocatable only code */
		case 'r':
		    Rel_only = TRUE;
		    continue;
		case 'S':		/* Use static link data model */
		case 's':
		    Static_link = TRUE;
		    Rel_only = TRUE;	/* Implied by static link option */
		    SegMap[SEG_UNINIT] = (SegMap[SEG_INIT] = SLBMODULE) + 2;
		    continue;
		case 'T':		/* Set trace-in/trace-out functions */
		case 't':
		    switch((int) *++p)
			{
		    case 'i':
			Tracein = TRUE;
			strncpy(Triname,++p,MAXLABEL);
			Triname[MAXLABEL] = '\0';
			break;
		    case 'o':
			Traceout = TRUE;
			strncpy(Troname,++p,MAXLABEL);
			Troname[MAXLABEL] = '\0';
			break;
		    default:
			usagerr();
			}
		    break;
		case 'V':
		case 'v':		/* Disable verbose */
		    Verbose = !Verbose;
		    continue;
		case 'W':
		case 'w':		/* Set warning level */
		    switch((int) *++p)
			{
		    case '0':
			Warn_level = 0;	/* No warnings */
			break;
		    case '1':
			Warn_level = 1;	/* Medium picky */
			break;
		    case '2':
			Warn_level = 2;	/* Maximally picky */
			break;
		    default:
			usagerr();
			}
		    continue;
		default:
		    usagerr();
		    }
		break;
		}
	    }
	else if (rename == 1)
	    {				/* Should be new output file name */
#if H_OPSYS == VMS			/* Only a directory or device? */
	    if (p[strlen(p) - 1] == ']' || p[strlen(p) - 1] == ':')
		{
		strcpy(Coutfil, p);	/* Save output dir */
		strcat(Coutfil, ((strrchr(Infil, ']') == NULL) ?
			((strrchr(Infil, ':') == NULL) ?
				Infil : strrchr(Infil, ':') + 1) :
			strrchr(Infil, ']') + 1));
		*(strrchr(Coutfil, '.') + 1) = '\0';
		strcat(Coutfil, Coutext);
		}
#endif
#if ((H_OPSYS == UNIX) || (H_OPSYS == MPW) || (H_OPSYS == BSD4_2) || (H_OPSYS == XENIX)) /* dir? */
	    if (p[strlen(p) - 1] == '/')
		{
		strcpy(Coutfil, p);	/* Save output dir */
		strcat(Coutfil, strrchr(Infil, '/') == NULL ?
			Infil :		/* Append infil.ext */
			strrchr(Infil,'/') + 1);
		*(strchr(strrchr(Coutfil,'/'),'.') + 1) = '\0';
		strcat(Coutfil, Coutext);
		}
#endif
#if H_OPSYS == MS_DOS			/* Only a drive or path? */
	    if (p[strlen(p)-1] == '\\' || p[strlen(p)-1] == ':')
		{
		strcpy(Coutfil, p);	/* Save drive/path */
		strcat(Coutfil, ((strrchr(Infil, '\\') == NULL) ?
			((strrchr(Infil, ':') == NULL) ?
				Infil :
				strrchr(Infil, ':') + 1) :
			strrchr(Infil, '\\') + 1));
		*(strchr(Coutfil, '.') + 1) = '\0';
		strcat(Coutfil, Coutext);
		}
#endif
	    else
		{
		getfilename(p, Coutfil, Coutext);
	/* Use output file .ext for library extension as well */
#if ((H_OPSYS == UNIX) || (H_OPSYS == BSD4_2))
		strcpy(Coutext, 
			strchr(((strrchr(Coutfil,'/') == NULL) ?
			Coutfil : strrchr(Coutfil, '/') + 1),'.') + 1);
#else
		strcpy(Coutext, strrchr(Coutfil, '.') + 1);
#endif
		}
	    if (streq(Infil, Coutfil))
		{
		fprintf(STDERR, "Output file name same as input\n");
		exit(BADEXIT);
		}
	    rename = 2; 		/* Remember we did it */
	    }
	else
	    usagerr();
	}
    if ((!rename) && (Coutput == NULL))
	{				/* Use Infil.defext for code output */
	strcpy(Coutfil, Infil);
#if ((H_OPSYS == UNIX) || (H_OPSYS == BSD4_2))
	*(strchr(((strrchr(Coutfil, '/') == NULL) ?
			Coutfil :
			strrchr(Coutfil, '/')),'.') + 1) = '\0';
#else
	*(strchr(Coutfil, '.')+1) = '\0'; /* End name after '.' */
#endif
	strcat(Coutfil, Coutext); /* Add default extension */
	}
    else if (rename == 1)
	usagerr();			/* New file name not found */

    segment(SEG_NONE);			/* Initialize to "none yet" */
    }


PRIVATE	VOID
usagerr()
	{
fprintf(STDERR,
"FATAL: General usage is 'tcx input_filename [options]* [output_filename]'\n");
fprintf(STDERR,"\tCompress output file (remove debugging info)   '-c'\n");
fprintf(STDERR,"\tEnable TCX internal debugging mode             '-d'\n");
fprintf(STDERR,"\tDon't delete output file on errors             '-e'\n");
fprintf(STDERR,"\tChoose floating pt mode: Force float to double '-f0'\n");
fprintf(STDERR,"\t                         Normal ANSI           '-f1' (default)\n");
fprintf(STDERR,"\t                         Single precision      '-f2'\n");
fprintf(STDERR,"\tUse temp file for internal data to save memory '-i'\n");
fprintf(STDERR,"\tSelect module number for data types (# is 0-255):\n");
fprintf(STDERR,"\t                         Code                  '-mc#'\n");
fprintf(STDERR,"\t                         Uninitialized data    '-mu#'\n");
fprintf(STDERR,"\t                         Initialized data      '-mi#'\n");
fprintf(STDERR,"\tSpecify different output filename              '-o'\n");
fprintf(STDERR,"\tSelect compiler options: Compile for T4XX/T8XX '-p0'\n");
fprintf(STDERR,"\t                         Compile for T212/T222 '-p2'\n");
fprintf(STDERR,"\t                         Compile for T225      '-p25'\n");
fprintf(STDERR,"\t                         Compile for T414      '-p4' (default)\n");
fprintf(STDERR,"\t                         Compile for T400/T425 '-p45'\n");
fprintf(STDERR,"\t                         Compile for T800      '-p8'\n");
fprintf(STDERR,"\t                         Compile for T801/T805 '-p85'\n");
fprintf(STDERR,"\t                         Normal 'char' signed  '-pc'\n");
fprintf(STDERR,"\t                         Shifts propagate sign '-ps'\n");
fprintf(STDERR,"\tDisable peephole optimization                  '-q0'\n");
fprintf(STDERR,"\tDisable intrinsic function substitution        '-q1'\n");
fprintf(STDERR,"\tDisable FPU concurrency optimization           '-q2'\n");
fprintf(STDERR,"\tGenerate relocatable code & data               '-r'\n");
fprintf(STDERR,"\tGenerate separately relocatable code & data    '-s'\n");
fprintf(STDERR,"\tSpecify trace-in function for function entry   '-ti[name]'\n");
fprintf(STDERR,"\tSpecify trace-out function for function exit   '-to[name]'\n");
#if	H_OPSYS != MPW
fprintf(STDERR,"\tDisable verbose output mode                    '-v'\n");
#else
fprintf(STDERR,"\tEnable verbose output mode                     '-v'\n");
#endif	/* !MPW */
fprintf(STDERR,"\tSelect warning level:    No warnings           '-w0'\n");
fprintf(STDERR,"\t                         Moderate nitpicking   '-w1'\n");
fprintf(STDERR,"\t                         Full nitpicking       '-w2' (default)\n");

	exit(BADEXIT);
	}

/*
 *	Obtain compiler table and buffer space from free space.
 */
PUBLIC	VOID
getspace()
	{
	extern	VOID		fatal();
	extern	char		*malloc();

	if((Line = malloc(LINESIZE)) == NULL)
		fatal("Cannot allocate enough memory");
	Casestack = NULL;
	Swqueue = NULL;
	}

/*
 *	Initkeywords
 *
 *	This routine loads the C reserved words into the symbol table.
 *	We use C reserved words in several different ways, so they
 *	are loaded in under several different name spaces.
 *
 *	Several pairs of these funny tables really should be combined into
 *	a structure.  It would be less error-prone, as well as more readable.
 *		DCR		6 July 1987
 */
PRIVATE	VOID
initkeywords()
	{
	extern	struct	id_node	*add_fixed();

	static	char		*keywords[] =
		{
		"break", "case", "continue", "default", "do", "else", "entry",
		"enum", "for", "goto", "if", "return", "sizeof", "switch",
		"while", NULL
		};

	static	char		*ctypes[] =
		{
		"int",  "char",  "unsigned",  "signed",   "struct",  "union",
		"enum",	"short", "long", "void",
		NULL
		};
	static	INT32		ctval[] =
		{
		INT_NODE,  CHAR_NODE, UNSIGNED_NODE, SIGNED_NODE, STRUCT_NODE,
		UNION_NODE, ENUM_NODE, SHORT_NODE, LONG_NODE, VOID_NODE
		};

	static	char		*ctypefps[] =
		{
		"float", "double", NULL
		};
	static	INT32		ctfpval[] =
		{
		FLOAT_NODE, DOUBLE_NODE
		};

	static	char		*operators[] =
		{
		"+=", "|=", "^=", "&=", "*=", "-=", "=", ">>=", "<<=", "/=",
		"%=", NULL
		};
	static	INT32		opvalues[] =
		{
		OP_AADD, OP_ABOR, OP_ABXOR, OP_ABAND, OP_AMUL, OP_ASUB,
		OP_ASSIGN, OP_ARSHIFT, OP_ALSHIFT, OP_ADIV, OP_AMOD
		};

	static	char		*storagecls[] =
		{
		"register", "static", "extern",  "typedef",  "auto", NULL
		};
	static	INT32		clsval[] =
		{
		SC_REG, SC_STATIC, SC_EXTERN, SC_TYPEDEF, SC_AUTO
		};

	static	char		*ifuncs[] =
		{
"_BUILTIN_ajw",		"_BUILTIN_eval",	"_BUILTIN_fabs",
"_BUILTIN_fabsf",	"_BUILTIN_sqrt",	"_BUILTIN_sqrtf",
"_BUILTIN_memcpy",	"_BUILTIN_ChanIn",	"_BUILTIN_ChanInChar",
"_BUILTIN_ChanInInt",	"_BUILTIN_ChanOut",	"_BUILTIN_ChanOutChar",
"_BUILTIN_ChanOutInt",	"_BUILTIN_ChanReset",	"_BUILTIN_ProcAfter",
"_BUILTIN_ProcGetPriority", "_BUILTIN_ProcReschedule", "_BUILTIN_ProcStop",
"_BUILTIN_ProcWait",	"_BUILTIN_SetTime",	"_BUILTIN_Time",
"_BUILTIN_SetLoPriQ",	"_BUILTIN_SetHiPriQ",	"_BUILTIN_GetLoPriQ",
"_BUILTIN_GetHiPriQ",	"_BUILTIN_BitCnt",	"_BUILTIN_BitRevNBits",
"_BUILTIN_BitRevWord",	"_BUILTIN_Move2D",	"_BUILTIN_Move2DNonZero",
"_BUILTIN_Move2DZero",	"_BUILTIN_PHalt",	"_BUILTIN_PRun",
NULL
		};
	static	INT32		ifuncsval[] =
		{
		BI_ajw,		BI_eval,	BI_fabs,
		BI_fabsf,	BI_sqrt,	BI_sqrtf,
		BI_memcpy,	BI_ChanIn,	BI_ChanInChar,
		BI_ChanInInt,	BI_ChanOut,	BI_ChanOutChar,
		BI_ChanOutInt,	BI_ChanReset,	BI_ProcAfter,
		BI_ProcGetPriority, BI_ProcReschedule, BI_ProcStop,
		BI_ProcWait,	BI_SetTime,	BI_Time,
		BI_SetLoPriQ,	BI_SetHiPriQ,	BI_GetLoPriQ,
		BI_GetHiPriQ,	BI_BitCnt,	BI_BitRevNBits,
		BI_BitRevWord,	BI_Move2D,	BI_Move2DNonZero,
		BI_Move2DZero,	BI_PHalt,	BI_PRun
		};

	char			**loop;
	INT32			*valloop;

	for(loop = keywords; *loop != NULL; loop++)
		(VOID) add_fixed(*loop,NS_CKEYWORD,(INT32) strlen(*loop));

	for(loop = ctypes, valloop = ctval; *loop != NULL; )
		(VOID) add_fixed(*loop++,NS_CTYPES,*valloop++);

	if(! (Cpu_type & (T212 | T2EX)))
		{
		for(loop = ctypefps, valloop = ctfpval; *loop != NULL; )
			(VOID) add_fixed(*loop++,NS_CTYPES,*valloop++);
		}

	for(loop = storagecls, valloop = clsval; *loop != NULL; )
		(VOID) add_fixed(*loop++,NS_CSTORAGE,*valloop++);

	for(loop = operators, valloop = opvalues; *loop != NULL; )
		(VOID) add_fixed(*loop++,NS_CKEYWORD,*valloop++);

	for(loop = ifuncs, valloop = ifuncsval; *loop != NULL; )
		(VOID) add_fixed(*loop++,NS_IFUNC,*valloop++);
	}

/* End of TCX10.C */
