/*
**	TCX Compiler (3 of 11)
**
**	Copyright 1987-1988 by Quality Computer Systems and Logical Systems.
**	All rights reserved.
**
**	Published by:	Logical Systems
**			P.O. Box 1702
**			Corvallis, OR 97339 (USA)
**
**	05/7/87
**
**	TCX general statement processor
**
**	Note:  The statement routine causes the Lattice C compiler to get
**	a stack overflow.  If your compiler blows up, this may be the routine.
**
**	Note that a function body and a compound statement are very much alike.
**	The major differences are handled in the "prologue" and "epilogue"
**	routines elsewhere.
*/

#include <stdio.h>
#include <ctype.h>
#include "tcx.h"
#include "tcxdecl.h"


/*
 *	comp_block
 *
 *	Process compound statements.
 */
PUBLIC VOID
comp_block(locptr)
  INT32 locptr;		/* Current stack offset for local vars */
    {
    extern VOID			block();
    extern struct id_node	*enter_scope();
    extern VOID			exit_scope();

    struct id_node	*old_scope;	/* Scope of symbol list upon entry */

    Ncmp++;				/* Entered a new bracket level	*/
    old_scope = enter_scope();		/* Save old scope		*/
    block(locptr);
    exit_scope(old_scope);		/* Remove local symbols, etc.	*/
    }


/*
 * Block -- Parse a block
 */
PUBLIC VOID
block(locptr)
	INT32 locptr;		/* Current stack offset for local vars */
	{
	extern	INT32		declarations();
	extern	BOOL		Eof;
	extern	BOOL		matchc();
	extern	VOID 		statement();

	locptr = declarations(locptr);	/* Do block level declarations	*/

	while((! matchc('}')) && (! Eof))
		{
		statement(locptr);
		blanks();
		}
	if(! Eof)
		Ncmp--;
	}


/*
 *	statement
 *
 *	Process a single C statement.
 */
PRIVATE VOID
statement(locptr)
  INT32 locptr;
    {
    extern int		ch();
    extern BOOL		chklabel();
    extern VOID		comp_block();
    extern VOID		dobreak();
    extern VOID		docase();
    extern VOID		docontinue();
    extern VOID		dodefault();
    extern VOID		dodowhile();
    extern VOID		doexpr();
    extern VOID		dofor();
    extern VOID		dogoto();
    extern VOID		doif();
    extern VOID		doreturn();
    extern VOID		doswitch();
    extern VOID		dowhile();
    extern VOID		error();
    extern BOOL		Eof;
    extern BOOL		matchc();
/*
 *	Note that labels may appear where other statement type are allowed but
 *	they do not alter the flow of control.  Thus the following "while"
 *	loop (this could also be done by having "chklabel" call "statement",
 *	but why waste the stack space...)
 */
    while(chklabel())
	;

    if ((ch() == 0) && Eof)
	return;

    if (matchc('{'))			comp_block(locptr);
    else if (amatch("break"))		dobreak();
    else if (amatch("case"))		docase();
    else if (amatch("continue"))	docontinue();
    else if (amatch("default"))		dodefault();
    else if (amatch("do"))		dodowhile(locptr);
    else if (amatch("else"))		error("else not matched with if");
    else if (amatch("for"))		dofor(locptr);
    else if (amatch("goto"))		dogoto();
    else if (amatch("if"))		doif(locptr);
    else if (amatch("return"))		doreturn();
    else if (amatch("switch"))		doswitch(locptr);
    else if (amatch("while"))		dowhile(locptr);
    else if (matchc(';'))		/* donothing() */ ;
    else				doexpr();
    }


/*
 *	chklabel
 *
 *	Check for a LABEL: on a line.
 */
PRIVATE BOOL
chklabel()
    {
    extern struct id_node	*add_label();
    extern BOOL			chkeyword();
    extern char			*chksym();
    extern LABEL		getlabel();
    extern VOID			keyerr();   /* Display keyword error message */
    extern VOID			printlabel();

    char			labname[NAMEMAX+1];
    register char		*p;
    struct id_node		*symbol;

    if (((p = chksym(labname)) == NULL) || chks("default"))
	return (FALSE);

    while (isspace(*p))
	++p;			/* Skip leading spaces */

    /* Note that this allows white space, but not newlines.  Nasty.
     *		DCR		25 May 87				*/
    if (*p != ':')
	return (FALSE);

    if (chkeyword())
	keyerr(labname);	/* Invalid keyword usage		*/

    Lptr = p + 1;		/* Advance past the label		*/

    if ((symbol = add_label(labname, getlabel(), TRUE)) == NULL)
	multidef(labname);
    printlabel(symbol->val.label);

    return (TRUE);
    }


/*
 *	Do local variable declarations for both function level stuff
 *	and block level variables.
 */

PRIVATE	INT32	help_locptr = 0;/* Helper variable used only by "declarations"
				 * and "decl_help".			*/

PRIVATE INT32
declarations(locptr)
  INT32	locptr;
    {
    extern VOID			decl_auto();
    extern VOID			decl_inits();
    extern VOID			decl_noninits();
    extern VOID			decl_regs();
    extern VOID			expr_out();
    extern VOID			instruction();
    extern BOOL			isdecl();
    extern VOID			loadconstant();
    extern VOID			scan_ripup();
    extern INT32		word();

    int				sc;
    struct id_node		*type;

    Initlist = NULL;	/* Nothing in initialization list		*/

    while (isdecl(&sc, &type, Inttype))
	{
	/* Process local declarations					*/
	/* if storage class not set, use default			*/
	declloc(((sc == SC_NONE) ? SC_AUTO : sc), type);
	}

    /*
     * Now we go through all of the local symbols we just declared, and
     * assign offsets to them starting with locptr (in terms of bytes
     * rounded to words).
     */
    help_locptr = locptr;

    scan_ripup(decl_inits);		/* Initializers			*/
    scan_ripup(decl_noninits);		/* Non-initialized aggregates	*/
    scan_ripup(decl_auto);		/* All other non-register autos	*/
    scan_ripup(decl_regs);		/* Registers			*/

    if (-help_locptr > Autospace)	/* Remember the high water mark */
	Autospace = -help_locptr;

    /*
     * Output initialization expressions.
     */
    expr_out(Initlist, FALSE);

    return (help_locptr);
    }


/*
 *	decl_auto
 *
 *	Assign offset for AUTO variables.
 */
PRIVATE VOID
decl_auto(sym)
  struct id_node *sym;
    {
    extern INT32 decl_help();

    help_locptr = decl_help(sym, SC_AUTO, help_locptr);
    }


/*
 *	decl_inits
 *
 *	Assign offset to initialized aggregates.
 */
PRIVATE VOID
decl_inits(sym)
  struct id_node *sym;
    {
    extern INT32 decl_aggreg();

    help_locptr = decl_aggreg(sym, /* init = */ TRUE, help_locptr);
    }


/*
 *	decl_noninits
 *
 *	Assign offset for non-initialized aggregates.
 */
PRIVATE VOID
decl_noninits(sym)
  struct id_node *sym;
    {
    extern INT32 decl_aggreg();

    help_locptr = decl_aggreg(sym, /* init = */ FALSE, help_locptr);
    }


/*
 *	decl_regs
 *
 *	Assign offset for Register variables.
 */
PRIVATE VOID
decl_regs(sym)
  struct id_node *sym;
    {
    extern INT32 decl_help();

    help_locptr = decl_help(sym, SC_REG, help_locptr);
    }

/*
 *	decl_help
 *
 *	This routine is called once for every variable in the local scope.
 *	We must assign offsets to all variables which are of AUTO storage
 *	class and don't have an offset already applied to them.  The current
 *	offset is passed and the updated value is returned.
 */
PUBLIC	INT32
decl_help(sym,sc,offset)
	struct	id_node		*sym;
	int			sc;	/* Storage class to address	*/
	INT32			offset;	/* Offset to use		*/
	{
	extern	VOID		post_symbol();

	INT32			need;	/* Number of words needed	*/

	if((! (sym->type & OFFSET_ASSIGNED)) &&
		(sym->info.storage_class == sc))
		{
		sym->type |= OFFSET_ASSIGNED;
		need = ((sym->val.integral + Sizeof_int - 1) / Sizeof_int) *
			Sizeof_int;
		sym->address.offset = offset - need;
		post_symbol(sym);	/* Opt. write symbol debug info	*/
		}
	else
		need = 0;

	return (offset - need);
	}

/*
 *	Decl_aggreg
 *
 *	This routine is called once for every variable in the local scope.
 *	We must assign offsets to all variables which are of AUTO storage
 *	class and don't have an offset already applied to them.  The current
 *	offset is passed and the updated value is returned. 
 *
 *	The "init" parameter tells whether to assign storage to initialized
 *	aggregates or to any aggregate.  An aggregate is any item that is
 *	is an ARRAY_NODE, STRUCT_NODE or UNION_NODE.
 */
PUBLIC	INT32
decl_aggreg(sym,init,offset)
	struct	id_node		*sym;
	BOOL			init;	/* TRUE/looking for init. vars	*/
	INT32			offset;	/* Offset to use		*/
	{
	INT32	need;			/* Number of words needed	*/

	if((! (sym->type & OFFSET_ASSIGNED))			&&
			((sym->info.storage_class == SC_AUTO) ||
	(sym->info.storage_class == SC_REG))			&&
	(sym->type & (ARRAY_NODE | STRUCT_NODE | UNION_NODE))	&&
	(! init))
		{
		sym->type |= OFFSET_ASSIGNED;
/*
 *	The access to "val.integral" should probably be an access to the
 *	type pointer information.  In general we need to examine where the
 *	size information is getting stored (type versus value nodes).  Part of
 *	TCX uses one, part uses the other.  KAB & DCR 4/10/88.
 */
		need = ((sym->val.integral + Sizeof_int - 1) / Sizeof_int) *
			Sizeof_int;

		sym->address.offset = offset - need;
		post_symbol(sym);	/* Opt. write symbol debug info	*/
		}
	else
		need = 0;

	return (offset - need);
	}

/*
 *	dobreak
 *
 *	Process the break statement.
 */
PRIVATE	VOID
dobreak()
    {
    extern VOID		error();
    extern VOID		instruction();
    extern VOID		ns();
    extern char		*strlabel();
    extern struct swq	*Swqueue;		/* Ptr to sw/loop queue	*/

    char		lblbuf[NAMEMAX+1];

    if ((Swqueue != NULL) && (Swqueue->exit != (LABEL) 0))
	instruction(TOP_J, REL, strlabel(Swqueue->exit, lblbuf),
							NULL, (INT32) 0, NULL);
    else
	error("No active switch or loop");

    ns();
    }


/*
 *	docase
 *
 *	Process case labels.
 */
PRIVATE	VOID
docase()
    {
    extern VOID			addcase();
    extern VOID			error();
    extern struct op_node	*expression();
    extern VOID			free_op_node();
    extern LABEL		getlabel();
    extern VOID			needpunc();
    extern VOID			printlabel();

    struct op_node		*case_expression;
    LABEL			label;

    case_expression = expression();
    needpunc(':');

    if(!ISCONSTANT(case_expression))
	error("Case label is not a constant");
    else if(!ISINTEGRAL(case_expression))
	error("Case label is not integral");
    else
	{
	/* Case label is a constant integral (comp w/ int) */
/*
 * ANSI says case expression is (cast) to type of switch expression.  We
 * don't need to do anything here for Transputer, but think about this
 * for other processors.  7-Jul-87 GO
 */
	addcase(case_expression->right.ptr.terminal->val.integral,
		label = getlabel());
	printlabel(label);
	}
    free_op_node(case_expression);
    }


/*
 *	docontinue
 *
 *	Process continue statement.
 */
PRIVATE VOID
docontinue()
    {
    extern VOID		error();
    extern VOID		instruction();
    extern struct swq	*Swqueue;
    extern VOID		ns();
    extern char		*strlabel();

    char		lblbuf[NAMEMAX+1];

    if ((Swqueue != NULL) && (Swqueue->loop != (LABEL) 0))
	instruction(TOP_J, REL, strlabel(Swqueue->loop, lblbuf), NULL,
							(INT32) 0, NULL);
    else
	error("No active loop statements");

    ns();
    }


/*
 *	dodefault
 *
 *	Process default statement
 */
PRIVATE VOID
dodefault()
    {
    extern struct swheader *Casestack;
    extern VOID		error();
    extern LABEL	getlabel();
    extern VOID		needpunc();
    extern VOID		printlabel();

    needpunc(':');

    if (Casestack != NULL)
	{
	if (Casestack->sh_default != (LABEL) 0)
	    error("Only one default label permitted");
	else
	    printlabel(Casestack->sh_default = getlabel());
	}
    else
	error("No active switch statement");
    }


/*
 *	dodowhile
 *
 *	Process the do..while statement.
 */
PRIVATE VOID
dodowhile(locptr)
  INT32 locptr;
    {
    extern VOID			addswq();
    extern VOID			delswq();
    extern VOID			dvrt_append();
    extern VOID			dvrt_stack();
    extern struct dvrt		*dvrt_unstack();
    extern VOID			error();
    extern VOID			expr_out();
    extern VOID			free_op_node();
    extern VOID			gen_while();
    extern LABEL		getlabel();
    extern VOID			instruction();
    extern VOID			ns();
    extern VOID			printlabel();
    extern char			*strlabel();
    extern struct op_node	*test_expression();

    char			lblbuf[NAMEMAX+1];
    LABEL			loop_top;
    struct swq			while_queue;	/* element for sw/loop queue */
    struct op_node		*while_expression;
    struct dvrt			*while_body;

    loop_top = getlabel();
    while_queue.loop = getlabel();
    while_queue.exit = getlabel();	/* Add labels to queue element */

    addswq(&while_queue);		/* Add this level to while queue */

    dvrt_stack();
    statement(locptr);			/* Capture body of while body */
    while_body = dvrt_unstack();

    if (!match("while"))
	error("Missing while in dowhile");

    while_expression = test_expression();

    printlabel(loop_top);
    dvrt_append(while_body);
    printlabel(while_queue.loop);
/*
 *	Check for constant "dowhile" expression.
 */
    if(! ISCONSTANT(while_expression))
	{
	expr_out(while_expression,NEED_VALUE | RELAX_BOOL);
	instruction(TOP_CJ, REL, strlabel(while_queue.exit, lblbuf), NULL,
							(INT32) 0, NULL);
	
	instruction(TOP_J, REL, strlabel(loop_top, lblbuf), NULL,
							(INT32) 0, NULL);
	}
    else
	{
	if(while_expression->right.ptr.terminal->val.integral != 0L)
	    {
	    instruction(TOP_J, REL, strlabel(loop_top, lblbuf), NULL,
							(INT32) 0, NULL);
	    }
	free_op_node(while_expression);
	}
    printlabel(while_queue.exit);

    delswq();

    ns();
    }


/*
 *	doexpr
 *
 *	If it wasn't a keyword, then we just have an expression (e.g. a
 *	function call or assignment).
 */
PRIVATE VOID
doexpr()
    {
    extern struct op_node	*expression();
    extern VOID			expr_out();
    extern VOID			ns();

    struct op_node 		*expr;	/* Hold the value		*/

    expr = expression();		/* Get the expression		*/
    expr_out(expr,NO_VALUE);		/* Write the code out		*/
    ns();
    }


/*
 *	dofor
 *
 *	Process the for statement.
 *
 *	Remember to handle the special case where the test expression is
 *	blank.
 *
 *	Generate code that looks like:
 *
 *		initializer
 *	top:
 *		test expression
 *		cj	@done
 *		. . . body . . .
 *	inclabel:
 *		increment expression
 *		j	@top
 *	done:
 *
 *	Note that on a machine where "descheduling" is not needed the
 *	above form can be improved on using in the following fashion:
 *
 *		initializer
 *		j	@testexpression
 *	top:
 *		. . . body . . .
 *	inclabel:
 *		increment expression
 *	testexpression:
 *		test expression
 *		negate
 *		cj	@not_done top
 *	done:
 */
PRIVATE VOID
dofor(locptr)
  INT32 locptr;
    {
    extern VOID			addswq();
    extern struct op_node	*bool_expr();
    extern VOID			delswq();
    extern VOID			dvrt_append();
    extern VOID			dvrt_stack();
    extern struct dvrt		*dvrt_unstack();
    extern struct op_node	*expression();
    extern VOID			expr_out();
    extern VOID			free_op_node();
    extern LABEL		getlabel();
    extern VOID			instruction();
    extern BOOL			matchc();
    extern VOID			needpunc();
    extern VOID			printlabel();
    extern char			*strlabel();

    char			lblbuf[NAMEMAX+1];
    struct op_node		*initializer, *test, *increment;
    struct dvrt			*body;
    struct swq			for_queue;	/* element for sw/loop queue */
    LABEL			toplab;		/* Start of the body	*/

    initializer = test = increment = NULL;	/* No expressions yet */

    needpunc('(');
    if (!matchc(';'))		/* If there is an initializer, get it	*/
	{
	initializer = expression();
	needpunc(';');
	}
    if (!matchc(';'))		/* If there is a test, get it		*/
	{
	test = bool_expr();
	needpunc(';');
	}
    if (!matchc(')'))		/* If there is an increment, get it	*/
	{
	increment = expression();
	needpunc(')');
	}

    for_queue.loop = getlabel();	/* Get the various labels	*/
    for_queue.exit = getlabel();
    toplab = getlabel();
    addswq(&for_queue);			/* Add this level to switch queue */

    dvrt_stack();			/* Get the body of the statement */
    statement(locptr);
    body = dvrt_unstack();

    expr_out(initializer,NO_VALUE);	/* Initializer */
/*
 *	Test for a constant valued "test" expression.
 */
    if((test != NULL) && ISCONSTANT(test))
	{
	if(test->right.ptr.terminal->val.integral == 0L)
	    {
	    dvrt_free(body);
	    free_op_node(test);
	    free_op_node(increment);
	    delswq();
	    return;
	    }
	else				/* Unconditional true */
	    {
	    free_op_node(test);
	    test = NULL;
	    }
	}
/*
 *	Now do the general case (including the unconditional true cases).
 */
    printlabel(toplab);
    if(test != NULL)
	{
	expr_out(test,NEED_VALUE | RELAX_BOOL);
	instruction(TOP_CJ, REL, strlabel(for_queue.exit, lblbuf), NULL,
							(INT32) 0, NULL);
	}
    dvrt_append(body);
    printlabel(for_queue.loop);
    expr_out(increment,NO_VALUE);
    instruction(TOP_J, REL, strlabel(toplab, lblbuf), NULL, (INT32) 0, NULL);
    printlabel(for_queue.exit);

    delswq();
    }


/*
 *	dogoto
 *
 *	Process the goto statement.
 *
 *	Slight problem:  A goto to a label that is never defined is not
 *	caught as an error.		DCR	20 May 87
 */
PRIVATE VOID
dogoto()
    {
    extern struct id_node	*add_label();
    extern VOID			error();
    extern LABEL		getlabel();
    extern VOID			instruction();
    extern VOID			junk();
    extern VOID			ns();
    extern char			*strlabel();
    extern BOOL			symname();

    char			goto_label[NAMEMAX+1];
    struct id_node		*label_sym;
    char 			lblbuf[MAXLABEL+1];

    if (symname(goto_label))
	{
	label_sym = add_label(goto_label, getlabel(), FALSE);
	instruction(TOP_J, REL, strlabel(label_sym->val.label, lblbuf), NULL,
							(INT32) 0, NULL);
	}
    else
	{
	error("Illegal goto symbol");
	junk();
	}

    ns();
    }


/*
 *	doif
 *
 *	Process the if (xx) ... else ... statement
 */
PRIVATE VOID
doif(locptr)
  INT32 locptr;
    {
    extern VOID			dvrt_append();
    extern VOID			dvrt_stack();
    extern struct dvrt		*dvrt_unstack();
    extern VOID			expr_out();
    extern VOID			free_op_node();
    extern LABEL		getlabel();
    extern VOID			instruction();
    extern VOID			printlabel();
    extern char			*strlabel();
    extern struct op_node	*test_expression();

    char			lblbuf[NAMEMAX+1];
    struct op_node		*if_expression;
    struct dvrt			*true_statement, *false_statement;
    LABEL			endlab;
    LABEL			falselab;

    if_expression = test_expression();

    dvrt_stack();
    statement(locptr);
    true_statement = dvrt_unstack();

    if (amatch("else"))
	{
	dvrt_stack();
	statement(locptr);
	false_statement = dvrt_unstack();
	}
    else
	false_statement = NULL;
/*
 *	Check for constant "if" expression.
 */
    if(ISCONSTANT(if_expression))
	{
	if(if_expression->right.ptr.terminal->val.integral != 0L)
	    {
	    dvrt_append(true_statement);
	    dvrt_free(false_statement);
	    }
	else
	    {
	    dvrt_append(false_statement);
	    dvrt_free(true_statement);
	    }
	free_op_node(if_expression);
	return;
	}
/*
 *	Wasn't a constant "if", do the real thing.
 */
    expr_out(if_expression,NEED_VALUE | RELAX_BOOL);
    endlab = getlabel();
    falselab = ((false_statement == NULL) ? endlab : getlabel());
    instruction(TOP_CJ, REL, strlabel(falselab, lblbuf), NULL, (INT32) 0, NULL);
    dvrt_append(true_statement);
    if (false_statement != NULL)
	{
	instruction(TOP_J, REL, strlabel(endlab, lblbuf), NULL, (INT32) 0, NULL);
	printlabel(falselab);
	dvrt_append(false_statement);
	}
    printlabel(endlab);
    }

/*
 *	doreturn
 *
 *	Process return statement.
 */
PRIVATE VOID
doreturn()
	{
	extern	VOID		abs_oper();
	extern	struct op_node	*cast_it();
	extern	BOOL		endst();
	extern	struct op_node	*expression();
	extern	VOID		expr_out();
	extern	VOID		fatal();
	extern	VOID		instruction();
	extern	VOID		loadconstant();
	extern	struct id_node	*lookup_node();
	extern	struct op_node	*lterminal();
	extern	VOID		modify_rslink();
	extern	VOID		no_oper();
	extern	VOID		ns();
	extern	BOOL		same_type();
	extern	VOID		warning();
	extern	INT32		word();

	struct	op_node		*return_expression;
	INT32			size;
	struct	id_node		*sym;

	if(! endst())
		{
		return_expression = expression();
		if(! same_type(return_expression->node_type,Func_type,
			/* ptr_eq_array = */ FALSE))
			{
			return_expression = cast_it(return_expression,
				Func_type);
			}
		if(EXPRTYPE(return_expression) & SU_NODE)
			{
			if((sym = lookup_node(RSLINK,NS_VAR)) == NULL)
				fatal("Compiler error: doreturn");
			if(return_expression->operator == OP_CALL)
				{
				/* Avoid duplicate copying of structure */
				modify_rslink(return_expression,
					lterminal(sym));
				expr_out(return_expression,NEED_VALUE);
				}
			else
				{
				expr_out(return_expression,
					(NEED_VALUE | EVAL_ADDR));
				size = return_expression->node_type->
					val.integral;
				if(size <= Sizeof_int)
					abs_oper(TOP_LDNL,(INT32) 0);
				instruction(TOP_LDL,ABS,WORKSPACE,NULL,
					-(word(-sym->address.offset)),
#ifdef	DEBUG
				(Debug ? "struct/union function return ptr" :
					NULL)
#else
						 NULL
#endif	/* DEBUG */
							);
				if(size <= Sizeof_int)
					abs_oper(TOP_STNL,(INT32) 0);
				else
					{
					loadconstant(size);
					no_oper(TOP_MOVE);
					}
				}
			instruction(TOP_LDL,ABS,WORKSPACE,NULL,
				-(word(-sym->address.offset)),
#ifdef	DEBUG
			(Debug ? "struct/union function return ptr" : NULL)
#else
					 NULL
#endif	/* DEBUG */
						);
			}
		else
			expr_out(return_expression,NEED_VALUE);
		}
	else
		{
		if(! same_type(Voidtype,Func_type,/* ptr_eq_array = */ FALSE))
			warning(2,"Inconsistent return type");
		}
	ns();

	if(Traceout)			/* Do function exit trace fn	*/
		instruction(TOP_CALL,REL,Troname,NULL,(INT32) 0,NULL);

	instruction(TOP_RETF,ABS,WORKSPACE,NULL,(INT32) 0,NULL);
	}

/*
 *	doswitch()
 *
 *	Process the switch statement
 */
PRIVATE VOID
doswitch(locptr)
	INT32 locptr;
	{
	extern	VOID		abs_oper();
	extern	VOID		addswheader();
	extern	VOID		addswq();
	extern	VOID		delswheader();
	extern	VOID		delswq();
	extern	VOID		dvrt_append();
	extern	VOID		dvrt_stack();
	extern	struct	dvrt	*dvrt_unstack();
	extern	VOID		error();
	extern	struct op_node	*expression();
	extern	VOID		expr_out();
	extern	VOID		free_temp();
	extern	TEMP		get_temp();
	extern	VOID		instruction();
	extern	VOID		loadconstant();
	extern	VOID		load_temp();
	extern	VOID		needpunc();
	extern	VOID		no_oper();
	extern	VOID		printlabel();
	extern	VOID		store_temp();
	extern	char		*strlabel();
	extern	struct swq	*Swqueue;

	BOOL			cdup;
	char			lblbuf[NAMEMAX+1];
	int			sw_count;
	struct	swheader	switch_header;
	struct	dvrt		*switch_body;
	struct	switem		*sl;
	TEMP			temp;

	needpunc('(');
	switch_header.sh_operand = expression();
	needpunc(')');

	switch_header.sh_swqueue.exit = getlabel(); /* Exit (break) label */
	switch_header.sh_swqueue.loop =
			((Swqueue != NULL) ? Swqueue->loop : (LABEL) 0);

	addswq(&switch_header.sh_swqueue);	/* Add sw/loop queue elem */
	switch_header.sh_default = (LABEL) 0;
	switch_header.sh_caselist = NULL;	/* Case list */
	addswheader(&switch_header);		/* Add to switch list */

	if(! ISINTEGRAL(switch_header.sh_operand))
		error("Switch expression not integral type");

	dvrt_stack();
	statement(locptr);		/* Get switch body */
	switch_body = dvrt_unstack();

	expr_out(switch_header.sh_operand,NEED_VALUE);
/*
 *	Figure out what template to generate the switch statement with
 *	depending on whether we have 3 or more cases (use "C" register self
 *	duplication feature), or less (use distributed re-fetch of switch
 *	value).
 */
	for(sw_count = 0, sl = switch_header.sh_caselist;
		((sl != NULL) && (sw_count < 4)); sw_count++, sl = sl->sw_next)
		;
	if(sw_count < 4)
		cdup = FALSE;		/* Use distributed re-fetch scheme */
	else
		cdup = TRUE;		/* Use "C" register auto-dup feature */

	if(cdup)
		{			/* Fill all registers */
		if(CPU_DUP)
			{
			no_oper(TOP_DUP);	/* Set value of C register */
			no_oper(TOP_DUP);
			}
		else
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			load_temp(Inttype,temp);
			load_temp(Inttype,temp);
			load_temp(Inttype,temp);	/* Set value of C */
			}
		}
	else
		{		/* Fill only A reg. (+temp if no dup) */
		if(CPU_NO_DUP && (sw_count > 1))
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			}
		}
/*
 *	For future versions, we could sort the case list in a tree order to
 *	build a binary-executable tree to select the appropriate case label.
 *	For now, take the simple (cowardly) way out.
 */
	/* Emit simple compare/jump clauses */
	for(sl = switch_header.sh_caselist; sl != NULL; sl = sl->sw_next)
		{
		if(! cdup)
			{
			if(CPU_NO_DUP && (sw_count > 1))
				load_temp(Inttype,temp);
			else
				{
				if(sl->sw_next != NULL)
					no_oper(TOP_DUP);
				}
			}
		if(sl->sw_value != (INT32) 0)
			abs_oper(TOP_ADC, (INT32) -sl->sw_value);
		instruction(TOP_CJ, REL, strlabel(sl->sw_label, lblbuf),
				NULL, (LABEL) 0, NULL);
		}
	if(switch_header.sh_default != (LABEL) 0)
		{
		instruction(TOP_J,REL,
			strlabel(switch_header.sh_default,lblbuf),
				NULL,(INT32) 0,NULL);
		}
	else
		{
		instruction(TOP_J,REL,
			strlabel(switch_header.sh_swqueue.exit,lblbuf),
				NULL,(INT32) 0,NULL);
		}

	if(CPU_NO_DUP && (sw_count > 1))
		free_temp(temp);

	dvrt_append(switch_body);	/* Output body of switch statement */
	printlabel(switch_header.sh_swqueue.exit);
	delswq();			/* Delete the active sw/loop element */
	delswheader();			/* And the switch header */
	}


/*
 *	dowhile
 *
 *	Process the while... statement.
 */
PRIVATE VOID
dowhile(locptr)
	INT32			locptr;
	{
	extern	VOID		addswq();
	extern	VOID		delswq();
	extern	VOID		dvrt_append();
	extern	VOID		dvrt_stack();
	extern	struct dvrt	*dvrt_unstack();
	extern	VOID		expr_out();
	extern	VOID		free_op_node();
	extern	LABEL		getlabel();
	extern	VOID		gen_while();
	extern	VOID		instruction();
	extern	VOID		printlabel();
	extern	char		*strlabel();
	extern	struct op_node	*test_expression();

	char			lblbuf[NAMEMAX+1];
	struct	swq		while_queue;	/* Element for sw/loop queue */
	struct	op_node		*while_expression;
	struct	dvrt		*while_body;

	while_queue.loop = getlabel();
	while_queue.exit = getlabel();	/* Add labels to queue element */

	addswq(&while_queue);		/* Add this level to while queue */

	while_expression = test_expression();

	dvrt_stack();
	statement(locptr);		/* Capture body of while body */
	while_body = dvrt_unstack();
/*
 *	Test for a FALSE constant expression "while".
 */
	if(ISCONSTANT(while_expression) &&
		(while_expression->right.ptr.terminal->val.integral == 0L))
		{
		dvrt_free(while_body);
		free_op_node(while_expression);
		}
/*
 *	The following code handles the general case of the while as well as
 *	the "while(TRUE)" case.
 */
	else
		{
		printlabel(while_queue.loop);
		if(! ISCONSTANT(while_expression))
			{
			expr_out(while_expression,(NEED_VALUE | RELAX_BOOL));
			instruction(TOP_CJ,REL,
				strlabel(while_queue.exit,lblbuf),NULL,
					(INT32) 0,NULL);
			}
		dvrt_append(while_body);
		instruction(TOP_J,REL,strlabel(while_queue.loop,lblbuf),NULL,
			(INT32) 0,NULL);
		printlabel(while_queue.exit);
		}

	delswq();
	}

/*
 *	Syntax requires a semicolon here
 */
PUBLIC	VOID
ns()
	{
	extern	VOID		error();
	extern	BOOL		matchc();

	if(! matchc(';'))
		error("Missing semicolon");
	}

/*
 *	Evaluate a expression for a boolean result.  If expression is a
 *	floating point constant convert to the equivalent integral value.
 *	If a floating point expression paste on a != 0.0
 */
PRIVATE	struct	op_node	*
bool_expr()
    {
    extern struct op_node	*binaryop();
    extern struct op_node	*expression();
    extern VOID			free_op_node();
    extern struct op_node	*make_const();
    extern struct op_node	*make_fconst();

    struct op_node		*op;

    op = expression();
    if(EXPRTYPE(op) & FP_NODE)
	{
	if(ISCONSTANT(op))
	    {
	    if(op->right.ptr.terminal->val.fvalue != 0.0)
		{
		free_op_node(op);
		op = make_const((INT32) 1, Inttype);
		}
	    else
		{
		free_op_node(op);
		op = make_const((INT32) 0, Inttype);
		}
	    }
	else
	    op = binaryop(OP_NE, op, make_fconst(0.0, op->node_type));
	}
    return (op);
    }


/*
 *	Recognize a "(", boolean expression, and ")"
 */
PRIVATE struct	op_node *
test_expression()
    {
    extern struct op_node	*bool_expr();
    extern VOID			needpunc();

    struct op_node		*op;

    needpunc('(');
    op = bool_expr();
    needpunc(')');

    return (op);
    }

/* End of TCX3.C */
