/*
**	TCX Compiler (6c of 11)	Expression Syntax Tree Walker
**
**	Copyright (c) 1987-1988 by Logical Systems.  All rights reserved.
**
**	Published by:	Logical Systems
**			P.O. Box 1702
**			Corvallis, OR 97339 (USA)
**
**	03/17/87
*/

#include <stdio.h>
#include <ctype.h>
#include "tcx.h"
#include "tcxdecl.h"

/*
 *	General skeleton routine for "op=" routines.  It uses the following
 *	helper routines:
 *
 *	"op_sl"	-	Called if the left expression is simple.  This is
 *			generally the "do_op" routine for the non-assignment
 *			form of the operator.  This routine is called after
 *			"op_cr1" so the trivial constant cases have already
 *			been done.
 *
 *	"op_cr1" -	Called first to handle trival constant cases (a *= 0,
 *			a *= 1, a &= 0xFFFFFFFF, etc).  This routine must
 *			perform all reloading of the result if needed etc.
 *			If the constant being examined doesn't qualify as
 *			trivial the special depth value of UNDEF_DEEP is
 *			returned.
 *
 *	"op_cr2" -	This routine is called if neither "op_cr1" or "op_sl"
 *			was appropriate and is responsible for all other
 *			code generation chores involving right constants.
 *
 *	"op_sfpr" -	If this function pointer is non-NULL then this function
 *			should generate an instruction to be used when the
 *			expression type is floating point and the right operand
 *			is a simple and of the same size as the expression.
 *			This means something like "FPLDNLMULSN"/"FPLDNLMULDB",
 *			etc.  This function is called after trying to use the
 *			above special case routines.
 *
 *	"op_g" -	Called to perform the general case of evaluation when
 *			none of the special cases apply.
 *
 *	Note that much of the entry and cleanup code the above routines is
 *	common to all the operators and is placed in this function.
 */

PUBLIC	unsigned int
do_aop(expr,vf,op_sl,op_cr1,op_cr2,op_sfpr,op_g)
  struct op_node		*expr;
  int				vf;
  unsigned	int		(*op_sl)();
  unsigned	int		(*op_cr1)();
  unsigned	int		(*op_cr2)();
  VOID				(*op_sfpr)();
  unsigned	int		(*op_g)();
    {
    extern unsigned int		assign_op();
    extern unsigned int		cnvt_type();
    extern unsigned int		depth_max();
    extern unsigned int		do_load();
    extern unsigned int		do_store();
    extern unsigned int		expr_code();
    extern VOID			free_temp();
    extern TEMP			get_temp();
    extern BOOL			is_simple();
    extern unsigned int		load_indirect();
    extern VOID			load_temp();
    extern VOID			no_oper();
    extern struct id_node	*op_types();
    extern unsigned int		store_indirect();
    extern VOID			store_temp();

    unsigned int		depth;
    BOOL			dupped;
    struct id_node		*optype;
    unsigned int		sdepth;
    TEMP			temp;

    optype = op_types(expr);

    if (ISRIGHTCONSTANT(expr) && ((depth = (*op_cr1)(expr,vf)) != UNDEF_DEEP))
	{
	/* This clause does the degenerative constant operators (e.g.
	 * add 0; mul by 1; etc.  It has the responsibility to reload
	 * (when necessary) the target value of the expression.
	 * However, we have nothing further to do here, as it was all
	 * accomplished by op_cr1().
	 */
	}
    else if (is_simple(LEFT(expr)))
	{
	depth = (*op_sl)(expr, NEED_VALUE);
	depth = depth_max(depth,cnvt_type(expr->node_type, optype));
	sdepth = 0;			/* Integer stack used below store */
	if (vf & NEED_VALUE)
	    {
	    if(LEFTTYPE(expr) & FP_NODE)
		{
		dupped = TRUE;
		no_oper(TOP_FPDUP);
		depth = depth_max(depth, FLT_DEEP(2));
		}
	    else if(CPU_DUP && (! (LEFTTYPE(expr) & CHAR_NODE)))
		{
		dupped = TRUE;
		no_oper(TOP_DUP);
		sdepth = INT_DEEP(1);
		}
	    else
		dupped = FALSE;
	    }
	depth = depth_max(do_store(LEFT(expr)) + sdepth, depth);

	if ((vf & NEED_VALUE) && (! dupped))
	    {
	    depth = depth_max(do_load(expr->node_type, LEFT(expr), vf), depth);
	    }
	}
    else
	{
	if (ISRIGHTCONSTANT(expr) ||
		((op_sfpr != NULL) && is_simple(RIGHT(expr)) &&
			(RIGHTTYPE(expr) & FP_NODE) &&
				(RIGHT(expr)->node_type->val.integral ==
				optype->val.integral)))
	    {
	    depth = expr_code(expr->node_type, LEFT(expr),
						NEED_VALUE | EVAL_ADDR);
	    temp = get_temp();

	    if(CPU_DUP)
		{
		no_oper(TOP_DUP);
		depth = depth_max(depth, INT_DEEP(2));
		}
	    store_temp(Inttype, temp);	/* Lvalue address */
	    if(CPU_NO_DUP)
		load_temp(Inttype, temp);

	    depth = depth_max(depth, load_indirect(LEFT(expr), optype,
						(INT32) 0, NEED_VALUE));

	    if(ISRIGHTCONSTANT(expr))
		{
		depth = depth_max(depth, (*op_cr2)(expr));
		}
	    else
		{
		depth = depth_max(depth, expr_code(NULL, RIGHT(expr),
						EVAL_ADDR | NEED_VALUE) +
							FLT_DEEP(1));
		(*op_sfpr)(expr);
		depth = depth_max(depth,FLT_DEEP(2));
		}
	    }
	else
	    {
	    /*
	     * Load by generating the code for the Lvalue (address) and the
	     * Rvalue of the right operand.  The right operand MUST be
	     * evaluated first to preserve right-associativity rules.  The
	     * address (Lvalue) is in A.
	     */
	    temp = get_temp();

	    depth = (*op_g)(expr, temp);
	    }

	if (vf & NEED_VALUE)
	    {
	    if(LEFTTYPE(expr) & FP_NODE)
		{
		dupped = TRUE;
		no_oper(TOP_FPDUP);
		depth = depth_max(depth, FLT_DEEP(2));
		}
	    else if(CPU_DUP && (! (LEFTTYPE(expr) & CHAR_NODE)) &&
		(! (RIGHTTYPE(expr) & FP_NODE)))
		{
		dupped = TRUE;
		no_oper(TOP_DUP);
		depth = depth_max(depth, INT_DEEP(3));
		}
	    else
		dupped = FALSE;
	    }

	load_temp(Inttype, temp);

	/* If target is integer, store_indirect will do conversion */
	depth = depth_max(depth, store_indirect(LEFT(expr),optype));

	if ((vf & NEED_VALUE) && (! dupped))
	    {
	    load_temp(Inttype, temp);
	    depth = depth_max(load_indirect(LEFT(expr),	NULL, (INT32) 0, vf),
						depth);
	    }
	free_temp(temp);
	}
    return (depth);
    }


PUBLIC	unsigned int
do_amul(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		amul_cr1();
    extern unsigned int		amul_cr2();
    extern unsigned int		amul_g();
    extern VOID			amul_sfpr();
    extern unsigned int		do_aop();
    extern unsigned int		do_mul();

    return (do_aop(expr,value_flag,do_mul,amul_cr1,amul_cr2,amul_sfpr,amul_g));
    }


PRIVATE unsigned int
amul_cr1(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		do_assign();
    extern unsigned int		expr_code();

    unsigned int		depth;

    if (! (RIGHTTYPE(expr) & FP_NODE) && (RIGHTCONSTANT(expr) == 0))
    /* "a *= 0" is equivalent to "a = 0" */
	{
	depth = do_assign(expr, value_flag);
	}
    else if (!(RIGHTTYPE(expr) & FP_NODE) && (RIGHTCONSTANT(expr) == 1))
	{
	/* Just do target for side-effects */
	depth = expr_code(expr->node_type, LEFT(expr), value_flag);
	}
    else
	depth = UNDEF_DEEP;

    return (depth);
    }


PRIVATE unsigned int
amul_cr2(expr)
  struct op_node		*expr;
    {
    extern VOID			fp_oper();
    extern unsigned int		fp_oplit();
    extern VOID			free_op_node();
    extern VOID			loadconstant();
    extern double		l_to_f();
    extern struct op_node	*make_fconst();
    extern VOID			no_oper();
    extern struct id_node	*op_types();

    unsigned int		depth;
    double			fvalue;
    struct op_node		*op;
    struct id_node		*optype;

    optype = op_types(expr);

    depth = INT_DEEP(1);

    if ((LEFTTYPE(expr) & FP_NODE) && !(RIGHTTYPE(expr) & FP_NODE) &&
		(RIGHTCONSTANT(expr) == 2))
	{
	no_oper(TOP_FPUMULBY2);
	depth |= FLT_DEEP(1);
	}
    else if (!(LEFTTYPE(expr) & FP_NODE) && !(RIGHTTYPE(expr) & FP_NODE) &&
		(RIGHTCONSTANT(expr) == Sizeof_int))
	{
	no_oper(TOP_BCNT);
	}
    else if (optype->type & FP_NODE)
	{
	if (RIGHTTYPE(expr) & FP_NODE)
	    fvalue = RIGHTFCONSTANT(expr);
	else
	    fvalue = l_to_f(RIGHT(expr));
	op = make_fconst(fvalue, optype);
	depth = fp_oplit(optype, op) + FLT_DEEP(2);
	fp_oper(optype, TOP_BAD, TOP_FPLDNLMULSN, TOP_FPLDNLMULDB);
	free_op_node(op);
	}
    else
	{
	loadconstant(RIGHTCONSTANT(expr));
	no_oper(TOP_PROD);
	depth = INT_DEEP(2);
	}
    return (depth);
    }


PRIVATE VOID
amul_sfpr(expr)
  struct op_node		*expr;
    {
    extern VOID			fp_oper();

    fp_oper(RIGHT(expr)->node_type, TOP_BAD, TOP_FPLDNLMULSN, TOP_FPLDNLMULDB);
    }


PRIVATE unsigned int
amul_g(expr, temp)
  struct op_node		*expr;
  TEMP				temp;
    {
    extern unsigned int		assign_op();
    extern VOID			fp_oper();
    extern VOID			no_oper();
    extern struct id_node	*op_types();

    unsigned int		depth;

    depth = assign_op(expr, temp);
    fp_oper(op_types(expr),TOP_PROD,TOP_FPMUL,TOP_FPMUL);

    return (depth);
    }


PUBLIC	unsigned int
do_adiv(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		adiv_cr1();
    extern unsigned int		adiv_cr2();
    extern unsigned int		adiv_g();
    extern unsigned int		do_aop();
    extern unsigned int		do_div();

    return (do_aop(expr,value_flag,do_div,adiv_cr1,adiv_cr2,NULL,adiv_g));
    }


PRIVATE unsigned int
adiv_cr1(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern VOID			error();
    extern unsigned int		expr_code();

    unsigned int		depth;

    if (! (RIGHTTYPE(expr) & FP_NODE) && (RIGHTCONSTANT(expr) == 0))
	{
	error("Division by zero");
	depth = 0;
	}
    else if (! (RIGHTTYPE(expr) & FP_NODE) && (RIGHTCONSTANT(expr) == 1))
	{
	/* Just do target for side-effects */
	depth = expr_code(expr->node_type, LEFT(expr), value_flag);
	}
    else
	depth = UNDEF_DEEP;

    return (depth);
    }


PRIVATE unsigned int
adiv_cr2(expr)
  struct op_node		*expr;
    {
    extern INT32		bitlog2();
    extern VOID			fp_oper();
    extern unsigned int		fp_oplit();
    extern VOID			free_op_node();
    extern VOID			loadconstant();
    extern double		l_to_f();
    extern struct op_node	*make_fconst();
    extern VOID			no_oper();
    extern struct id_node	*op_types();
    extern VOID			rev();

    unsigned int		depth;
    double			fvalue;
    struct op_node		*op;
    struct id_node		*optype;
    INT32			shift;

    optype = op_types(expr);

    if ((LEFTTYPE(expr) & FP_NODE) && !(RIGHTTYPE(expr) & FP_NODE) &&
	(RIGHTCONSTANT(expr) == 2))
	{
	no_oper(TOP_FPUDIVBY2);
	depth = INT_DEEP(1) | FLT_DEEP(1);
	}
    else if (optype->type & FP_NODE)
	{
	if (RIGHTTYPE(expr) & FP_NODE)
	    fvalue = RIGHTFCONSTANT(expr);
	else
	    fvalue = l_to_f(RIGHT(expr));

	op = make_fconst(1.0 / fvalue, optype);	/* Mul by reciprocal */
	depth = fp_oplit(optype, op) + FLT_DEEP(2);
	fp_oper(optype, TOP_BAD, TOP_FPLDNLMULSN, TOP_FPLDNLMULDB);
	free_op_node(op);
	}
    else
	{
	if (optype->type & UNSIGNED_NODE)	/* Unsigned? */
	    {
	    if((shift = bitlog2(RIGHTCONSTANT(expr))) >= 0)
		{
		loadconstant(shift);		/* We can shift instead */
		no_oper(TOP_SHR);
		depth = INT_DEEP(2);
		}
	    else				/* Must do unsigned divide */
		{
		loadconstant((INT32) 0);
		rev(optype);			/* Bury zero extension */
		loadconstant(RIGHTCONSTANT(expr));
		no_oper(TOP_LDIV);
		depth = INT_DEEP(3);
		}
	    }
	else				/* Was signed integral */
	    {
	    loadconstant(RIGHTCONSTANT(expr));
	    no_oper(TOP_DIV);
	    depth = INT_DEEP(2);
	    }
	}
    return (depth);
    }


PRIVATE unsigned int
adiv_g(expr, temp)
  struct op_node		*expr;
  TEMP				temp;
    {
    extern unsigned int		assign_op();
    extern unsigned int		audm_op();
    extern VOID			no_oper();
    extern struct id_node	*op_types();

    unsigned int		depth;
    struct id_node		*optype;

    optype = op_types(expr);

    if(optype->type & FP_NODE)		/* Floating point */
	{
	depth = assign_op(expr, temp);
	no_oper(TOP_FPDIV);
	}
    else				/* Was integral type */
	{
	if (optype->type & UNSIGNED_NODE)	/* Unsigned? */
	    {
	    depth = audm_op(expr, temp);
	    }
	else				/* Signed integral */
	    {
	    depth = assign_op(expr, temp);
	    no_oper(TOP_DIV);
	    }
	}
    return (depth);
    }


PUBLIC	unsigned int
do_amod(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		adiv_cr1();
    extern unsigned int		amod_cr2();
    extern unsigned int		amod_g();
    extern unsigned int		do_aop();
    extern unsigned int		do_mod();

    return (do_aop(expr,value_flag,do_mod,adiv_cr1,amod_cr2,NULL,amod_g));
    }

PRIVATE unsigned int
amod_cr2(expr)
  struct op_node		*expr;
    {
    extern VOID			loadconstant();
    extern VOID			no_oper();
    extern struct id_node	*op_types();
    extern VOID			rev();

    unsigned int		depth;
    struct id_node		*optype;

    optype = op_types(expr);

    if (optype->type & UNSIGNED_NODE)	/* Unsigned? */
	{
	loadconstant((INT32) 0);
	rev(optype);			/* Bury zero extension */
	loadconstant(RIGHTCONSTANT(expr));
	no_oper(TOP_LDIV);
	rev(optype);
	depth = INT_DEEP(3);
	}
    else				/* Was signed integral */
	{
	loadconstant(RIGHTCONSTANT(expr));
	no_oper(TOP_REM);
	depth = INT_DEEP(2);
	}
    return (depth);
    }


PRIVATE unsigned int
amod_g(expr, temp)
  struct op_node		*expr;
  TEMP				temp;
    {
    extern unsigned int		assign_op();
    extern unsigned int		audm_op();
    extern VOID			no_oper();
    extern struct id_node	*op_types();
    extern VOID			rev();

    unsigned int		depth;
    struct id_node		*optype;

    optype = op_types(expr);

    if (optype->type & UNSIGNED_NODE)	/* Unsigned? */
	{
	depth = audm_op(expr, temp);
	rev(optype);
	}
    else				/* Signed integral */
	{
	depth = assign_op(expr, temp);
	no_oper(TOP_REM);
	}
    return (depth);
    }


/*
 *	Use double precision signed division to handle single precision
 *	unsigned.  This routine is constructed for the /= and %= operators.
 *	It is functionally equivalent to "assign_op" except it also does a
 *	trailing "LDIV".
 */
PUBLIC unsigned int
audm_op(expr,temp)
  struct op_node	*expr;
  TEMP			temp;
    {
    extern unsigned int	depth_max();
    extern VOID		dvrt_append();
    extern VOID		dvrt_free();
    extern VOID		dvrt_stack();
    extern struct dvrt	*dvrt_unstack();
    extern VOID		free_temp();
    extern unsigned int	expr_code();
    extern TEMP		get_temp();
    extern VOID		loadconstant();
    extern unsigned int	load_indirect();
    extern VOID		load_temp();
    extern VOID		no_oper();
    extern VOID		store_temp();
    extern VOID		rev();

    unsigned int	deep_lop;
    unsigned int	deep_rop;
    unsigned int	ideep_lop;
    unsigned int	ideep_rop;
    TEMP		temp2;
    struct dvrt		*dvrt_lop;
    struct dvrt		*dvrt_rop;

    dvrt_stack();
    deep_lop = expr_code(expr->node_type, LEFT(expr), NEED_VALUE | EVAL_ADDR);
    if(CPU_DUP)
	{
	no_oper(TOP_DUP);
	deep_lop = depth_max(deep_lop, INT_DEEP(2));
	}
    store_temp(Inttype, temp);
    if(CPU_NO_DUP)
	load_temp(Inttype, temp);
    deep_lop = depth_max(deep_lop, load_indirect(LEFT(expr), NULL, (INT32) 0,
						NEED_VALUE));
    dvrt_lop = dvrt_unstack();

    dvrt_stack();
    deep_rop = expr_code(expr->node_type, RIGHT(expr), NEED_VALUE);
    dvrt_rop = dvrt_unstack();

    /*
     *	Given the left and right depths, figure out how to set up the stack
     *	to minimize the use of "rev" instructions and "temp" variables.
     */
    ideep_rop = deep_rop & INT_STACK;
    ideep_lop = deep_lop & INT_STACK;

    if (ideep_rop == INT_DEEP(1))
	{
	if (ideep_lop < INT_DEEP(3))
	    {
	    loadconstant((INT32) 0);
	    dvrt_append(dvrt_lop);
	    dvrt_append(dvrt_rop);
	    }
	else
	    {
	    dvrt_append(dvrt_lop);
	    loadconstant((INT32) 0);
	    rev(Inttype);
	    dvrt_append(dvrt_rop);
	    }
	}
    else
    if ((ideep_rop == INT_DEEP(2)) && (ideep_lop == INT_DEEP(1)))
	{
	loadconstant((INT32) 0);
	dvrt_append(dvrt_rop);
	dvrt_append(dvrt_lop);
	rev(Inttype);
	}
    else
    if ((ideep_rop == INT_DEEP(3)) && (ideep_lop == INT_DEEP(1)))
	{
	dvrt_append(dvrt_rop);
	loadconstant((INT32) 0);
	rev(Inttype);
	dvrt_append(dvrt_lop);
	rev(Inttype);
	}
    /*
     *	Well, no way to avoid using another "temp".  Do the code so as to
     *   minimize the number of "rev" instructions needed.
     */
    else
	{
	/* For details about expr_code reevaluation, see bin_op */
	dvrt_free(dvrt_rop);	/* Flush the original */
	dvrt_free(dvrt_lop);	/* And the other side */
	(VOID) expr_code(expr->node_type, RIGHT(expr), NEED_VALUE);

	temp2 = get_temp();
	store_temp(Inttype, temp2);

	if (ideep_lop == INT_DEEP(2))
	    {
	    loadconstant((INT32) 0);
	    (VOID) expr_code(expr->node_type, LEFT(expr), NEED_VALUE);
	    }
	else
	    {
	    (VOID) expr_code(expr->node_type, LEFT(expr), NEED_VALUE);
	    loadconstant((INT32) 0);
	    rev(Inttype);
	    }

	load_temp(Inttype, temp2);
	free_temp(temp2);
	}

    no_oper(TOP_LDIV);

    /* Return FP stack usage + int stack usage of 3 */
    return (depth_max(depth_max(deep_lop,  deep_rop), INT_DEEP(3)));
    }


PUBLIC	unsigned int
do_aadd(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		aadd_cr1();
    extern unsigned int		aadd_cr2();
    extern unsigned int		aadd_g();
    extern VOID			aadd_sfpr();
    extern unsigned int		do_add();
    extern unsigned int		do_aop();

    return (do_aop(expr,value_flag,do_add,aadd_cr1,aadd_cr2,aadd_sfpr,aadd_g));
    }


PRIVATE unsigned int
aadd_cr1(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		expr_code();

    unsigned int		depth;

    if (!(RIGHTTYPE(expr) & FP_NODE) && (RIGHTCONSTANT(expr) == 0))
	{
	/* Just do target for side-effects */
	depth = expr_code(expr->node_type, LEFT(expr), value_flag);
	}
    else
	depth = UNDEF_DEEP;

    return (depth);
    }


PRIVATE unsigned int
aadd_cr2(expr)
  struct op_node		*expr;
    {
    extern VOID			abs_oper();
    extern VOID			fp_oper();
    extern unsigned int		fp_oplit();
    extern VOID			free_op_node();
    extern double		l_to_f();
    extern struct op_node	*make_fconst();
    extern struct id_node	*op_types();

    unsigned int		depth;
    double			fvalue;
    struct op_node		*op;
    struct id_node		*optype;

    optype = op_types(expr);

    if (optype->type & FP_NODE)
	{
	if (RIGHTTYPE(expr) & FP_NODE)
	    fvalue = RIGHTFCONSTANT(expr);
	else
	    fvalue = l_to_f(RIGHT(expr));

	op = make_fconst(fvalue, optype);
	depth = fp_oplit(optype, op) + FLT_DEEP(2);
	fp_oper(optype, TOP_BAD, TOP_FPLDNLADDSN, TOP_FPLDNLADDDB);
	free_op_node(op);
	}
    else
	{
	abs_oper(TOP_ADC, RIGHTCONSTANT(expr));
	depth = INT_DEEP(1);
	}
    return (depth);
    }


PRIVATE VOID
aadd_sfpr(expr)
  struct op_node		*expr;
    {
    extern VOID			fp_oper();

    fp_oper(RIGHT(expr)->node_type, TOP_BAD, TOP_FPLDNLADDSN, TOP_FPLDNLADDDB);
    }


PRIVATE unsigned int
aadd_g(expr, temp)
  struct op_node		*expr;
  TEMP				temp;
    {
    extern unsigned int		assign_op();
    extern VOID			fp_oper();
    extern VOID			no_oper();
    extern struct id_node	*op_types();

    unsigned int		depth;

    depth = assign_op(expr, temp);

    fp_oper(op_types(expr),TOP_ADD,TOP_FPADD,TOP_FPADD);

    return (depth);
    }


PUBLIC	unsigned int
do_asub(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		aadd_cr1();
    extern unsigned int		asub_cr2();
    extern unsigned int		asub_g();
    extern unsigned int		do_aop();
    extern unsigned int		do_sub();

    return (do_aop(expr,value_flag,do_sub,aadd_cr1,asub_cr2,NULL,asub_g));
    }


PRIVATE unsigned int
asub_cr2(expr)
  struct op_node		*expr;
    {
    extern VOID			abs_oper();
    extern VOID			fp_oper();
    extern unsigned int		fp_oplit();
    extern VOID			free_op_node();
    extern double		l_to_f();
    extern struct op_node	*make_fconst();
    extern struct id_node	*op_types();

    unsigned int		depth;
    double			fvalue;
    struct op_node		*op;
    struct id_node		*optype;

    optype = op_types(expr);

    if (optype->type & FP_NODE)
	{
	if (RIGHTTYPE(expr) & FP_NODE)
	    fvalue = RIGHTFCONSTANT(expr);
	else
	    fvalue = l_to_f(RIGHT(expr));

	op = make_fconst(-fvalue, optype);
	depth = fp_oplit(optype, op) + FLT_DEEP(2);
	fp_oper(optype, TOP_BAD, TOP_FPLDNLADDSN, TOP_FPLDNLADDDB);
	free_op_node(op);
	}
    else
	{
	abs_oper(TOP_ADC,-RIGHTCONSTANT(expr));
	depth = INT_DEEP(1);
	}
    return (depth);
    }


PRIVATE unsigned int
asub_g(expr, temp)
  struct op_node		*expr;
  TEMP				temp;
    {
    extern unsigned int		assign_op();
    extern VOID			fp_oper();
    extern VOID			no_oper();
    extern struct id_node	*op_types();

    unsigned int		depth;

    depth = assign_op(expr, temp);

    fp_oper(op_types(expr),TOP_SUB,TOP_FPSUB,TOP_FPSUB);

    return (depth);
    }


PUBLIC	unsigned int
do_alshift(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		aadd_cr1();
    extern unsigned int		alsf_cr2();
    extern unsigned int		alsf_g();
    extern unsigned int		do_aop();
    extern unsigned int		do_lshift();

    return (do_aop(expr,value_flag,do_lshift,aadd_cr1,alsf_cr2,NULL,alsf_g));
    }


PRIVATE unsigned int
alsf_cr2(expr)
  struct op_node		*expr;
    {
    extern VOID			loadconstant();
    extern VOID			no_oper();

    loadconstant(RIGHTCONSTANT(expr));
    no_oper(TOP_SHL);

    return (INT_DEEP(2));
    }


PRIVATE unsigned int
alsf_g(expr, temp)
  struct op_node		*expr;
  TEMP				temp;
    {
    extern unsigned int		assign_op();
    extern VOID			no_oper();

    unsigned int		depth;

    depth = assign_op(expr, temp);

    no_oper(TOP_SHL);

    return (depth);
    }


PUBLIC	unsigned int
do_arshift(expr,value_flag)
  struct op_node		*expr;
  int				value_flag;
    {
    extern unsigned int		aadd_cr1();
    extern unsigned int		arsf_cr2();
    extern unsigned int		arsf_g();
    extern unsigned int		do_aop();
    extern unsigned int		do_rshift();

    return (do_aop(expr,value_flag,do_rshift,aadd_cr1,arsf_cr2,NULL,arsf_g));
    }


PRIVATE unsigned int
arsf_cr2(expr)
  struct op_node		*expr;
    {
    extern VOID			loadconstant();
    extern VOID			no_oper();

    unsigned int		depth;

    if (Cpu_srshift && (RIGHTTYPE(expr) & SIGNED_NODE))
	{
	no_oper(TOP_XDBLE);
	loadconstant(RIGHTCONSTANT(expr));
	no_oper(TOP_LSHR);
	depth = INT_DEEP(3);
	}
    else
	{
	loadconstant(RIGHTCONSTANT(expr));
	no_oper(TOP_SHR);
	depth = INT_DEEP(2);
	}
    return (depth);
    }


PRIVATE unsigned int
arsf_g(expr, temp)
  struct op_node		*expr;
  TEMP				temp;
    {
    extern unsigned int		assign_op();
    extern unsigned int		depth_max();
    extern VOID			dvrt_free();
    extern VOID			dvrt_stack();
    extern struct dvrt		*dvrt_unstack();
    extern unsigned int		expr_code();
    extern VOID			free_temp();
    extern TEMP			get_temp();
    extern unsigned int		load_indirect();
    extern VOID			load_temp();
    extern VOID			no_oper();
    extern VOID			store_temp();

    unsigned int		depth;
    TEMP			temp2;

/*
 *	Do "signed" right shifts if necessary.
 */
    if (Cpu_srshift && (RIGHTTYPE(expr) & SIGNED_NODE))
	{
	dvrt_stack();
	depth = expr_code(expr->node_type, RIGHT(expr), NEED_VALUE);
	dvrt_free(dvrt_unstack());

	if ((depth & INT_STACK) > INT_DEEP(1))
	    {
	    depth = depth_max(depth, expr_code(expr->node_type, RIGHT(expr),
				NEED_VALUE));
	    temp2 = get_temp();	/* For saving value */
	    store_temp(Inttype, temp2);
	    depth = depth_max(depth, expr_code(expr->node_type,	LEFT(expr),
				 NEED_VALUE | EVAL_ADDR));
	    if(CPU_DUP)
		no_oper(TOP_DUP);
	    store_temp(Inttype, temp);	/* Save address of target */
	    if(CPU_NO_DUP)
		load_temp(Inttype, temp);
	    /* Ignore return depth - we know total is 3 */
	    (VOID) load_indirect(LEFT(expr), NULL, (INT32) 0, NEED_VALUE);
	    no_oper(TOP_XDBLE);
	    load_temp(Inttype, temp2);
	    free_temp(temp2);
	    }
	else
	    {
	    depth = depth_max(depth, expr_code(NULL, LEFT(expr),
				 NEED_VALUE | EVAL_ADDR));
	    if(CPU_DUP)
		no_oper(TOP_DUP);
	    store_temp(Inttype, temp);
	    if(CPU_NO_DUP)
		load_temp(Inttype, temp);
	    /* Ignore return depth - we know total is 3 */
	    (VOID) load_indirect(LEFT(expr), NULL, (INT32) 0, NEED_VALUE);
	    no_oper(TOP_XDBLE);
	    (VOID) expr_code(expr->node_type, RIGHT(expr), NEED_VALUE);
	    }

	no_oper(TOP_LSHR);
	depth = depth_max(depth, INT_DEEP(3));
	}
/*
 *	Do right shift unsigned (the easy way).
 */
    else
	{
	depth = assign_op(expr, temp);
	no_oper(TOP_SHR);
	depth = INT_DEEP(2);
	}

    return (depth);
    }


PRIVATE	unsigned int
comp_op(expr,vf,dir)
	struct	op_node		*expr;
	int			vf;
	BOOL			dir;
	{
	extern	unsigned int	bin_op();
	extern	VOID		comp_append();
	extern unsigned int	depth_max();
	extern	VOID		dvrt_stack();
	extern	struct	dvrt	*dvrt_unstack();
	extern	unsigned int	expr_code();
	extern	VOID		free_dtemp();
	extern	VOID		free_temp();
	extern	TEMP		get_dtemp();
	extern	TEMP		get_temp();
	extern	VOID		load_temp();
	extern	struct	id_node	*op_types();
	extern	VOID		rev();
	extern	VOID		store_temp();

	unsigned int		deep_lop;
	unsigned int		deep_rop;
	unsigned int		deep_type;
	struct	dvrt		*dvrt_lop;
	struct	dvrt		*dvrt_rop;
	TEMP			temp;
	struct id_node		*types;
	BOOL			usign;
/*
 *	The values loaded are not needed by a higher level, so emit any
 *	code (generated by side-effects) and leave.
 */
	if (! (vf & NEED_VALUE))
		return (bin_op(expr,vf));
/*
 *	Do usual binary conversions to determine sign/unsigned type of
 *	of comparison.
 */
	types = op_types(expr);
	if(types->type & UNSIGNED_NODE)
		usign = TRUE;
	else
		usign = FALSE;
/*
 *	Figure out how much stack each operand will take to evaluate.
 */
	if(types->type & FP_NODE)
		deep_type = FLT_DEEP(1);
	else
		deep_type = INT_DEEP(1);

	if(! (types->type & FP_NODE) && ISLEFTCONSTANT(expr))
		deep_lop = deep_type;
	else
		{
		dvrt_stack();
		deep_lop = expr_code(types, LEFT(expr), vf);
		dvrt_lop = dvrt_unstack();
		if(usign)
			deep_lop = depth_max(deep_lop,2 * deep_type);
		}

	if( !(types->type & FP_NODE) && ISRIGHTCONSTANT(expr))
		deep_rop = deep_type;
	else
		{
		dvrt_stack();
		deep_rop = expr_code(types, RIGHT(expr), vf);
		dvrt_rop = dvrt_unstack();
		if(usign)
			deep_rop = depth_max(deep_rop,2 * deep_type);
		}
/*
 *	Always load diversion which uses most stack space first.
 */
	if(deep_lop < deep_rop)
		{
		comp_append(types,dvrt_rop,RIGHT(expr),usign,vf);
		comp_append(types,dvrt_lop,LEFT(expr),usign,vf);
		if(dir)
			rev(types);
		return (depth_max(deep_lop + deep_type,deep_rop));
		}
	else if(deep_rop < deep_lop)
		{
		comp_append(types,dvrt_lop,LEFT(expr),usign,vf);
		comp_append(types,dvrt_rop,RIGHT(expr),usign,vf);
		if(! dir)
			rev(types);
		return (depth_max(deep_lop,deep_rop + deep_type));
		}
/*
 *	If both diversions use the same amount of stack space, load in the
 *	order which eliminates the extra "rev" instruction.
 */
	else if(deep_lop < (3 * deep_type))
		{
		if(dir)
			{
			comp_append(types,dvrt_lop,LEFT(expr),usign,vf);
			comp_append(types,dvrt_rop,RIGHT(expr),usign,vf);
			return (depth_max(deep_lop,deep_rop + deep_type));
			}
		else
			{
			comp_append(types,dvrt_rop,RIGHT(expr),usign,vf);
			comp_append(types,dvrt_lop,LEFT(expr),usign,vf);
			return (depth_max(deep_lop + deep_type,deep_rop));
			}
		}
/*
 *	Well, both diversions use the full three elements.  Load in the order
 *	which eliminates the extra "rev" instruction, using a "temp".
 */
	if((types->type & FP_NODE) && (types->val.integral > Sizeof_int))
		temp = get_dtemp();
	else
		temp = get_temp();
	if(dir)
		{
		comp_append(types,dvrt_rop,RIGHT(expr),usign,vf);
		store_temp(types, temp);
		comp_append(types,dvrt_lop,LEFT(expr),usign,vf);
		load_temp(types, temp);
		}
	else
		{
		comp_append(types,dvrt_lop,LEFT(expr),usign,vf);
		store_temp(types, temp);
		comp_append(types,dvrt_rop,RIGHT(expr),usign,vf);
		load_temp(types, temp);
		}
	if((types->type & FP_NODE) && (types->val.integral > Sizeof_int))
		free_dtemp(temp);
	else
		free_temp(temp);

	return (depth_max(depth_max(deep_lop, deep_rop), 3 * deep_type));
	}

PRIVATE	VOID
comp_append(types,dvr,expr,usign,vf)
	struct	id_node		*types;
	struct	dvrt		*dvr;
	struct	op_node		*expr;
	BOOL			usign;
	int			vf;
	{
	extern VOID		dvrt_append(),loadconstant();
	extern VOID		dvrt_free();
	extern unsigned int	expr_code();
	extern VOID		no_oper();

	if(!(types->type & FP_NODE) && ISCONSTANT(expr))
		{
		if(usign)
			{
			if(Sizeof_int == WORD_SIZE_16B)
				{
				loadconstant((INT32)
					(expr->right.ptr.terminal->
						val.integral ^ 0x8000L));
				}
			else
				{
				loadconstant((INT32)
					(expr->right.ptr.terminal->
						val.integral ^ 0x80000000L));
				}
			}
		else
			loadconstant((INT32)
				expr->right.ptr.terminal->val.integral);
		}
	else
		{
		dvrt_free(dvr);

		(VOID) expr_code(types, expr, vf);
		if(! (types->type & FP_NODE) && usign)
			{
			no_oper(TOP_MINT);
			no_oper(TOP_XOR);
			}
		}
	}

PUBLIC	unsigned int
do_lt(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	extern	unsigned int	comp_op();
	extern	unsigned int	depth_max();
	extern	VOID		fp_oper();
	extern	VOID		no_oper();
	extern struct id_node	*op_types();

	unsigned int		depth;

	depth = comp_op(expr,value_flag,FALSE);
	if (value_flag & NEED_VALUE)
		{
		fp_oper(op_types(expr),TOP_GT,TOP_FPGT,TOP_FPGT);
		depth = depth_max(depth, INT_DEEP(1));
		}
	return (depth);
	}


PUBLIC	unsigned int
do_gt(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	extern	unsigned int	comp_op();
	extern	unsigned int	depth_max();
	extern	VOID		fp_oper();
	extern	VOID		no_oper();
	extern struct id_node	*op_types();

	unsigned int		depth;

	depth = comp_op(expr,value_flag,TRUE);
	if (value_flag & NEED_VALUE)
		{
		fp_oper(op_types(expr),TOP_GT,TOP_FPGT,TOP_FPGT);
		depth = depth_max(depth, INT_DEEP(1));
		}
	return (depth);
	}


PUBLIC	unsigned int
do_le(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	extern	unsigned int	comp_op();
	extern	unsigned int	depth_max();
	extern	VOID		eqc();
	extern	VOID		fp_oper();
	extern	VOID		no_oper();
	extern struct id_node	*op_types();

	unsigned int		depth;

	depth = comp_op(expr,value_flag,TRUE);
	if (value_flag & NEED_VALUE)
		{
		fp_oper(op_types(expr),TOP_GT,TOP_FPGT,TOP_FPGT);
		depth = depth_max(depth, INT_DEEP(1));
		eqc();
		}
	return (depth);
	}


PUBLIC	unsigned int
do_ge(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	extern	unsigned int	comp_op();
	extern	unsigned int	depth_max();
	extern	VOID		eqc();
	extern	VOID		fp_oper();
	extern	VOID		no_oper();
	extern struct id_node	*op_types();

	unsigned int		depth;

	depth = comp_op(expr,value_flag,FALSE);
	if (value_flag & NEED_VALUE)
		{
		fp_oper(op_types(expr),TOP_GT,TOP_FPGT,TOP_FPGT);
		depth = depth_max(depth, INT_DEEP(1));
		eqc();
		}
	return (depth);
	}


PUBLIC	unsigned int
do_ecomp(expr,value_flag,is_ne)
	struct	op_node		*expr;
	int			value_flag;
	BOOL			is_ne;
	{
	extern	VOID		abs_oper();
	extern	unsigned int	bin_op();
	extern	unsigned int	depth_max();
	extern	VOID		eqc();
	extern	unsigned int	expr_code();
	extern	unsigned int	fp_val();
	extern	INT32		f_to_l();
	extern	VOID		no_oper();
	extern	struct	id_node	*op_types();
	extern	VOID		warning();

	unsigned int		depth;
	struct	id_node		*optype;
	INT32			value;


	if (! (value_flag & NEED_VALUE))
		return (bin_op(expr, value_flag));

	optype = op_types(expr);

	if (ISRIGHTCONSTANT(expr))
		{
		if(LEFTTYPE(expr) & FP_NODE)
			{
			/* Comparison must be floating point */
			depth = expr_code(NULL, LEFT(expr), NEED_VALUE);
			depth = depth_max(depth, fp_val(LEFT(expr)->node_type,
					RIGHT(expr)));
			no_oper(TOP_FPEQ);
	    		if(is_ne)
				eqc();
			depth = depth_max(depth, INT_DEEP(1) | FLT_DEEP(2));
			}
		else
			{
			if((RIGHTTYPE(expr) & FP_NODE))
				value = f_to_l(RIGHTFCONSTANT(expr), TRUE);
			else
				value = RIGHTCONSTANT(expr);
/*
 *	Don't sign extend if child is some char type and value fits
 *	without sign extension.
 */
			if (LEFTTYPE(expr) & CHAR_NODE)
				{
				if (LEFTTYPE(expr) & UNSIGNED_NODE)
					{
					/* Ensure value is between 0 and 255 */
					if((value < 0) || (value > 255))
						{
				warning(1,"Unsigned char compare error");
						}
					}
				else
					{
					/* Ensure value is between -128, 127 */
					if((value < -128) || (value > 127))
						{
					warning(1,"Signed char compare error");
						}
					LEFT(expr)->node_type = Uchartype;
					}
				value = value & 0xFF;
				}
			depth = expr_code(NULL, LEFT(expr), NEED_VALUE);
			if(is_ne)
				{
				if((value != 0) ||
					(! (value_flag & RELAX_BOOL)))
					{
					abs_oper(TOP_EQC, (INT32) value);
					eqc();
					}
				}
			else
				abs_oper(TOP_EQC, (INT32) value);
			}
		}
	else
		{
		depth = bin_op(expr, NEED_VALUE);
		if(optype->type & FP_NODE)
			{
			no_oper(TOP_FPEQ);
			if(is_ne)
				eqc();
			depth = depth_max(depth, INT_DEEP(1));
			}
		else
			{
			no_oper(TOP_DIFF);
			if(is_ne)
				{
				if(! (value_flag & RELAX_BOOL))
					{
					eqc();
					eqc();
					}
				}
			else
				eqc();
			}
		}
	return (depth);
	}

/* End of tcx6c.c */
